//! This file is auto-generated by tools/update_cpu_features.zig.

const std = @import("../std.zig");
const CpuFeature = std.Target.Cpu.Feature;
const CpuModel = std.Target.Cpu.Model;

pub const Feature = enum {
    ptx32,
    ptx40,
    ptx41,
    ptx42,
    ptx43,
    ptx50,
    ptx60,
    ptx61,
    ptx62,
    ptx63,
    ptx64,
    ptx65,
    ptx70,
    ptx71,
    ptx72,
    ptx73,
    ptx74,
    ptx75,
    ptx76,
    ptx77,
    ptx78,
    ptx80,
    ptx81,
    ptx82,
    ptx83,
    ptx84,
    ptx85,
    ptx86,
    ptx87,
    ptx88,
    ptx90,
    sm_100,
    sm_100a,
    sm_100f,
    sm_101,
    sm_101a,
    sm_101f,
    sm_103,
    sm_103a,
    sm_103f,
    sm_110,
    sm_110a,
    sm_110f,
    sm_120,
    sm_120a,
    sm_120f,
    sm_121,
    sm_121a,
    sm_121f,
    sm_20,
    sm_21,
    sm_30,
    sm_32,
    sm_35,
    sm_37,
    sm_50,
    sm_52,
    sm_53,
    sm_60,
    sm_61,
    sm_62,
    sm_70,
    sm_72,
    sm_75,
    sm_80,
    sm_86,
    sm_87,
    sm_88,
    sm_89,
    sm_90,
    sm_90a,
};

pub const featureSet = CpuFeature.FeatureSetFns(Feature).featureSet;
pub const featureSetHas = CpuFeature.FeatureSetFns(Feature).featureSetHas;
pub const featureSetHasAny = CpuFeature.FeatureSetFns(Feature).featureSetHasAny;
pub const featureSetHasAll = CpuFeature.FeatureSetFns(Feature).featureSetHasAll;

pub const all_features = blk: {
    const len = @typeInfo(Feature).@"enum".field_names.len;
    std.debug.assert(len <= CpuFeature.Set.needed_bit_count);
    var result: [len]CpuFeature = undefined;
    result[@backingInt(Feature.ptx32)] = .{
        .llvm_name = "ptx32",
        .description = "Use PTX version 32",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx40)] = .{
        .llvm_name = "ptx40",
        .description = "Use PTX version 40",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx41)] = .{
        .llvm_name = "ptx41",
        .description = "Use PTX version 41",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx42)] = .{
        .llvm_name = "ptx42",
        .description = "Use PTX version 42",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx43)] = .{
        .llvm_name = "ptx43",
        .description = "Use PTX version 43",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx50)] = .{
        .llvm_name = "ptx50",
        .description = "Use PTX version 50",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx60)] = .{
        .llvm_name = "ptx60",
        .description = "Use PTX version 60",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx61)] = .{
        .llvm_name = "ptx61",
        .description = "Use PTX version 61",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx62)] = .{
        .llvm_name = "ptx62",
        .description = "Use PTX version 62",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx63)] = .{
        .llvm_name = "ptx63",
        .description = "Use PTX version 63",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx64)] = .{
        .llvm_name = "ptx64",
        .description = "Use PTX version 64",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx65)] = .{
        .llvm_name = "ptx65",
        .description = "Use PTX version 65",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx70)] = .{
        .llvm_name = "ptx70",
        .description = "Use PTX version 70",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx71)] = .{
        .llvm_name = "ptx71",
        .description = "Use PTX version 71",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx72)] = .{
        .llvm_name = "ptx72",
        .description = "Use PTX version 72",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx73)] = .{
        .llvm_name = "ptx73",
        .description = "Use PTX version 73",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx74)] = .{
        .llvm_name = "ptx74",
        .description = "Use PTX version 74",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx75)] = .{
        .llvm_name = "ptx75",
        .description = "Use PTX version 75",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx76)] = .{
        .llvm_name = "ptx76",
        .description = "Use PTX version 76",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx77)] = .{
        .llvm_name = "ptx77",
        .description = "Use PTX version 77",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx78)] = .{
        .llvm_name = "ptx78",
        .description = "Use PTX version 78",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx80)] = .{
        .llvm_name = "ptx80",
        .description = "Use PTX version 80",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx81)] = .{
        .llvm_name = "ptx81",
        .description = "Use PTX version 81",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx82)] = .{
        .llvm_name = "ptx82",
        .description = "Use PTX version 82",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx83)] = .{
        .llvm_name = "ptx83",
        .description = "Use PTX version 83",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx84)] = .{
        .llvm_name = "ptx84",
        .description = "Use PTX version 84",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx85)] = .{
        .llvm_name = "ptx85",
        .description = "Use PTX version 85",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx86)] = .{
        .llvm_name = "ptx86",
        .description = "Use PTX version 86",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx87)] = .{
        .llvm_name = "ptx87",
        .description = "Use PTX version 87",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx88)] = .{
        .llvm_name = "ptx88",
        .description = "Use PTX version 88",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.ptx90)] = .{
        .llvm_name = "ptx90",
        .description = "Use PTX version 90",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_100)] = .{
        .llvm_name = "sm_100",
        .description = "Target SM 100",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_100a)] = .{
        .llvm_name = "sm_100a",
        .description = "Target SM 100a",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_100f)] = .{
        .llvm_name = "sm_100f",
        .description = "Target SM 100f",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_101)] = .{
        .llvm_name = "sm_101",
        .description = "Target SM 101",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_101a)] = .{
        .llvm_name = "sm_101a",
        .description = "Target SM 101a",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_101f)] = .{
        .llvm_name = "sm_101f",
        .description = "Target SM 101f",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_103)] = .{
        .llvm_name = "sm_103",
        .description = "Target SM 103",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_103a)] = .{
        .llvm_name = "sm_103a",
        .description = "Target SM 103a",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_103f)] = .{
        .llvm_name = "sm_103f",
        .description = "Target SM 103f",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_110)] = .{
        .llvm_name = "sm_110",
        .description = "Target SM 110",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_110a)] = .{
        .llvm_name = "sm_110a",
        .description = "Target SM 110a",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_110f)] = .{
        .llvm_name = "sm_110f",
        .description = "Target SM 110f",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_120)] = .{
        .llvm_name = "sm_120",
        .description = "Target SM 120",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_120a)] = .{
        .llvm_name = "sm_120a",
        .description = "Target SM 120a",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_120f)] = .{
        .llvm_name = "sm_120f",
        .description = "Target SM 120f",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_121)] = .{
        .llvm_name = "sm_121",
        .description = "Target SM 121",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_121a)] = .{
        .llvm_name = "sm_121a",
        .description = "Target SM 121a",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_121f)] = .{
        .llvm_name = "sm_121f",
        .description = "Target SM 121f",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_20)] = .{
        .llvm_name = "sm_20",
        .description = "Target SM 20",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_21)] = .{
        .llvm_name = "sm_21",
        .description = "Target SM 21",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_30)] = .{
        .llvm_name = "sm_30",
        .description = "Target SM 30",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_32)] = .{
        .llvm_name = "sm_32",
        .description = "Target SM 32",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_35)] = .{
        .llvm_name = "sm_35",
        .description = "Target SM 35",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_37)] = .{
        .llvm_name = "sm_37",
        .description = "Target SM 37",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_50)] = .{
        .llvm_name = "sm_50",
        .description = "Target SM 50",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_52)] = .{
        .llvm_name = "sm_52",
        .description = "Target SM 52",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_53)] = .{
        .llvm_name = "sm_53",
        .description = "Target SM 53",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_60)] = .{
        .llvm_name = "sm_60",
        .description = "Target SM 60",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_61)] = .{
        .llvm_name = "sm_61",
        .description = "Target SM 61",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_62)] = .{
        .llvm_name = "sm_62",
        .description = "Target SM 62",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_70)] = .{
        .llvm_name = "sm_70",
        .description = "Target SM 70",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_72)] = .{
        .llvm_name = "sm_72",
        .description = "Target SM 72",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_75)] = .{
        .llvm_name = "sm_75",
        .description = "Target SM 75",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_80)] = .{
        .llvm_name = "sm_80",
        .description = "Target SM 80",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_86)] = .{
        .llvm_name = "sm_86",
        .description = "Target SM 86",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_87)] = .{
        .llvm_name = "sm_87",
        .description = "Target SM 87",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_88)] = .{
        .llvm_name = "sm_88",
        .description = "Target SM 88",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_89)] = .{
        .llvm_name = "sm_89",
        .description = "Target SM 89",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_90)] = .{
        .llvm_name = "sm_90",
        .description = "Target SM 90",
        .dependencies = featureSet(&[_]Feature{}),
    };
    result[@backingInt(Feature.sm_90a)] = .{
        .llvm_name = "sm_90a",
        .description = "Target SM 90a",
        .dependencies = featureSet(&[_]Feature{}),
    };
    const ti = @typeInfo(Feature);
    for (&result, 0..) |*elem, i| {
        elem.index = i;
        elem.name = ti.@"enum".field_names[i];
    }
    break :blk result;
};

pub const cpu = struct {
    pub const sm_100: CpuModel = .{
        .name = "sm_100",
        .llvm_name = "sm_100",
        .features = featureSet(&[_]Feature{
            .sm_100,
        }),
    };
    pub const sm_100a: CpuModel = .{
        .name = "sm_100a",
        .llvm_name = "sm_100a",
        .features = featureSet(&[_]Feature{
            .sm_100a,
        }),
    };
    pub const sm_100f: CpuModel = .{
        .name = "sm_100f",
        .llvm_name = "sm_100f",
        .features = featureSet(&[_]Feature{
            .sm_100f,
        }),
    };
    pub const sm_101: CpuModel = .{
        .name = "sm_101",
        .llvm_name = "sm_101",
        .features = featureSet(&[_]Feature{
            .sm_101,
        }),
    };
    pub const sm_101a: CpuModel = .{
        .name = "sm_101a",
        .llvm_name = "sm_101a",
        .features = featureSet(&[_]Feature{
            .sm_101a,
        }),
    };
    pub const sm_101f: CpuModel = .{
        .name = "sm_101f",
        .llvm_name = "sm_101f",
        .features = featureSet(&[_]Feature{
            .sm_101f,
        }),
    };
    pub const sm_103: CpuModel = .{
        .name = "sm_103",
        .llvm_name = "sm_103",
        .features = featureSet(&[_]Feature{
            .sm_103,
        }),
    };
    pub const sm_103a: CpuModel = .{
        .name = "sm_103a",
        .llvm_name = "sm_103a",
        .features = featureSet(&[_]Feature{
            .sm_103a,
        }),
    };
    pub const sm_103f: CpuModel = .{
        .name = "sm_103f",
        .llvm_name = "sm_103f",
        .features = featureSet(&[_]Feature{
            .sm_103f,
        }),
    };
    pub const sm_110: CpuModel = .{
        .name = "sm_110",
        .llvm_name = "sm_110",
        .features = featureSet(&[_]Feature{
            .sm_110,
        }),
    };
    pub const sm_110a: CpuModel = .{
        .name = "sm_110a",
        .llvm_name = "sm_110a",
        .features = featureSet(&[_]Feature{
            .sm_110a,
        }),
    };
    pub const sm_110f: CpuModel = .{
        .name = "sm_110f",
        .llvm_name = "sm_110f",
        .features = featureSet(&[_]Feature{
            .sm_110f,
        }),
    };
    pub const sm_120: CpuModel = .{
        .name = "sm_120",
        .llvm_name = "sm_120",
        .features = featureSet(&[_]Feature{
            .sm_120,
        }),
    };
    pub const sm_120a: CpuModel = .{
        .name = "sm_120a",
        .llvm_name = "sm_120a",
        .features = featureSet(&[_]Feature{
            .sm_120a,
        }),
    };
    pub const sm_120f: CpuModel = .{
        .name = "sm_120f",
        .llvm_name = "sm_120f",
        .features = featureSet(&[_]Feature{
            .sm_120f,
        }),
    };
    pub const sm_121: CpuModel = .{
        .name = "sm_121",
        .llvm_name = "sm_121",
        .features = featureSet(&[_]Feature{
            .sm_121,
        }),
    };
    pub const sm_121a: CpuModel = .{
        .name = "sm_121a",
        .llvm_name = "sm_121a",
        .features = featureSet(&[_]Feature{
            .sm_121a,
        }),
    };
    pub const sm_121f: CpuModel = .{
        .name = "sm_121f",
        .llvm_name = "sm_121f",
        .features = featureSet(&[_]Feature{
            .sm_121f,
        }),
    };
    pub const sm_20: CpuModel = .{
        .name = "sm_20",
        .llvm_name = "sm_20",
        .features = featureSet(&[_]Feature{
            .sm_20,
        }),
    };
    pub const sm_21: CpuModel = .{
        .name = "sm_21",
        .llvm_name = "sm_21",
        .features = featureSet(&[_]Feature{
            .sm_21,
        }),
    };
    pub const sm_30: CpuModel = .{
        .name = "sm_30",
        .llvm_name = "sm_30",
        .features = featureSet(&[_]Feature{
            .sm_30,
        }),
    };
    pub const sm_32: CpuModel = .{
        .name = "sm_32",
        .llvm_name = "sm_32",
        .features = featureSet(&[_]Feature{
            .sm_32,
        }),
    };
    pub const sm_35: CpuModel = .{
        .name = "sm_35",
        .llvm_name = "sm_35",
        .features = featureSet(&[_]Feature{
            .sm_35,
        }),
    };
    pub const sm_37: CpuModel = .{
        .name = "sm_37",
        .llvm_name = "sm_37",
        .features = featureSet(&[_]Feature{
            .sm_37,
        }),
    };
    pub const sm_50: CpuModel = .{
        .name = "sm_50",
        .llvm_name = "sm_50",
        .features = featureSet(&[_]Feature{
            .sm_50,
        }),
    };
    pub const sm_52: CpuModel = .{
        .name = "sm_52",
        .llvm_name = "sm_52",
        .features = featureSet(&[_]Feature{
            .sm_52,
        }),
    };
    pub const sm_53: CpuModel = .{
        .name = "sm_53",
        .llvm_name = "sm_53",
        .features = featureSet(&[_]Feature{
            .sm_53,
        }),
    };
    pub const sm_60: CpuModel = .{
        .name = "sm_60",
        .llvm_name = "sm_60",
        .features = featureSet(&[_]Feature{
            .sm_60,
        }),
    };
    pub const sm_61: CpuModel = .{
        .name = "sm_61",
        .llvm_name = "sm_61",
        .features = featureSet(&[_]Feature{
            .sm_61,
        }),
    };
    pub const sm_62: CpuModel = .{
        .name = "sm_62",
        .llvm_name = "sm_62",
        .features = featureSet(&[_]Feature{
            .sm_62,
        }),
    };
    pub const sm_70: CpuModel = .{
        .name = "sm_70",
        .llvm_name = "sm_70",
        .features = featureSet(&[_]Feature{
            .sm_70,
        }),
    };
    pub const sm_72: CpuModel = .{
        .name = "sm_72",
        .llvm_name = "sm_72",
        .features = featureSet(&[_]Feature{
            .sm_72,
        }),
    };
    pub const sm_75: CpuModel = .{
        .name = "sm_75",
        .llvm_name = "sm_75",
        .features = featureSet(&[_]Feature{
            .sm_75,
        }),
    };
    pub const sm_80: CpuModel = .{
        .name = "sm_80",
        .llvm_name = "sm_80",
        .features = featureSet(&[_]Feature{
            .sm_80,
        }),
    };
    pub const sm_86: CpuModel = .{
        .name = "sm_86",
        .llvm_name = "sm_86",
        .features = featureSet(&[_]Feature{
            .sm_86,
        }),
    };
    pub const sm_87: CpuModel = .{
        .name = "sm_87",
        .llvm_name = "sm_87",
        .features = featureSet(&[_]Feature{
            .sm_87,
        }),
    };
    pub const sm_88: CpuModel = .{
        .name = "sm_88",
        .llvm_name = "sm_88",
        .features = featureSet(&[_]Feature{
            .sm_88,
        }),
    };
    pub const sm_89: CpuModel = .{
        .name = "sm_89",
        .llvm_name = "sm_89",
        .features = featureSet(&[_]Feature{
            .sm_89,
        }),
    };
    pub const sm_90: CpuModel = .{
        .name = "sm_90",
        .llvm_name = "sm_90",
        .features = featureSet(&[_]Feature{
            .sm_90,
        }),
    };
    pub const sm_90a: CpuModel = .{
        .name = "sm_90a",
        .llvm_name = "sm_90a",
        .features = featureSet(&[_]Feature{
            .sm_90a,
        }),
    };
};
