pub const C89 = 0x0001;
pub const C = 0x0002;
pub const Ada83 = 0x0003;
pub const C_plus_plus = 0x0004;
pub const Cobol74 = 0x0005;
pub const Cobol85 = 0x0006;
pub const Fortran77 = 0x0007;
pub const Fortran90 = 0x0008;
pub const Pascal83 = 0x0009;
pub const Modula2 = 0x000a;
pub const Java = 0x000b;
pub const C99 = 0x000c;
pub const Ada95 = 0x000d;
pub const Fortran95 = 0x000e;
pub const PLI = 0x000f;
pub const ObjC = 0x0010;
pub const ObjC_plus_plus = 0x0011;
pub const UPC = 0x0012;
pub const D = 0x0013;
pub const Python = 0x0014;
pub const OpenCL = 0x0015;
pub const Go = 0x0016;
pub const Modula3 = 0x0017;
pub const Haskell = 0x0018;
pub const C_plus_plus_03 = 0x0019;
pub const C_plus_plus_11 = 0x001a;
pub const OCaml = 0x001b;
pub const Rust = 0x001c;
pub const C11 = 0x001d;
pub const Swift = 0x001e;
pub const Julia = 0x001f;
pub const Dylan = 0x0020;
pub const C_plus_plus_14 = 0x0021;
pub const Fortran03 = 0x0022;
pub const Fortran08 = 0x0023;
pub const RenderScript = 0x0024;
pub const BLISS = 0x0025;
pub const Kotlin = 0x0026;
pub const Zig = 0x0027;
pub const Crystal = 0x0028;
pub const C_plus_plus_17 = 0x002a;
pub const C_plus_plus_20 = 0x002b;
pub const C17 = 0x002c;
pub const Fortran18 = 0x002d;
pub const Ada2005 = 0x002e;
pub const Ada2012 = 0x002f;
pub const HIP = 0x0030;
pub const Assembly = 0x0031;
pub const C_sharp = 0x0032;
pub const Mojo = 0x0033;
pub const GLSL = 0x0034;
pub const GLSL_ES = 0x0035;
pub const HLSL = 0x0036;
pub const OpenCL_CPP = 0x0037;
pub const CPP_for_OpenCL = 0x0038;
pub const SYCL = 0x0039;
pub const C_plus_plus_23 = 0x003a;
pub const Odin = 0x003b;
pub const Ruby = 0x0040;
pub const Move = 0x0041;
pub const Hylo = 0x0042;

pub const lo_user = 0x8000;
pub const hi_user = 0xffff;

pub const Mips_Assembler = 0x8001;
pub const Upc = 0x8765;
pub const HP_Bliss = 0x8003;
pub const HP_Basic91 = 0x8004;
pub const HP_Pascal91 = 0x8005;
pub const HP_IMacro = 0x8006;
pub const HP_Assembler = 0x8007;
