// This file is automatically generated by tools/generate_linux_syscalls.zig
// This list current as of kernel: 7.1.0

pub const X86 = enum(usize) {
    restart_syscall = 0,
    exit = 1,
    fork = 2,
    read = 3,
    write = 4,
    open = 5,
    close = 6,
    waitpid = 7,
    creat = 8,
    link = 9,
    unlink = 10,
    execve = 11,
    chdir = 12,
    time = 13,
    mknod = 14,
    chmod = 15,
    lchown = 16,
    @"break" = 17,
    oldstat = 18,
    lseek = 19,
    getpid = 20,
    mount = 21,
    umount = 22,
    setuid = 23,
    getuid = 24,
    stime = 25,
    ptrace = 26,
    alarm = 27,
    oldfstat = 28,
    pause = 29,
    utime = 30,
    stty = 31,
    gtty = 32,
    access = 33,
    nice = 34,
    ftime = 35,
    sync = 36,
    kill = 37,
    rename = 38,
    mkdir = 39,
    rmdir = 40,
    dup = 41,
    pipe = 42,
    times = 43,
    prof = 44,
    brk = 45,
    setgid = 46,
    getgid = 47,
    signal = 48,
    geteuid = 49,
    getegid = 50,
    acct = 51,
    umount2 = 52,
    lock = 53,
    ioctl = 54,
    fcntl = 55,
    mpx = 56,
    setpgid = 57,
    ulimit = 58,
    oldolduname = 59,
    umask = 60,
    chroot = 61,
    ustat = 62,
    dup2 = 63,
    getppid = 64,
    getpgrp = 65,
    setsid = 66,
    sigaction = 67,
    sgetmask = 68,
    ssetmask = 69,
    setreuid = 70,
    setregid = 71,
    sigsuspend = 72,
    sigpending = 73,
    sethostname = 74,
    setrlimit = 75,
    getrlimit = 76,
    getrusage = 77,
    gettimeofday = 78,
    settimeofday = 79,
    getgroups = 80,
    setgroups = 81,
    select = 82,
    symlink = 83,
    oldlstat = 84,
    readlink = 85,
    uselib = 86,
    swapon = 87,
    reboot = 88,
    readdir = 89,
    mmap = 90,
    munmap = 91,
    truncate = 92,
    ftruncate = 93,
    fchmod = 94,
    fchown = 95,
    getpriority = 96,
    setpriority = 97,
    profil = 98,
    statfs = 99,
    fstatfs = 100,
    ioperm = 101,
    socketcall = 102,
    syslog = 103,
    setitimer = 104,
    getitimer = 105,
    stat = 106,
    lstat = 107,
    fstat = 108,
    olduname = 109,
    iopl = 110,
    vhangup = 111,
    idle = 112,
    vm86old = 113,
    wait4 = 114,
    swapoff = 115,
    sysinfo = 116,
    ipc = 117,
    fsync = 118,
    sigreturn = 119,
    clone = 120,
    setdomainname = 121,
    uname = 122,
    modify_ldt = 123,
    adjtimex = 124,
    mprotect = 125,
    sigprocmask = 126,
    create_module = 127,
    init_module = 128,
    delete_module = 129,
    get_kernel_syms = 130,
    quotactl = 131,
    getpgid = 132,
    fchdir = 133,
    bdflush = 134,
    sysfs = 135,
    personality = 136,
    afs_syscall = 137,
    setfsuid = 138,
    setfsgid = 139,
    llseek = 140,
    getdents = 141,
    newselect = 142,
    flock = 143,
    msync = 144,
    readv = 145,
    writev = 146,
    getsid = 147,
    fdatasync = 148,
    sysctl = 149,
    mlock = 150,
    munlock = 151,
    mlockall = 152,
    munlockall = 153,
    sched_setparam = 154,
    sched_getparam = 155,
    sched_setscheduler = 156,
    sched_getscheduler = 157,
    sched_yield = 158,
    sched_get_priority_max = 159,
    sched_get_priority_min = 160,
    sched_rr_get_interval = 161,
    nanosleep = 162,
    mremap = 163,
    setresuid = 164,
    getresuid = 165,
    vm86 = 166,
    query_module = 167,
    poll = 168,
    nfsservctl = 169,
    setresgid = 170,
    getresgid = 171,
    prctl = 172,
    rt_sigreturn = 173,
    rt_sigaction = 174,
    rt_sigprocmask = 175,
    rt_sigpending = 176,
    rt_sigtimedwait = 177,
    rt_sigqueueinfo = 178,
    rt_sigsuspend = 179,
    pread64 = 180,
    pwrite64 = 181,
    chown = 182,
    getcwd = 183,
    capget = 184,
    capset = 185,
    sigaltstack = 186,
    sendfile = 187,
    getpmsg = 188,
    putpmsg = 189,
    vfork = 190,
    ugetrlimit = 191,
    mmap2 = 192,
    truncate64 = 193,
    ftruncate64 = 194,
    stat64 = 195,
    lstat64 = 196,
    fstat64 = 197,
    lchown32 = 198,
    getuid32 = 199,
    getgid32 = 200,
    geteuid32 = 201,
    getegid32 = 202,
    setreuid32 = 203,
    setregid32 = 204,
    getgroups32 = 205,
    setgroups32 = 206,
    fchown32 = 207,
    setresuid32 = 208,
    getresuid32 = 209,
    setresgid32 = 210,
    getresgid32 = 211,
    chown32 = 212,
    setuid32 = 213,
    setgid32 = 214,
    setfsuid32 = 215,
    setfsgid32 = 216,
    pivot_root = 217,
    mincore = 218,
    madvise = 219,
    getdents64 = 220,
    fcntl64 = 221,
    gettid = 224,
    readahead = 225,
    setxattr = 226,
    lsetxattr = 227,
    fsetxattr = 228,
    getxattr = 229,
    lgetxattr = 230,
    fgetxattr = 231,
    listxattr = 232,
    llistxattr = 233,
    flistxattr = 234,
    removexattr = 235,
    lremovexattr = 236,
    fremovexattr = 237,
    tkill = 238,
    sendfile64 = 239,
    futex = 240,
    sched_setaffinity = 241,
    sched_getaffinity = 242,
    set_thread_area = 243,
    get_thread_area = 244,
    io_setup = 245,
    io_destroy = 246,
    io_getevents = 247,
    io_submit = 248,
    io_cancel = 249,
    fadvise64 = 250,
    exit_group = 252,
    lookup_dcookie = 253,
    epoll_create = 254,
    epoll_ctl = 255,
    epoll_wait = 256,
    remap_file_pages = 257,
    set_tid_address = 258,
    timer_create = 259,
    timer_settime = 260,
    timer_gettime = 261,
    timer_getoverrun = 262,
    timer_delete = 263,
    clock_settime = 264,
    clock_gettime = 265,
    clock_getres = 266,
    clock_nanosleep = 267,
    statfs64 = 268,
    fstatfs64 = 269,
    tgkill = 270,
    utimes = 271,
    fadvise64_64 = 272,
    vserver = 273,
    mbind = 274,
    get_mempolicy = 275,
    set_mempolicy = 276,
    mq_open = 277,
    mq_unlink = 278,
    mq_timedsend = 279,
    mq_timedreceive = 280,
    mq_notify = 281,
    mq_getsetattr = 282,
    kexec_load = 283,
    waitid = 284,
    add_key = 286,
    request_key = 287,
    keyctl = 288,
    ioprio_set = 289,
    ioprio_get = 290,
    inotify_init = 291,
    inotify_add_watch = 292,
    inotify_rm_watch = 293,
    migrate_pages = 294,
    openat = 295,
    mkdirat = 296,
    mknodat = 297,
    fchownat = 298,
    futimesat = 299,
    fstatat64 = 300,
    unlinkat = 301,
    renameat = 302,
    linkat = 303,
    symlinkat = 304,
    readlinkat = 305,
    fchmodat = 306,
    faccessat = 307,
    pselect6 = 308,
    ppoll = 309,
    unshare = 310,
    set_robust_list = 311,
    get_robust_list = 312,
    splice = 313,
    sync_file_range = 314,
    tee = 315,
    vmsplice = 316,
    move_pages = 317,
    getcpu = 318,
    epoll_pwait = 319,
    utimensat = 320,
    signalfd = 321,
    timerfd_create = 322,
    eventfd = 323,
    fallocate = 324,
    timerfd_settime = 325,
    timerfd_gettime = 326,
    signalfd4 = 327,
    eventfd2 = 328,
    epoll_create1 = 329,
    dup3 = 330,
    pipe2 = 331,
    inotify_init1 = 332,
    preadv = 333,
    pwritev = 334,
    rt_tgsigqueueinfo = 335,
    perf_event_open = 336,
    recvmmsg = 337,
    fanotify_init = 338,
    fanotify_mark = 339,
    prlimit64 = 340,
    name_to_handle_at = 341,
    open_by_handle_at = 342,
    clock_adjtime = 343,
    syncfs = 344,
    sendmmsg = 345,
    setns = 346,
    process_vm_readv = 347,
    process_vm_writev = 348,
    kcmp = 349,
    finit_module = 350,
    sched_setattr = 351,
    sched_getattr = 352,
    renameat2 = 353,
    seccomp = 354,
    getrandom = 355,
    memfd_create = 356,
    bpf = 357,
    execveat = 358,
    socket = 359,
    socketpair = 360,
    bind = 361,
    connect = 362,
    listen = 363,
    accept4 = 364,
    getsockopt = 365,
    setsockopt = 366,
    getsockname = 367,
    getpeername = 368,
    sendto = 369,
    sendmsg = 370,
    recvfrom = 371,
    recvmsg = 372,
    shutdown = 373,
    userfaultfd = 374,
    membarrier = 375,
    mlock2 = 376,
    copy_file_range = 377,
    preadv2 = 378,
    pwritev2 = 379,
    pkey_mprotect = 380,
    pkey_alloc = 381,
    pkey_free = 382,
    statx = 383,
    arch_prctl = 384,
    io_pgetevents = 385,
    rseq = 386,
    semget = 393,
    semctl = 394,
    shmget = 395,
    shmctl = 396,
    shmat = 397,
    shmdt = 398,
    msgget = 399,
    msgsnd = 400,
    msgrcv = 401,
    msgctl = 402,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    memfd_secret = 447,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const X64 = enum(usize) {
    read = 0,
    write = 1,
    open = 2,
    close = 3,
    stat = 4,
    fstat = 5,
    lstat = 6,
    poll = 7,
    lseek = 8,
    mmap = 9,
    mprotect = 10,
    munmap = 11,
    brk = 12,
    rt_sigaction = 13,
    rt_sigprocmask = 14,
    rt_sigreturn = 15,
    ioctl = 16,
    pread64 = 17,
    pwrite64 = 18,
    readv = 19,
    writev = 20,
    access = 21,
    pipe = 22,
    select = 23,
    sched_yield = 24,
    mremap = 25,
    msync = 26,
    mincore = 27,
    madvise = 28,
    shmget = 29,
    shmat = 30,
    shmctl = 31,
    dup = 32,
    dup2 = 33,
    pause = 34,
    nanosleep = 35,
    getitimer = 36,
    alarm = 37,
    setitimer = 38,
    getpid = 39,
    sendfile = 40,
    socket = 41,
    connect = 42,
    accept = 43,
    sendto = 44,
    recvfrom = 45,
    sendmsg = 46,
    recvmsg = 47,
    shutdown = 48,
    bind = 49,
    listen = 50,
    getsockname = 51,
    getpeername = 52,
    socketpair = 53,
    setsockopt = 54,
    getsockopt = 55,
    clone = 56,
    fork = 57,
    vfork = 58,
    execve = 59,
    exit = 60,
    wait4 = 61,
    kill = 62,
    uname = 63,
    semget = 64,
    semop = 65,
    semctl = 66,
    shmdt = 67,
    msgget = 68,
    msgsnd = 69,
    msgrcv = 70,
    msgctl = 71,
    fcntl = 72,
    flock = 73,
    fsync = 74,
    fdatasync = 75,
    truncate = 76,
    ftruncate = 77,
    getdents = 78,
    getcwd = 79,
    chdir = 80,
    fchdir = 81,
    rename = 82,
    mkdir = 83,
    rmdir = 84,
    creat = 85,
    link = 86,
    unlink = 87,
    symlink = 88,
    readlink = 89,
    chmod = 90,
    fchmod = 91,
    chown = 92,
    fchown = 93,
    lchown = 94,
    umask = 95,
    gettimeofday = 96,
    getrlimit = 97,
    getrusage = 98,
    sysinfo = 99,
    times = 100,
    ptrace = 101,
    getuid = 102,
    syslog = 103,
    getgid = 104,
    setuid = 105,
    setgid = 106,
    geteuid = 107,
    getegid = 108,
    setpgid = 109,
    getppid = 110,
    getpgrp = 111,
    setsid = 112,
    setreuid = 113,
    setregid = 114,
    getgroups = 115,
    setgroups = 116,
    setresuid = 117,
    getresuid = 118,
    setresgid = 119,
    getresgid = 120,
    getpgid = 121,
    setfsuid = 122,
    setfsgid = 123,
    getsid = 124,
    capget = 125,
    capset = 126,
    rt_sigpending = 127,
    rt_sigtimedwait = 128,
    rt_sigqueueinfo = 129,
    rt_sigsuspend = 130,
    sigaltstack = 131,
    utime = 132,
    mknod = 133,
    uselib = 134,
    personality = 135,
    ustat = 136,
    statfs = 137,
    fstatfs = 138,
    sysfs = 139,
    getpriority = 140,
    setpriority = 141,
    sched_setparam = 142,
    sched_getparam = 143,
    sched_setscheduler = 144,
    sched_getscheduler = 145,
    sched_get_priority_max = 146,
    sched_get_priority_min = 147,
    sched_rr_get_interval = 148,
    mlock = 149,
    munlock = 150,
    mlockall = 151,
    munlockall = 152,
    vhangup = 153,
    modify_ldt = 154,
    pivot_root = 155,
    sysctl = 156,
    prctl = 157,
    arch_prctl = 158,
    adjtimex = 159,
    setrlimit = 160,
    chroot = 161,
    sync = 162,
    acct = 163,
    settimeofday = 164,
    mount = 165,
    umount2 = 166,
    swapon = 167,
    swapoff = 168,
    reboot = 169,
    sethostname = 170,
    setdomainname = 171,
    iopl = 172,
    ioperm = 173,
    create_module = 174,
    init_module = 175,
    delete_module = 176,
    get_kernel_syms = 177,
    query_module = 178,
    quotactl = 179,
    nfsservctl = 180,
    getpmsg = 181,
    putpmsg = 182,
    afs_syscall = 183,
    tuxcall = 184,
    security = 185,
    gettid = 186,
    readahead = 187,
    setxattr = 188,
    lsetxattr = 189,
    fsetxattr = 190,
    getxattr = 191,
    lgetxattr = 192,
    fgetxattr = 193,
    listxattr = 194,
    llistxattr = 195,
    flistxattr = 196,
    removexattr = 197,
    lremovexattr = 198,
    fremovexattr = 199,
    tkill = 200,
    time = 201,
    futex = 202,
    sched_setaffinity = 203,
    sched_getaffinity = 204,
    set_thread_area = 205,
    io_setup = 206,
    io_destroy = 207,
    io_getevents = 208,
    io_submit = 209,
    io_cancel = 210,
    get_thread_area = 211,
    lookup_dcookie = 212,
    epoll_create = 213,
    epoll_ctl_old = 214,
    epoll_wait_old = 215,
    remap_file_pages = 216,
    getdents64 = 217,
    set_tid_address = 218,
    restart_syscall = 219,
    semtimedop = 220,
    fadvise64 = 221,
    timer_create = 222,
    timer_settime = 223,
    timer_gettime = 224,
    timer_getoverrun = 225,
    timer_delete = 226,
    clock_settime = 227,
    clock_gettime = 228,
    clock_getres = 229,
    clock_nanosleep = 230,
    exit_group = 231,
    epoll_wait = 232,
    epoll_ctl = 233,
    tgkill = 234,
    utimes = 235,
    vserver = 236,
    mbind = 237,
    set_mempolicy = 238,
    get_mempolicy = 239,
    mq_open = 240,
    mq_unlink = 241,
    mq_timedsend = 242,
    mq_timedreceive = 243,
    mq_notify = 244,
    mq_getsetattr = 245,
    kexec_load = 246,
    waitid = 247,
    add_key = 248,
    request_key = 249,
    keyctl = 250,
    ioprio_set = 251,
    ioprio_get = 252,
    inotify_init = 253,
    inotify_add_watch = 254,
    inotify_rm_watch = 255,
    migrate_pages = 256,
    openat = 257,
    mkdirat = 258,
    mknodat = 259,
    fchownat = 260,
    futimesat = 261,
    newfstatat = 262,
    unlinkat = 263,
    renameat = 264,
    linkat = 265,
    symlinkat = 266,
    readlinkat = 267,
    fchmodat = 268,
    faccessat = 269,
    pselect6 = 270,
    ppoll = 271,
    unshare = 272,
    set_robust_list = 273,
    get_robust_list = 274,
    splice = 275,
    tee = 276,
    sync_file_range = 277,
    vmsplice = 278,
    move_pages = 279,
    utimensat = 280,
    epoll_pwait = 281,
    signalfd = 282,
    timerfd_create = 283,
    eventfd = 284,
    fallocate = 285,
    timerfd_settime = 286,
    timerfd_gettime = 287,
    accept4 = 288,
    signalfd4 = 289,
    eventfd2 = 290,
    epoll_create1 = 291,
    dup3 = 292,
    pipe2 = 293,
    inotify_init1 = 294,
    preadv = 295,
    pwritev = 296,
    rt_tgsigqueueinfo = 297,
    perf_event_open = 298,
    recvmmsg = 299,
    fanotify_init = 300,
    fanotify_mark = 301,
    prlimit64 = 302,
    name_to_handle_at = 303,
    open_by_handle_at = 304,
    clock_adjtime = 305,
    syncfs = 306,
    sendmmsg = 307,
    setns = 308,
    getcpu = 309,
    process_vm_readv = 310,
    process_vm_writev = 311,
    kcmp = 312,
    finit_module = 313,
    sched_setattr = 314,
    sched_getattr = 315,
    renameat2 = 316,
    seccomp = 317,
    getrandom = 318,
    memfd_create = 319,
    kexec_file_load = 320,
    bpf = 321,
    execveat = 322,
    userfaultfd = 323,
    membarrier = 324,
    mlock2 = 325,
    copy_file_range = 326,
    preadv2 = 327,
    pwritev2 = 328,
    pkey_mprotect = 329,
    pkey_alloc = 330,
    pkey_free = 331,
    statx = 332,
    io_pgetevents = 333,
    rseq = 334,
    uretprobe = 335,
    uprobe = 336,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    memfd_secret = 447,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const X32 = enum(usize) {
    read = 1073741824,
    write = 1073741825,
    open = 1073741826,
    close = 1073741827,
    stat = 1073741828,
    fstat = 1073741829,
    lstat = 1073741830,
    poll = 1073741831,
    lseek = 1073741832,
    mmap = 1073741833,
    mprotect = 1073741834,
    munmap = 1073741835,
    brk = 1073741836,
    rt_sigprocmask = 1073741838,
    pread64 = 1073741841,
    pwrite64 = 1073741842,
    access = 1073741845,
    pipe = 1073741846,
    select = 1073741847,
    sched_yield = 1073741848,
    mremap = 1073741849,
    msync = 1073741850,
    mincore = 1073741851,
    madvise = 1073741852,
    shmget = 1073741853,
    shmat = 1073741854,
    shmctl = 1073741855,
    dup = 1073741856,
    dup2 = 1073741857,
    pause = 1073741858,
    nanosleep = 1073741859,
    getitimer = 1073741860,
    alarm = 1073741861,
    setitimer = 1073741862,
    getpid = 1073741863,
    sendfile = 1073741864,
    socket = 1073741865,
    connect = 1073741866,
    accept = 1073741867,
    sendto = 1073741868,
    shutdown = 1073741872,
    bind = 1073741873,
    listen = 1073741874,
    getsockname = 1073741875,
    getpeername = 1073741876,
    socketpair = 1073741877,
    clone = 1073741880,
    fork = 1073741881,
    vfork = 1073741882,
    exit = 1073741884,
    wait4 = 1073741885,
    kill = 1073741886,
    uname = 1073741887,
    semget = 1073741888,
    semop = 1073741889,
    semctl = 1073741890,
    shmdt = 1073741891,
    msgget = 1073741892,
    msgsnd = 1073741893,
    msgrcv = 1073741894,
    msgctl = 1073741895,
    fcntl = 1073741896,
    flock = 1073741897,
    fsync = 1073741898,
    fdatasync = 1073741899,
    truncate = 1073741900,
    ftruncate = 1073741901,
    getdents = 1073741902,
    getcwd = 1073741903,
    chdir = 1073741904,
    fchdir = 1073741905,
    rename = 1073741906,
    mkdir = 1073741907,
    rmdir = 1073741908,
    creat = 1073741909,
    link = 1073741910,
    unlink = 1073741911,
    symlink = 1073741912,
    readlink = 1073741913,
    chmod = 1073741914,
    fchmod = 1073741915,
    chown = 1073741916,
    fchown = 1073741917,
    lchown = 1073741918,
    umask = 1073741919,
    gettimeofday = 1073741920,
    getrlimit = 1073741921,
    getrusage = 1073741922,
    sysinfo = 1073741923,
    times = 1073741924,
    getuid = 1073741926,
    syslog = 1073741927,
    getgid = 1073741928,
    setuid = 1073741929,
    setgid = 1073741930,
    geteuid = 1073741931,
    getegid = 1073741932,
    setpgid = 1073741933,
    getppid = 1073741934,
    getpgrp = 1073741935,
    setsid = 1073741936,
    setreuid = 1073741937,
    setregid = 1073741938,
    getgroups = 1073741939,
    setgroups = 1073741940,
    setresuid = 1073741941,
    getresuid = 1073741942,
    setresgid = 1073741943,
    getresgid = 1073741944,
    getpgid = 1073741945,
    setfsuid = 1073741946,
    setfsgid = 1073741947,
    getsid = 1073741948,
    capget = 1073741949,
    capset = 1073741950,
    rt_sigsuspend = 1073741954,
    utime = 1073741956,
    mknod = 1073741957,
    personality = 1073741959,
    ustat = 1073741960,
    statfs = 1073741961,
    fstatfs = 1073741962,
    sysfs = 1073741963,
    getpriority = 1073741964,
    setpriority = 1073741965,
    sched_setparam = 1073741966,
    sched_getparam = 1073741967,
    sched_setscheduler = 1073741968,
    sched_getscheduler = 1073741969,
    sched_get_priority_max = 1073741970,
    sched_get_priority_min = 1073741971,
    sched_rr_get_interval = 1073741972,
    mlock = 1073741973,
    munlock = 1073741974,
    mlockall = 1073741975,
    munlockall = 1073741976,
    vhangup = 1073741977,
    modify_ldt = 1073741978,
    pivot_root = 1073741979,
    prctl = 1073741981,
    arch_prctl = 1073741982,
    adjtimex = 1073741983,
    setrlimit = 1073741984,
    chroot = 1073741985,
    sync = 1073741986,
    acct = 1073741987,
    settimeofday = 1073741988,
    mount = 1073741989,
    umount2 = 1073741990,
    swapon = 1073741991,
    swapoff = 1073741992,
    reboot = 1073741993,
    sethostname = 1073741994,
    setdomainname = 1073741995,
    iopl = 1073741996,
    ioperm = 1073741997,
    init_module = 1073741999,
    delete_module = 1073742000,
    quotactl = 1073742003,
    getpmsg = 1073742005,
    putpmsg = 1073742006,
    afs_syscall = 1073742007,
    tuxcall = 1073742008,
    security = 1073742009,
    gettid = 1073742010,
    readahead = 1073742011,
    setxattr = 1073742012,
    lsetxattr = 1073742013,
    fsetxattr = 1073742014,
    getxattr = 1073742015,
    lgetxattr = 1073742016,
    fgetxattr = 1073742017,
    listxattr = 1073742018,
    llistxattr = 1073742019,
    flistxattr = 1073742020,
    removexattr = 1073742021,
    lremovexattr = 1073742022,
    fremovexattr = 1073742023,
    tkill = 1073742024,
    time = 1073742025,
    futex = 1073742026,
    sched_setaffinity = 1073742027,
    sched_getaffinity = 1073742028,
    io_destroy = 1073742031,
    io_getevents = 1073742032,
    io_cancel = 1073742034,
    lookup_dcookie = 1073742036,
    epoll_create = 1073742037,
    remap_file_pages = 1073742040,
    getdents64 = 1073742041,
    set_tid_address = 1073742042,
    restart_syscall = 1073742043,
    semtimedop = 1073742044,
    fadvise64 = 1073742045,
    timer_settime = 1073742047,
    timer_gettime = 1073742048,
    timer_getoverrun = 1073742049,
    timer_delete = 1073742050,
    clock_settime = 1073742051,
    clock_gettime = 1073742052,
    clock_getres = 1073742053,
    clock_nanosleep = 1073742054,
    exit_group = 1073742055,
    epoll_wait = 1073742056,
    epoll_ctl = 1073742057,
    tgkill = 1073742058,
    utimes = 1073742059,
    mbind = 1073742061,
    set_mempolicy = 1073742062,
    get_mempolicy = 1073742063,
    mq_open = 1073742064,
    mq_unlink = 1073742065,
    mq_timedsend = 1073742066,
    mq_timedreceive = 1073742067,
    mq_getsetattr = 1073742069,
    add_key = 1073742072,
    request_key = 1073742073,
    keyctl = 1073742074,
    ioprio_set = 1073742075,
    ioprio_get = 1073742076,
    inotify_init = 1073742077,
    inotify_add_watch = 1073742078,
    inotify_rm_watch = 1073742079,
    migrate_pages = 1073742080,
    openat = 1073742081,
    mkdirat = 1073742082,
    mknodat = 1073742083,
    fchownat = 1073742084,
    futimesat = 1073742085,
    newfstatat = 1073742086,
    unlinkat = 1073742087,
    renameat = 1073742088,
    linkat = 1073742089,
    symlinkat = 1073742090,
    readlinkat = 1073742091,
    fchmodat = 1073742092,
    faccessat = 1073742093,
    pselect6 = 1073742094,
    ppoll = 1073742095,
    unshare = 1073742096,
    splice = 1073742099,
    tee = 1073742100,
    sync_file_range = 1073742101,
    utimensat = 1073742104,
    epoll_pwait = 1073742105,
    signalfd = 1073742106,
    timerfd_create = 1073742107,
    eventfd = 1073742108,
    fallocate = 1073742109,
    timerfd_settime = 1073742110,
    timerfd_gettime = 1073742111,
    accept4 = 1073742112,
    signalfd4 = 1073742113,
    eventfd2 = 1073742114,
    epoll_create1 = 1073742115,
    dup3 = 1073742116,
    pipe2 = 1073742117,
    inotify_init1 = 1073742118,
    perf_event_open = 1073742122,
    fanotify_init = 1073742124,
    fanotify_mark = 1073742125,
    prlimit64 = 1073742126,
    name_to_handle_at = 1073742127,
    open_by_handle_at = 1073742128,
    clock_adjtime = 1073742129,
    syncfs = 1073742130,
    setns = 1073742132,
    getcpu = 1073742133,
    kcmp = 1073742136,
    finit_module = 1073742137,
    sched_setattr = 1073742138,
    sched_getattr = 1073742139,
    renameat2 = 1073742140,
    seccomp = 1073742141,
    getrandom = 1073742142,
    memfd_create = 1073742143,
    kexec_file_load = 1073742144,
    bpf = 1073742145,
    userfaultfd = 1073742147,
    membarrier = 1073742148,
    mlock2 = 1073742149,
    copy_file_range = 1073742150,
    pkey_mprotect = 1073742153,
    pkey_alloc = 1073742154,
    pkey_free = 1073742155,
    statx = 1073742156,
    io_pgetevents = 1073742157,
    rseq = 1073742158,
    uretprobe = 1073742159,
    uprobe = 1073742160,
    pidfd_send_signal = 1073742248,
    io_uring_setup = 1073742249,
    io_uring_enter = 1073742250,
    io_uring_register = 1073742251,
    open_tree = 1073742252,
    move_mount = 1073742253,
    fsopen = 1073742254,
    fsconfig = 1073742255,
    fsmount = 1073742256,
    fspick = 1073742257,
    pidfd_open = 1073742258,
    clone3 = 1073742259,
    close_range = 1073742260,
    openat2 = 1073742261,
    pidfd_getfd = 1073742262,
    faccessat2 = 1073742263,
    process_madvise = 1073742264,
    epoll_pwait2 = 1073742265,
    mount_setattr = 1073742266,
    quotactl_fd = 1073742267,
    landlock_create_ruleset = 1073742268,
    landlock_add_rule = 1073742269,
    landlock_restrict_self = 1073742270,
    memfd_secret = 1073742271,
    process_mrelease = 1073742272,
    futex_waitv = 1073742273,
    set_mempolicy_home_node = 1073742274,
    cachestat = 1073742275,
    fchmodat2 = 1073742276,
    map_shadow_stack = 1073742277,
    futex_wake = 1073742278,
    futex_wait = 1073742279,
    futex_requeue = 1073742280,
    statmount = 1073742281,
    listmount = 1073742282,
    lsm_get_self_attr = 1073742283,
    lsm_set_self_attr = 1073742284,
    lsm_list_modules = 1073742285,
    mseal = 1073742286,
    setxattrat = 1073742287,
    getxattrat = 1073742288,
    listxattrat = 1073742289,
    removexattrat = 1073742290,
    open_tree_attr = 1073742291,
    file_getattr = 1073742292,
    file_setattr = 1073742293,
    listns = 1073742294,
    rseq_slice_yield = 1073742295,
    rt_sigaction = 1073742336,
    rt_sigreturn = 1073742337,
    ioctl = 1073742338,
    readv = 1073742339,
    writev = 1073742340,
    recvfrom = 1073742341,
    sendmsg = 1073742342,
    recvmsg = 1073742343,
    execve = 1073742344,
    ptrace = 1073742345,
    rt_sigpending = 1073742346,
    rt_sigtimedwait = 1073742347,
    rt_sigqueueinfo = 1073742348,
    sigaltstack = 1073742349,
    timer_create = 1073742350,
    mq_notify = 1073742351,
    kexec_load = 1073742352,
    waitid = 1073742353,
    set_robust_list = 1073742354,
    get_robust_list = 1073742355,
    vmsplice = 1073742356,
    move_pages = 1073742357,
    preadv = 1073742358,
    pwritev = 1073742359,
    rt_tgsigqueueinfo = 1073742360,
    recvmmsg = 1073742361,
    sendmmsg = 1073742362,
    process_vm_readv = 1073742363,
    process_vm_writev = 1073742364,
    setsockopt = 1073742365,
    getsockopt = 1073742366,
    io_setup = 1073742367,
    io_submit = 1073742368,
    execveat = 1073742369,
    preadv2 = 1073742370,
    pwritev2 = 1073742371,
};

pub const Arm = enum(usize) {
    const arm_base = 0x0f0000;

    restart_syscall = 0,
    exit = 1,
    fork = 2,
    read = 3,
    write = 4,
    open = 5,
    close = 6,
    creat = 8,
    link = 9,
    unlink = 10,
    execve = 11,
    chdir = 12,
    mknod = 14,
    chmod = 15,
    lchown = 16,
    lseek = 19,
    getpid = 20,
    mount = 21,
    setuid = 23,
    getuid = 24,
    ptrace = 26,
    pause = 29,
    access = 33,
    nice = 34,
    sync = 36,
    kill = 37,
    rename = 38,
    mkdir = 39,
    rmdir = 40,
    dup = 41,
    pipe = 42,
    times = 43,
    brk = 45,
    setgid = 46,
    getgid = 47,
    geteuid = 49,
    getegid = 50,
    acct = 51,
    umount2 = 52,
    ioctl = 54,
    fcntl = 55,
    setpgid = 57,
    umask = 60,
    chroot = 61,
    ustat = 62,
    dup2 = 63,
    getppid = 64,
    getpgrp = 65,
    setsid = 66,
    sigaction = 67,
    setreuid = 70,
    setregid = 71,
    sigsuspend = 72,
    sigpending = 73,
    sethostname = 74,
    setrlimit = 75,
    getrusage = 77,
    gettimeofday = 78,
    settimeofday = 79,
    getgroups = 80,
    setgroups = 81,
    symlink = 83,
    readlink = 85,
    uselib = 86,
    swapon = 87,
    reboot = 88,
    munmap = 91,
    truncate = 92,
    ftruncate = 93,
    fchmod = 94,
    fchown = 95,
    getpriority = 96,
    setpriority = 97,
    statfs = 99,
    fstatfs = 100,
    syslog = 103,
    setitimer = 104,
    getitimer = 105,
    stat = 106,
    lstat = 107,
    fstat = 108,
    vhangup = 111,
    wait4 = 114,
    swapoff = 115,
    sysinfo = 116,
    fsync = 118,
    sigreturn = 119,
    clone = 120,
    setdomainname = 121,
    uname = 122,
    adjtimex = 124,
    mprotect = 125,
    sigprocmask = 126,
    init_module = 128,
    delete_module = 129,
    quotactl = 131,
    getpgid = 132,
    fchdir = 133,
    bdflush = 134,
    sysfs = 135,
    personality = 136,
    setfsuid = 138,
    setfsgid = 139,
    llseek = 140,
    getdents = 141,
    newselect = 142,
    flock = 143,
    msync = 144,
    readv = 145,
    writev = 146,
    getsid = 147,
    fdatasync = 148,
    sysctl = 149,
    mlock = 150,
    munlock = 151,
    mlockall = 152,
    munlockall = 153,
    sched_setparam = 154,
    sched_getparam = 155,
    sched_setscheduler = 156,
    sched_getscheduler = 157,
    sched_yield = 158,
    sched_get_priority_max = 159,
    sched_get_priority_min = 160,
    sched_rr_get_interval = 161,
    nanosleep = 162,
    mremap = 163,
    setresuid = 164,
    getresuid = 165,
    poll = 168,
    nfsservctl = 169,
    setresgid = 170,
    getresgid = 171,
    prctl = 172,
    rt_sigreturn = 173,
    rt_sigaction = 174,
    rt_sigprocmask = 175,
    rt_sigpending = 176,
    rt_sigtimedwait = 177,
    rt_sigqueueinfo = 178,
    rt_sigsuspend = 179,
    pread64 = 180,
    pwrite64 = 181,
    chown = 182,
    getcwd = 183,
    capget = 184,
    capset = 185,
    sigaltstack = 186,
    sendfile = 187,
    vfork = 190,
    ugetrlimit = 191,
    mmap2 = 192,
    truncate64 = 193,
    ftruncate64 = 194,
    stat64 = 195,
    lstat64 = 196,
    fstat64 = 197,
    lchown32 = 198,
    getuid32 = 199,
    getgid32 = 200,
    geteuid32 = 201,
    getegid32 = 202,
    setreuid32 = 203,
    setregid32 = 204,
    getgroups32 = 205,
    setgroups32 = 206,
    fchown32 = 207,
    setresuid32 = 208,
    getresuid32 = 209,
    setresgid32 = 210,
    getresgid32 = 211,
    chown32 = 212,
    setuid32 = 213,
    setgid32 = 214,
    setfsuid32 = 215,
    setfsgid32 = 216,
    getdents64 = 217,
    pivot_root = 218,
    mincore = 219,
    madvise = 220,
    fcntl64 = 221,
    gettid = 224,
    readahead = 225,
    setxattr = 226,
    lsetxattr = 227,
    fsetxattr = 228,
    getxattr = 229,
    lgetxattr = 230,
    fgetxattr = 231,
    listxattr = 232,
    llistxattr = 233,
    flistxattr = 234,
    removexattr = 235,
    lremovexattr = 236,
    fremovexattr = 237,
    tkill = 238,
    sendfile64 = 239,
    futex = 240,
    sched_setaffinity = 241,
    sched_getaffinity = 242,
    io_setup = 243,
    io_destroy = 244,
    io_getevents = 245,
    io_submit = 246,
    io_cancel = 247,
    exit_group = 248,
    lookup_dcookie = 249,
    epoll_create = 250,
    epoll_ctl = 251,
    epoll_wait = 252,
    remap_file_pages = 253,
    set_tid_address = 256,
    timer_create = 257,
    timer_settime = 258,
    timer_gettime = 259,
    timer_getoverrun = 260,
    timer_delete = 261,
    clock_settime = 262,
    clock_gettime = 263,
    clock_getres = 264,
    clock_nanosleep = 265,
    statfs64 = 266,
    fstatfs64 = 267,
    tgkill = 268,
    utimes = 269,
    fadvise64_64 = 270,
    pciconfig_iobase = 271,
    pciconfig_read = 272,
    pciconfig_write = 273,
    mq_open = 274,
    mq_unlink = 275,
    mq_timedsend = 276,
    mq_timedreceive = 277,
    mq_notify = 278,
    mq_getsetattr = 279,
    waitid = 280,
    socket = 281,
    bind = 282,
    connect = 283,
    listen = 284,
    accept = 285,
    getsockname = 286,
    getpeername = 287,
    socketpair = 288,
    send = 289,
    sendto = 290,
    recv = 291,
    recvfrom = 292,
    shutdown = 293,
    setsockopt = 294,
    getsockopt = 295,
    sendmsg = 296,
    recvmsg = 297,
    semop = 298,
    semget = 299,
    semctl = 300,
    msgsnd = 301,
    msgrcv = 302,
    msgget = 303,
    msgctl = 304,
    shmat = 305,
    shmdt = 306,
    shmget = 307,
    shmctl = 308,
    add_key = 309,
    request_key = 310,
    keyctl = 311,
    semtimedop = 312,
    vserver = 313,
    ioprio_set = 314,
    ioprio_get = 315,
    inotify_init = 316,
    inotify_add_watch = 317,
    inotify_rm_watch = 318,
    mbind = 319,
    get_mempolicy = 320,
    set_mempolicy = 321,
    openat = 322,
    mkdirat = 323,
    mknodat = 324,
    fchownat = 325,
    futimesat = 326,
    fstatat64 = 327,
    unlinkat = 328,
    renameat = 329,
    linkat = 330,
    symlinkat = 331,
    readlinkat = 332,
    fchmodat = 333,
    faccessat = 334,
    pselect6 = 335,
    ppoll = 336,
    unshare = 337,
    set_robust_list = 338,
    get_robust_list = 339,
    splice = 340,
    sync_file_range = 341,
    tee = 342,
    vmsplice = 343,
    move_pages = 344,
    getcpu = 345,
    epoll_pwait = 346,
    kexec_load = 347,
    utimensat = 348,
    signalfd = 349,
    timerfd_create = 350,
    eventfd = 351,
    fallocate = 352,
    timerfd_settime = 353,
    timerfd_gettime = 354,
    signalfd4 = 355,
    eventfd2 = 356,
    epoll_create1 = 357,
    dup3 = 358,
    pipe2 = 359,
    inotify_init1 = 360,
    preadv = 361,
    pwritev = 362,
    rt_tgsigqueueinfo = 363,
    perf_event_open = 364,
    recvmmsg = 365,
    accept4 = 366,
    fanotify_init = 367,
    fanotify_mark = 368,
    prlimit64 = 369,
    name_to_handle_at = 370,
    open_by_handle_at = 371,
    clock_adjtime = 372,
    syncfs = 373,
    sendmmsg = 374,
    setns = 375,
    process_vm_readv = 376,
    process_vm_writev = 377,
    kcmp = 378,
    finit_module = 379,
    sched_setattr = 380,
    sched_getattr = 381,
    renameat2 = 382,
    seccomp = 383,
    getrandom = 384,
    memfd_create = 385,
    bpf = 386,
    execveat = 387,
    userfaultfd = 388,
    membarrier = 389,
    mlock2 = 390,
    copy_file_range = 391,
    preadv2 = 392,
    pwritev2 = 393,
    pkey_mprotect = 394,
    pkey_alloc = 395,
    pkey_free = 396,
    statx = 397,
    rseq = 398,
    io_pgetevents = 399,
    migrate_pages = 400,
    kexec_file_load = 401,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,

    breakpoint = arm_base + 1,
    cacheflush = arm_base + 2,
    usr26 = arm_base + 3,
    usr32 = arm_base + 4,
    set_tls = arm_base + 5,
    get_tls = arm_base + 6,
};

pub const Sparc = enum(usize) {
    restart_syscall = 0,
    exit = 1,
    fork = 2,
    read = 3,
    write = 4,
    open = 5,
    close = 6,
    wait4 = 7,
    creat = 8,
    link = 9,
    unlink = 10,
    execv = 11,
    chdir = 12,
    chown = 13,
    mknod = 14,
    chmod = 15,
    lchown = 16,
    brk = 17,
    perfctr = 18,
    lseek = 19,
    getpid = 20,
    capget = 21,
    capset = 22,
    setuid = 23,
    getuid = 24,
    vmsplice = 25,
    ptrace = 26,
    alarm = 27,
    sigaltstack = 28,
    pause = 29,
    utime = 30,
    lchown32 = 31,
    fchown32 = 32,
    access = 33,
    nice = 34,
    chown32 = 35,
    sync = 36,
    kill = 37,
    stat = 38,
    sendfile = 39,
    lstat = 40,
    dup = 41,
    pipe = 42,
    times = 43,
    getuid32 = 44,
    umount2 = 45,
    setgid = 46,
    getgid = 47,
    signal = 48,
    geteuid = 49,
    getegid = 50,
    acct = 51,
    getgid32 = 53,
    ioctl = 54,
    reboot = 55,
    mmap2 = 56,
    symlink = 57,
    readlink = 58,
    execve = 59,
    umask = 60,
    chroot = 61,
    fstat = 62,
    fstat64 = 63,
    getpagesize = 64,
    msync = 65,
    vfork = 66,
    pread64 = 67,
    pwrite64 = 68,
    geteuid32 = 69,
    getegid32 = 70,
    mmap = 71,
    setreuid32 = 72,
    munmap = 73,
    mprotect = 74,
    madvise = 75,
    vhangup = 76,
    truncate64 = 77,
    mincore = 78,
    getgroups = 79,
    setgroups = 80,
    getpgrp = 81,
    setgroups32 = 82,
    setitimer = 83,
    ftruncate64 = 84,
    swapon = 85,
    getitimer = 86,
    setuid32 = 87,
    sethostname = 88,
    setgid32 = 89,
    dup2 = 90,
    setfsuid32 = 91,
    fcntl = 92,
    select = 93,
    setfsgid32 = 94,
    fsync = 95,
    setpriority = 96,
    socket = 97,
    connect = 98,
    accept = 99,
    getpriority = 100,
    rt_sigreturn = 101,
    rt_sigaction = 102,
    rt_sigprocmask = 103,
    rt_sigpending = 104,
    rt_sigtimedwait = 105,
    rt_sigqueueinfo = 106,
    rt_sigsuspend = 107,
    setresuid32 = 108,
    getresuid32 = 109,
    setresgid32 = 110,
    getresgid32 = 111,
    setregid32 = 112,
    recvmsg = 113,
    sendmsg = 114,
    getgroups32 = 115,
    gettimeofday = 116,
    getrusage = 117,
    getsockopt = 118,
    getcwd = 119,
    readv = 120,
    writev = 121,
    settimeofday = 122,
    fchown = 123,
    fchmod = 124,
    recvfrom = 125,
    setreuid = 126,
    setregid = 127,
    rename = 128,
    truncate = 129,
    ftruncate = 130,
    flock = 131,
    lstat64 = 132,
    sendto = 133,
    shutdown = 134,
    socketpair = 135,
    mkdir = 136,
    rmdir = 137,
    utimes = 138,
    stat64 = 139,
    sendfile64 = 140,
    getpeername = 141,
    futex = 142,
    gettid = 143,
    getrlimit = 144,
    setrlimit = 145,
    pivot_root = 146,
    prctl = 147,
    pciconfig_read = 148,
    pciconfig_write = 149,
    getsockname = 150,
    inotify_init = 151,
    inotify_add_watch = 152,
    poll = 153,
    getdents64 = 154,
    fcntl64 = 155,
    inotify_rm_watch = 156,
    statfs = 157,
    fstatfs = 158,
    umount = 159,
    sched_set_affinity = 160,
    sched_get_affinity = 161,
    getdomainname = 162,
    setdomainname = 163,
    quotactl = 165,
    set_tid_address = 166,
    mount = 167,
    ustat = 168,
    setxattr = 169,
    lsetxattr = 170,
    fsetxattr = 171,
    getxattr = 172,
    lgetxattr = 173,
    getdents = 174,
    setsid = 175,
    fchdir = 176,
    fgetxattr = 177,
    listxattr = 178,
    llistxattr = 179,
    flistxattr = 180,
    removexattr = 181,
    lremovexattr = 182,
    sigpending = 183,
    query_module = 184,
    setpgid = 185,
    fremovexattr = 186,
    tkill = 187,
    exit_group = 188,
    uname = 189,
    init_module = 190,
    personality = 191,
    remap_file_pages = 192,
    epoll_create = 193,
    epoll_ctl = 194,
    epoll_wait = 195,
    ioprio_set = 196,
    getppid = 197,
    sigaction = 198,
    sgetmask = 199,
    ssetmask = 200,
    sigsuspend = 201,
    oldlstat = 202,
    uselib = 203,
    readdir = 204,
    readahead = 205,
    socketcall = 206,
    syslog = 207,
    lookup_dcookie = 208,
    fadvise64 = 209,
    fadvise64_64 = 210,
    tgkill = 211,
    waitpid = 212,
    swapoff = 213,
    sysinfo = 214,
    ipc = 215,
    sigreturn = 216,
    clone = 217,
    ioprio_get = 218,
    adjtimex = 219,
    sigprocmask = 220,
    create_module = 221,
    delete_module = 222,
    get_kernel_syms = 223,
    getpgid = 224,
    bdflush = 225,
    sysfs = 226,
    afs_syscall = 227,
    setfsuid = 228,
    setfsgid = 229,
    newselect = 230,
    time = 231,
    splice = 232,
    stime = 233,
    statfs64 = 234,
    fstatfs64 = 235,
    llseek = 236,
    mlock = 237,
    munlock = 238,
    mlockall = 239,
    munlockall = 240,
    sched_setparam = 241,
    sched_getparam = 242,
    sched_setscheduler = 243,
    sched_getscheduler = 244,
    sched_yield = 245,
    sched_get_priority_max = 246,
    sched_get_priority_min = 247,
    sched_rr_get_interval = 248,
    nanosleep = 249,
    mremap = 250,
    sysctl = 251,
    getsid = 252,
    fdatasync = 253,
    nfsservctl = 254,
    sync_file_range = 255,
    clock_settime = 256,
    clock_gettime = 257,
    clock_getres = 258,
    clock_nanosleep = 259,
    sched_getaffinity = 260,
    sched_setaffinity = 261,
    timer_settime = 262,
    timer_gettime = 263,
    timer_getoverrun = 264,
    timer_delete = 265,
    timer_create = 266,
    vserver = 267,
    io_setup = 268,
    io_destroy = 269,
    io_submit = 270,
    io_cancel = 271,
    io_getevents = 272,
    mq_open = 273,
    mq_unlink = 274,
    mq_timedsend = 275,
    mq_timedreceive = 276,
    mq_notify = 277,
    mq_getsetattr = 278,
    waitid = 279,
    tee = 280,
    add_key = 281,
    request_key = 282,
    keyctl = 283,
    openat = 284,
    mkdirat = 285,
    mknodat = 286,
    fchownat = 287,
    futimesat = 288,
    fstatat64 = 289,
    unlinkat = 290,
    renameat = 291,
    linkat = 292,
    symlinkat = 293,
    readlinkat = 294,
    fchmodat = 295,
    faccessat = 296,
    pselect6 = 297,
    ppoll = 298,
    unshare = 299,
    set_robust_list = 300,
    get_robust_list = 301,
    migrate_pages = 302,
    mbind = 303,
    get_mempolicy = 304,
    set_mempolicy = 305,
    kexec_load = 306,
    move_pages = 307,
    getcpu = 308,
    epoll_pwait = 309,
    utimensat = 310,
    signalfd = 311,
    timerfd_create = 312,
    eventfd = 313,
    fallocate = 314,
    timerfd_settime = 315,
    timerfd_gettime = 316,
    signalfd4 = 317,
    eventfd2 = 318,
    epoll_create1 = 319,
    dup3 = 320,
    pipe2 = 321,
    inotify_init1 = 322,
    accept4 = 323,
    preadv = 324,
    pwritev = 325,
    rt_tgsigqueueinfo = 326,
    perf_event_open = 327,
    recvmmsg = 328,
    fanotify_init = 329,
    fanotify_mark = 330,
    prlimit64 = 331,
    name_to_handle_at = 332,
    open_by_handle_at = 333,
    clock_adjtime = 334,
    syncfs = 335,
    sendmmsg = 336,
    setns = 337,
    process_vm_readv = 338,
    process_vm_writev = 339,
    kern_features = 340,
    kcmp = 341,
    finit_module = 342,
    sched_setattr = 343,
    sched_getattr = 344,
    renameat2 = 345,
    seccomp = 346,
    getrandom = 347,
    memfd_create = 348,
    bpf = 349,
    execveat = 350,
    membarrier = 351,
    userfaultfd = 352,
    bind = 353,
    listen = 354,
    setsockopt = 355,
    mlock2 = 356,
    copy_file_range = 357,
    preadv2 = 358,
    pwritev2 = 359,
    statx = 360,
    io_pgetevents = 361,
    pkey_mprotect = 362,
    pkey_alloc = 363,
    pkey_free = 364,
    rseq = 365,
    semget = 393,
    semctl = 394,
    shmget = 395,
    shmctl = 396,
    shmat = 397,
    shmdt = 398,
    msgget = 399,
    msgsnd = 400,
    msgrcv = 401,
    msgctl = 402,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const Sparc64 = enum(usize) {
    restart_syscall = 0,
    exit = 1,
    fork = 2,
    read = 3,
    write = 4,
    open = 5,
    close = 6,
    wait4 = 7,
    creat = 8,
    link = 9,
    unlink = 10,
    execv = 11,
    chdir = 12,
    chown = 13,
    mknod = 14,
    chmod = 15,
    lchown = 16,
    brk = 17,
    perfctr = 18,
    lseek = 19,
    getpid = 20,
    capget = 21,
    capset = 22,
    setuid = 23,
    getuid = 24,
    vmsplice = 25,
    ptrace = 26,
    alarm = 27,
    sigaltstack = 28,
    pause = 29,
    utime = 30,
    access = 33,
    nice = 34,
    sync = 36,
    kill = 37,
    stat = 38,
    sendfile = 39,
    lstat = 40,
    dup = 41,
    pipe = 42,
    times = 43,
    umount2 = 45,
    setgid = 46,
    getgid = 47,
    signal = 48,
    geteuid = 49,
    getegid = 50,
    acct = 51,
    memory_ordering = 52,
    ioctl = 54,
    reboot = 55,
    symlink = 57,
    readlink = 58,
    execve = 59,
    umask = 60,
    chroot = 61,
    fstat = 62,
    fstat64 = 63,
    getpagesize = 64,
    msync = 65,
    vfork = 66,
    pread64 = 67,
    pwrite64 = 68,
    mmap = 71,
    munmap = 73,
    mprotect = 74,
    madvise = 75,
    vhangup = 76,
    mincore = 78,
    getgroups = 79,
    setgroups = 80,
    getpgrp = 81,
    setitimer = 83,
    swapon = 85,
    getitimer = 86,
    sethostname = 88,
    dup2 = 90,
    fcntl = 92,
    select = 93,
    fsync = 95,
    setpriority = 96,
    socket = 97,
    connect = 98,
    accept = 99,
    getpriority = 100,
    rt_sigreturn = 101,
    rt_sigaction = 102,
    rt_sigprocmask = 103,
    rt_sigpending = 104,
    rt_sigtimedwait = 105,
    rt_sigqueueinfo = 106,
    rt_sigsuspend = 107,
    setresuid = 108,
    getresuid = 109,
    setresgid = 110,
    getresgid = 111,
    recvmsg = 113,
    sendmsg = 114,
    gettimeofday = 116,
    getrusage = 117,
    getsockopt = 118,
    getcwd = 119,
    readv = 120,
    writev = 121,
    settimeofday = 122,
    fchown = 123,
    fchmod = 124,
    recvfrom = 125,
    setreuid = 126,
    setregid = 127,
    rename = 128,
    truncate = 129,
    ftruncate = 130,
    flock = 131,
    lstat64 = 132,
    sendto = 133,
    shutdown = 134,
    socketpair = 135,
    mkdir = 136,
    rmdir = 137,
    utimes = 138,
    stat64 = 139,
    sendfile64 = 140,
    getpeername = 141,
    futex = 142,
    gettid = 143,
    getrlimit = 144,
    setrlimit = 145,
    pivot_root = 146,
    prctl = 147,
    pciconfig_read = 148,
    pciconfig_write = 149,
    getsockname = 150,
    inotify_init = 151,
    inotify_add_watch = 152,
    poll = 153,
    getdents64 = 154,
    inotify_rm_watch = 156,
    statfs = 157,
    fstatfs = 158,
    umount = 159,
    sched_set_affinity = 160,
    sched_get_affinity = 161,
    getdomainname = 162,
    setdomainname = 163,
    utrap_install = 164,
    quotactl = 165,
    set_tid_address = 166,
    mount = 167,
    ustat = 168,
    setxattr = 169,
    lsetxattr = 170,
    fsetxattr = 171,
    getxattr = 172,
    lgetxattr = 173,
    getdents = 174,
    setsid = 175,
    fchdir = 176,
    fgetxattr = 177,
    listxattr = 178,
    llistxattr = 179,
    flistxattr = 180,
    removexattr = 181,
    lremovexattr = 182,
    sigpending = 183,
    query_module = 184,
    setpgid = 185,
    fremovexattr = 186,
    tkill = 187,
    exit_group = 188,
    uname = 189,
    init_module = 190,
    personality = 191,
    remap_file_pages = 192,
    epoll_create = 193,
    epoll_ctl = 194,
    epoll_wait = 195,
    ioprio_set = 196,
    getppid = 197,
    sigaction = 198,
    sgetmask = 199,
    ssetmask = 200,
    sigsuspend = 201,
    oldlstat = 202,
    uselib = 203,
    readdir = 204,
    readahead = 205,
    socketcall = 206,
    syslog = 207,
    lookup_dcookie = 208,
    fadvise64 = 209,
    fadvise64_64 = 210,
    tgkill = 211,
    waitpid = 212,
    swapoff = 213,
    sysinfo = 214,
    ipc = 215,
    sigreturn = 216,
    clone = 217,
    ioprio_get = 218,
    adjtimex = 219,
    sigprocmask = 220,
    create_module = 221,
    delete_module = 222,
    get_kernel_syms = 223,
    getpgid = 224,
    bdflush = 225,
    sysfs = 226,
    afs_syscall = 227,
    setfsuid = 228,
    setfsgid = 229,
    newselect = 230,
    splice = 232,
    stime = 233,
    statfs64 = 234,
    fstatfs64 = 235,
    llseek = 236,
    mlock = 237,
    munlock = 238,
    mlockall = 239,
    munlockall = 240,
    sched_setparam = 241,
    sched_getparam = 242,
    sched_setscheduler = 243,
    sched_getscheduler = 244,
    sched_yield = 245,
    sched_get_priority_max = 246,
    sched_get_priority_min = 247,
    sched_rr_get_interval = 248,
    nanosleep = 249,
    mremap = 250,
    sysctl = 251,
    getsid = 252,
    fdatasync = 253,
    nfsservctl = 254,
    sync_file_range = 255,
    clock_settime = 256,
    clock_gettime = 257,
    clock_getres = 258,
    clock_nanosleep = 259,
    sched_getaffinity = 260,
    sched_setaffinity = 261,
    timer_settime = 262,
    timer_gettime = 263,
    timer_getoverrun = 264,
    timer_delete = 265,
    timer_create = 266,
    vserver = 267,
    io_setup = 268,
    io_destroy = 269,
    io_submit = 270,
    io_cancel = 271,
    io_getevents = 272,
    mq_open = 273,
    mq_unlink = 274,
    mq_timedsend = 275,
    mq_timedreceive = 276,
    mq_notify = 277,
    mq_getsetattr = 278,
    waitid = 279,
    tee = 280,
    add_key = 281,
    request_key = 282,
    keyctl = 283,
    openat = 284,
    mkdirat = 285,
    mknodat = 286,
    fchownat = 287,
    futimesat = 288,
    fstatat64 = 289,
    unlinkat = 290,
    renameat = 291,
    linkat = 292,
    symlinkat = 293,
    readlinkat = 294,
    fchmodat = 295,
    faccessat = 296,
    pselect6 = 297,
    ppoll = 298,
    unshare = 299,
    set_robust_list = 300,
    get_robust_list = 301,
    migrate_pages = 302,
    mbind = 303,
    get_mempolicy = 304,
    set_mempolicy = 305,
    kexec_load = 306,
    move_pages = 307,
    getcpu = 308,
    epoll_pwait = 309,
    utimensat = 310,
    signalfd = 311,
    timerfd_create = 312,
    eventfd = 313,
    fallocate = 314,
    timerfd_settime = 315,
    timerfd_gettime = 316,
    signalfd4 = 317,
    eventfd2 = 318,
    epoll_create1 = 319,
    dup3 = 320,
    pipe2 = 321,
    inotify_init1 = 322,
    accept4 = 323,
    preadv = 324,
    pwritev = 325,
    rt_tgsigqueueinfo = 326,
    perf_event_open = 327,
    recvmmsg = 328,
    fanotify_init = 329,
    fanotify_mark = 330,
    prlimit64 = 331,
    name_to_handle_at = 332,
    open_by_handle_at = 333,
    clock_adjtime = 334,
    syncfs = 335,
    sendmmsg = 336,
    setns = 337,
    process_vm_readv = 338,
    process_vm_writev = 339,
    kern_features = 340,
    kcmp = 341,
    finit_module = 342,
    sched_setattr = 343,
    sched_getattr = 344,
    renameat2 = 345,
    seccomp = 346,
    getrandom = 347,
    memfd_create = 348,
    bpf = 349,
    execveat = 350,
    membarrier = 351,
    userfaultfd = 352,
    bind = 353,
    listen = 354,
    setsockopt = 355,
    mlock2 = 356,
    copy_file_range = 357,
    preadv2 = 358,
    pwritev2 = 359,
    statx = 360,
    io_pgetevents = 361,
    pkey_mprotect = 362,
    pkey_alloc = 363,
    pkey_free = 364,
    rseq = 365,
    semtimedop = 392,
    semget = 393,
    semctl = 394,
    shmget = 395,
    shmctl = 396,
    shmat = 397,
    shmdt = 398,
    msgget = 399,
    msgsnd = 400,
    msgrcv = 401,
    msgctl = 402,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const M68k = enum(usize) {
    restart_syscall = 0,
    exit = 1,
    fork = 2,
    read = 3,
    write = 4,
    open = 5,
    close = 6,
    waitpid = 7,
    creat = 8,
    link = 9,
    unlink = 10,
    execve = 11,
    chdir = 12,
    time = 13,
    mknod = 14,
    chmod = 15,
    chown = 16,
    oldstat = 18,
    lseek = 19,
    getpid = 20,
    mount = 21,
    umount = 22,
    setuid = 23,
    getuid = 24,
    stime = 25,
    ptrace = 26,
    alarm = 27,
    oldfstat = 28,
    pause = 29,
    utime = 30,
    access = 33,
    nice = 34,
    sync = 36,
    kill = 37,
    rename = 38,
    mkdir = 39,
    rmdir = 40,
    dup = 41,
    pipe = 42,
    times = 43,
    brk = 45,
    setgid = 46,
    getgid = 47,
    signal = 48,
    geteuid = 49,
    getegid = 50,
    acct = 51,
    umount2 = 52,
    ioctl = 54,
    fcntl = 55,
    setpgid = 57,
    umask = 60,
    chroot = 61,
    ustat = 62,
    dup2 = 63,
    getppid = 64,
    getpgrp = 65,
    setsid = 66,
    sigaction = 67,
    sgetmask = 68,
    ssetmask = 69,
    setreuid = 70,
    setregid = 71,
    sigsuspend = 72,
    sigpending = 73,
    sethostname = 74,
    setrlimit = 75,
    getrlimit = 76,
    getrusage = 77,
    gettimeofday = 78,
    settimeofday = 79,
    getgroups = 80,
    setgroups = 81,
    select = 82,
    symlink = 83,
    oldlstat = 84,
    readlink = 85,
    uselib = 86,
    swapon = 87,
    reboot = 88,
    readdir = 89,
    mmap = 90,
    munmap = 91,
    truncate = 92,
    ftruncate = 93,
    fchmod = 94,
    fchown = 95,
    getpriority = 96,
    setpriority = 97,
    statfs = 99,
    fstatfs = 100,
    socketcall = 102,
    syslog = 103,
    setitimer = 104,
    getitimer = 105,
    stat = 106,
    lstat = 107,
    fstat = 108,
    vhangup = 111,
    wait4 = 114,
    swapoff = 115,
    sysinfo = 116,
    ipc = 117,
    fsync = 118,
    sigreturn = 119,
    clone = 120,
    setdomainname = 121,
    uname = 122,
    cacheflush = 123,
    adjtimex = 124,
    mprotect = 125,
    sigprocmask = 126,
    create_module = 127,
    init_module = 128,
    delete_module = 129,
    get_kernel_syms = 130,
    quotactl = 131,
    getpgid = 132,
    fchdir = 133,
    bdflush = 134,
    sysfs = 135,
    personality = 136,
    setfsuid = 138,
    setfsgid = 139,
    llseek = 140,
    getdents = 141,
    newselect = 142,
    flock = 143,
    msync = 144,
    readv = 145,
    writev = 146,
    getsid = 147,
    fdatasync = 148,
    sysctl = 149,
    mlock = 150,
    munlock = 151,
    mlockall = 152,
    munlockall = 153,
    sched_setparam = 154,
    sched_getparam = 155,
    sched_setscheduler = 156,
    sched_getscheduler = 157,
    sched_yield = 158,
    sched_get_priority_max = 159,
    sched_get_priority_min = 160,
    sched_rr_get_interval = 161,
    nanosleep = 162,
    mremap = 163,
    setresuid = 164,
    getresuid = 165,
    getpagesize = 166,
    query_module = 167,
    poll = 168,
    nfsservctl = 169,
    setresgid = 170,
    getresgid = 171,
    prctl = 172,
    rt_sigreturn = 173,
    rt_sigaction = 174,
    rt_sigprocmask = 175,
    rt_sigpending = 176,
    rt_sigtimedwait = 177,
    rt_sigqueueinfo = 178,
    rt_sigsuspend = 179,
    pread64 = 180,
    pwrite64 = 181,
    lchown = 182,
    getcwd = 183,
    capget = 184,
    capset = 185,
    sigaltstack = 186,
    sendfile = 187,
    getpmsg = 188,
    putpmsg = 189,
    vfork = 190,
    ugetrlimit = 191,
    mmap2 = 192,
    truncate64 = 193,
    ftruncate64 = 194,
    stat64 = 195,
    lstat64 = 196,
    fstat64 = 197,
    chown32 = 198,
    getuid32 = 199,
    getgid32 = 200,
    geteuid32 = 201,
    getegid32 = 202,
    setreuid32 = 203,
    setregid32 = 204,
    getgroups32 = 205,
    setgroups32 = 206,
    fchown32 = 207,
    setresuid32 = 208,
    getresuid32 = 209,
    setresgid32 = 210,
    getresgid32 = 211,
    lchown32 = 212,
    setuid32 = 213,
    setgid32 = 214,
    setfsuid32 = 215,
    setfsgid32 = 216,
    pivot_root = 217,
    getdents64 = 220,
    gettid = 221,
    tkill = 222,
    setxattr = 223,
    lsetxattr = 224,
    fsetxattr = 225,
    getxattr = 226,
    lgetxattr = 227,
    fgetxattr = 228,
    listxattr = 229,
    llistxattr = 230,
    flistxattr = 231,
    removexattr = 232,
    lremovexattr = 233,
    fremovexattr = 234,
    futex = 235,
    sendfile64 = 236,
    mincore = 237,
    madvise = 238,
    fcntl64 = 239,
    readahead = 240,
    io_setup = 241,
    io_destroy = 242,
    io_getevents = 243,
    io_submit = 244,
    io_cancel = 245,
    fadvise64 = 246,
    exit_group = 247,
    lookup_dcookie = 248,
    epoll_create = 249,
    epoll_ctl = 250,
    epoll_wait = 251,
    remap_file_pages = 252,
    set_tid_address = 253,
    timer_create = 254,
    timer_settime = 255,
    timer_gettime = 256,
    timer_getoverrun = 257,
    timer_delete = 258,
    clock_settime = 259,
    clock_gettime = 260,
    clock_getres = 261,
    clock_nanosleep = 262,
    statfs64 = 263,
    fstatfs64 = 264,
    tgkill = 265,
    utimes = 266,
    fadvise64_64 = 267,
    mbind = 268,
    get_mempolicy = 269,
    set_mempolicy = 270,
    mq_open = 271,
    mq_unlink = 272,
    mq_timedsend = 273,
    mq_timedreceive = 274,
    mq_notify = 275,
    mq_getsetattr = 276,
    waitid = 277,
    add_key = 279,
    request_key = 280,
    keyctl = 281,
    ioprio_set = 282,
    ioprio_get = 283,
    inotify_init = 284,
    inotify_add_watch = 285,
    inotify_rm_watch = 286,
    migrate_pages = 287,
    openat = 288,
    mkdirat = 289,
    mknodat = 290,
    fchownat = 291,
    futimesat = 292,
    fstatat64 = 293,
    unlinkat = 294,
    renameat = 295,
    linkat = 296,
    symlinkat = 297,
    readlinkat = 298,
    fchmodat = 299,
    faccessat = 300,
    pselect6 = 301,
    ppoll = 302,
    unshare = 303,
    set_robust_list = 304,
    get_robust_list = 305,
    splice = 306,
    sync_file_range = 307,
    tee = 308,
    vmsplice = 309,
    move_pages = 310,
    sched_setaffinity = 311,
    sched_getaffinity = 312,
    kexec_load = 313,
    getcpu = 314,
    epoll_pwait = 315,
    utimensat = 316,
    signalfd = 317,
    timerfd_create = 318,
    eventfd = 319,
    fallocate = 320,
    timerfd_settime = 321,
    timerfd_gettime = 322,
    signalfd4 = 323,
    eventfd2 = 324,
    epoll_create1 = 325,
    dup3 = 326,
    pipe2 = 327,
    inotify_init1 = 328,
    preadv = 329,
    pwritev = 330,
    rt_tgsigqueueinfo = 331,
    perf_event_open = 332,
    get_thread_area = 333,
    set_thread_area = 334,
    atomic_cmpxchg_32 = 335,
    atomic_barrier = 336,
    fanotify_init = 337,
    fanotify_mark = 338,
    prlimit64 = 339,
    name_to_handle_at = 340,
    open_by_handle_at = 341,
    clock_adjtime = 342,
    syncfs = 343,
    setns = 344,
    process_vm_readv = 345,
    process_vm_writev = 346,
    kcmp = 347,
    finit_module = 348,
    sched_setattr = 349,
    sched_getattr = 350,
    renameat2 = 351,
    getrandom = 352,
    memfd_create = 353,
    bpf = 354,
    execveat = 355,
    socket = 356,
    socketpair = 357,
    bind = 358,
    connect = 359,
    listen = 360,
    accept4 = 361,
    getsockopt = 362,
    setsockopt = 363,
    getsockname = 364,
    getpeername = 365,
    sendto = 366,
    sendmsg = 367,
    recvfrom = 368,
    recvmsg = 369,
    shutdown = 370,
    recvmmsg = 371,
    sendmmsg = 372,
    userfaultfd = 373,
    membarrier = 374,
    mlock2 = 375,
    copy_file_range = 376,
    preadv2 = 377,
    pwritev2 = 378,
    statx = 379,
    seccomp = 380,
    pkey_mprotect = 381,
    pkey_alloc = 382,
    pkey_free = 383,
    rseq = 384,
    semget = 393,
    semctl = 394,
    shmget = 395,
    shmctl = 396,
    shmat = 397,
    shmdt = 398,
    msgget = 399,
    msgsnd = 400,
    msgrcv = 401,
    msgctl = 402,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const MipsO32 = enum(usize) {
    syscall = 4000,
    exit = 4001,
    fork = 4002,
    read = 4003,
    write = 4004,
    open = 4005,
    close = 4006,
    waitpid = 4007,
    creat = 4008,
    link = 4009,
    unlink = 4010,
    execve = 4011,
    chdir = 4012,
    time = 4013,
    mknod = 4014,
    chmod = 4015,
    lchown = 4016,
    @"break" = 4017,
    lseek = 4019,
    getpid = 4020,
    mount = 4021,
    umount = 4022,
    setuid = 4023,
    getuid = 4024,
    stime = 4025,
    ptrace = 4026,
    alarm = 4027,
    pause = 4029,
    utime = 4030,
    stty = 4031,
    gtty = 4032,
    access = 4033,
    nice = 4034,
    ftime = 4035,
    sync = 4036,
    kill = 4037,
    rename = 4038,
    mkdir = 4039,
    rmdir = 4040,
    dup = 4041,
    pipe = 4042,
    times = 4043,
    prof = 4044,
    brk = 4045,
    setgid = 4046,
    getgid = 4047,
    signal = 4048,
    geteuid = 4049,
    getegid = 4050,
    acct = 4051,
    umount2 = 4052,
    lock = 4053,
    ioctl = 4054,
    fcntl = 4055,
    mpx = 4056,
    setpgid = 4057,
    ulimit = 4058,
    umask = 4060,
    chroot = 4061,
    ustat = 4062,
    dup2 = 4063,
    getppid = 4064,
    getpgrp = 4065,
    setsid = 4066,
    sigaction = 4067,
    sgetmask = 4068,
    ssetmask = 4069,
    setreuid = 4070,
    setregid = 4071,
    sigsuspend = 4072,
    sigpending = 4073,
    sethostname = 4074,
    setrlimit = 4075,
    getrlimit = 4076,
    getrusage = 4077,
    gettimeofday = 4078,
    settimeofday = 4079,
    getgroups = 4080,
    setgroups = 4081,
    symlink = 4083,
    readlink = 4085,
    uselib = 4086,
    swapon = 4087,
    reboot = 4088,
    readdir = 4089,
    mmap = 4090,
    munmap = 4091,
    truncate = 4092,
    ftruncate = 4093,
    fchmod = 4094,
    fchown = 4095,
    getpriority = 4096,
    setpriority = 4097,
    profil = 4098,
    statfs = 4099,
    fstatfs = 4100,
    ioperm = 4101,
    socketcall = 4102,
    syslog = 4103,
    setitimer = 4104,
    getitimer = 4105,
    stat = 4106,
    lstat = 4107,
    fstat = 4108,
    iopl = 4110,
    vhangup = 4111,
    idle = 4112,
    vm86 = 4113,
    wait4 = 4114,
    swapoff = 4115,
    sysinfo = 4116,
    ipc = 4117,
    fsync = 4118,
    sigreturn = 4119,
    clone = 4120,
    setdomainname = 4121,
    uname = 4122,
    modify_ldt = 4123,
    adjtimex = 4124,
    mprotect = 4125,
    sigprocmask = 4126,
    create_module = 4127,
    init_module = 4128,
    delete_module = 4129,
    get_kernel_syms = 4130,
    quotactl = 4131,
    getpgid = 4132,
    fchdir = 4133,
    bdflush = 4134,
    sysfs = 4135,
    personality = 4136,
    afs_syscall = 4137,
    setfsuid = 4138,
    setfsgid = 4139,
    llseek = 4140,
    getdents = 4141,
    newselect = 4142,
    flock = 4143,
    msync = 4144,
    readv = 4145,
    writev = 4146,
    cacheflush = 4147,
    cachectl = 4148,
    sysmips = 4149,
    getsid = 4151,
    fdatasync = 4152,
    sysctl = 4153,
    mlock = 4154,
    munlock = 4155,
    mlockall = 4156,
    munlockall = 4157,
    sched_setparam = 4158,
    sched_getparam = 4159,
    sched_setscheduler = 4160,
    sched_getscheduler = 4161,
    sched_yield = 4162,
    sched_get_priority_max = 4163,
    sched_get_priority_min = 4164,
    sched_rr_get_interval = 4165,
    nanosleep = 4166,
    mremap = 4167,
    accept = 4168,
    bind = 4169,
    connect = 4170,
    getpeername = 4171,
    getsockname = 4172,
    getsockopt = 4173,
    listen = 4174,
    recv = 4175,
    recvfrom = 4176,
    recvmsg = 4177,
    send = 4178,
    sendmsg = 4179,
    sendto = 4180,
    setsockopt = 4181,
    shutdown = 4182,
    socket = 4183,
    socketpair = 4184,
    setresuid = 4185,
    getresuid = 4186,
    query_module = 4187,
    poll = 4188,
    nfsservctl = 4189,
    setresgid = 4190,
    getresgid = 4191,
    prctl = 4192,
    rt_sigreturn = 4193,
    rt_sigaction = 4194,
    rt_sigprocmask = 4195,
    rt_sigpending = 4196,
    rt_sigtimedwait = 4197,
    rt_sigqueueinfo = 4198,
    rt_sigsuspend = 4199,
    pread64 = 4200,
    pwrite64 = 4201,
    chown = 4202,
    getcwd = 4203,
    capget = 4204,
    capset = 4205,
    sigaltstack = 4206,
    sendfile = 4207,
    getpmsg = 4208,
    putpmsg = 4209,
    mmap2 = 4210,
    truncate64 = 4211,
    ftruncate64 = 4212,
    stat64 = 4213,
    lstat64 = 4214,
    fstat64 = 4215,
    pivot_root = 4216,
    mincore = 4217,
    madvise = 4218,
    getdents64 = 4219,
    fcntl64 = 4220,
    gettid = 4222,
    readahead = 4223,
    setxattr = 4224,
    lsetxattr = 4225,
    fsetxattr = 4226,
    getxattr = 4227,
    lgetxattr = 4228,
    fgetxattr = 4229,
    listxattr = 4230,
    llistxattr = 4231,
    flistxattr = 4232,
    removexattr = 4233,
    lremovexattr = 4234,
    fremovexattr = 4235,
    tkill = 4236,
    sendfile64 = 4237,
    futex = 4238,
    sched_setaffinity = 4239,
    sched_getaffinity = 4240,
    io_setup = 4241,
    io_destroy = 4242,
    io_getevents = 4243,
    io_submit = 4244,
    io_cancel = 4245,
    exit_group = 4246,
    lookup_dcookie = 4247,
    epoll_create = 4248,
    epoll_ctl = 4249,
    epoll_wait = 4250,
    remap_file_pages = 4251,
    set_tid_address = 4252,
    restart_syscall = 4253,
    fadvise64 = 4254,
    statfs64 = 4255,
    fstatfs64 = 4256,
    timer_create = 4257,
    timer_settime = 4258,
    timer_gettime = 4259,
    timer_getoverrun = 4260,
    timer_delete = 4261,
    clock_settime = 4262,
    clock_gettime = 4263,
    clock_getres = 4264,
    clock_nanosleep = 4265,
    tgkill = 4266,
    utimes = 4267,
    mbind = 4268,
    get_mempolicy = 4269,
    set_mempolicy = 4270,
    mq_open = 4271,
    mq_unlink = 4272,
    mq_timedsend = 4273,
    mq_timedreceive = 4274,
    mq_notify = 4275,
    mq_getsetattr = 4276,
    vserver = 4277,
    waitid = 4278,
    add_key = 4280,
    request_key = 4281,
    keyctl = 4282,
    set_thread_area = 4283,
    inotify_init = 4284,
    inotify_add_watch = 4285,
    inotify_rm_watch = 4286,
    migrate_pages = 4287,
    openat = 4288,
    mkdirat = 4289,
    mknodat = 4290,
    fchownat = 4291,
    futimesat = 4292,
    fstatat64 = 4293,
    unlinkat = 4294,
    renameat = 4295,
    linkat = 4296,
    symlinkat = 4297,
    readlinkat = 4298,
    fchmodat = 4299,
    faccessat = 4300,
    pselect6 = 4301,
    ppoll = 4302,
    unshare = 4303,
    splice = 4304,
    sync_file_range = 4305,
    tee = 4306,
    vmsplice = 4307,
    move_pages = 4308,
    set_robust_list = 4309,
    get_robust_list = 4310,
    kexec_load = 4311,
    getcpu = 4312,
    epoll_pwait = 4313,
    ioprio_set = 4314,
    ioprio_get = 4315,
    utimensat = 4316,
    signalfd = 4317,
    timerfd = 4318,
    eventfd = 4319,
    fallocate = 4320,
    timerfd_create = 4321,
    timerfd_gettime = 4322,
    timerfd_settime = 4323,
    signalfd4 = 4324,
    eventfd2 = 4325,
    epoll_create1 = 4326,
    dup3 = 4327,
    pipe2 = 4328,
    inotify_init1 = 4329,
    preadv = 4330,
    pwritev = 4331,
    rt_tgsigqueueinfo = 4332,
    perf_event_open = 4333,
    accept4 = 4334,
    recvmmsg = 4335,
    fanotify_init = 4336,
    fanotify_mark = 4337,
    prlimit64 = 4338,
    name_to_handle_at = 4339,
    open_by_handle_at = 4340,
    clock_adjtime = 4341,
    syncfs = 4342,
    sendmmsg = 4343,
    setns = 4344,
    process_vm_readv = 4345,
    process_vm_writev = 4346,
    kcmp = 4347,
    finit_module = 4348,
    sched_setattr = 4349,
    sched_getattr = 4350,
    renameat2 = 4351,
    seccomp = 4352,
    getrandom = 4353,
    memfd_create = 4354,
    bpf = 4355,
    execveat = 4356,
    userfaultfd = 4357,
    membarrier = 4358,
    mlock2 = 4359,
    copy_file_range = 4360,
    preadv2 = 4361,
    pwritev2 = 4362,
    pkey_mprotect = 4363,
    pkey_alloc = 4364,
    pkey_free = 4365,
    statx = 4366,
    rseq = 4367,
    io_pgetevents = 4368,
    semget = 4393,
    semctl = 4394,
    shmget = 4395,
    shmctl = 4396,
    shmat = 4397,
    shmdt = 4398,
    msgget = 4399,
    msgsnd = 4400,
    msgrcv = 4401,
    msgctl = 4402,
    clock_gettime64 = 4403,
    clock_settime64 = 4404,
    clock_adjtime64 = 4405,
    clock_getres_time64 = 4406,
    clock_nanosleep_time64 = 4407,
    timer_gettime64 = 4408,
    timer_settime64 = 4409,
    timerfd_gettime64 = 4410,
    timerfd_settime64 = 4411,
    utimensat_time64 = 4412,
    pselect6_time64 = 4413,
    ppoll_time64 = 4414,
    io_pgetevents_time64 = 4416,
    recvmmsg_time64 = 4417,
    mq_timedsend_time64 = 4418,
    mq_timedreceive_time64 = 4419,
    semtimedop_time64 = 4420,
    rt_sigtimedwait_time64 = 4421,
    futex_time64 = 4422,
    sched_rr_get_interval_time64 = 4423,
    pidfd_send_signal = 4424,
    io_uring_setup = 4425,
    io_uring_enter = 4426,
    io_uring_register = 4427,
    open_tree = 4428,
    move_mount = 4429,
    fsopen = 4430,
    fsconfig = 4431,
    fsmount = 4432,
    fspick = 4433,
    pidfd_open = 4434,
    clone3 = 4435,
    close_range = 4436,
    openat2 = 4437,
    pidfd_getfd = 4438,
    faccessat2 = 4439,
    process_madvise = 4440,
    epoll_pwait2 = 4441,
    mount_setattr = 4442,
    quotactl_fd = 4443,
    landlock_create_ruleset = 4444,
    landlock_add_rule = 4445,
    landlock_restrict_self = 4446,
    process_mrelease = 4448,
    futex_waitv = 4449,
    set_mempolicy_home_node = 4450,
    cachestat = 4451,
    fchmodat2 = 4452,
    map_shadow_stack = 4453,
    futex_wake = 4454,
    futex_wait = 4455,
    futex_requeue = 4456,
    statmount = 4457,
    listmount = 4458,
    lsm_get_self_attr = 4459,
    lsm_set_self_attr = 4460,
    lsm_list_modules = 4461,
    mseal = 4462,
    setxattrat = 4463,
    getxattrat = 4464,
    listxattrat = 4465,
    removexattrat = 4466,
    open_tree_attr = 4467,
    file_getattr = 4468,
    file_setattr = 4469,
    listns = 4470,
    rseq_slice_yield = 4471,
};

pub const MipsN64 = enum(usize) {
    read = 5000,
    write = 5001,
    open = 5002,
    close = 5003,
    stat = 5004,
    fstat = 5005,
    lstat = 5006,
    poll = 5007,
    lseek = 5008,
    mmap = 5009,
    mprotect = 5010,
    munmap = 5011,
    brk = 5012,
    rt_sigaction = 5013,
    rt_sigprocmask = 5014,
    ioctl = 5015,
    pread64 = 5016,
    pwrite64 = 5017,
    readv = 5018,
    writev = 5019,
    access = 5020,
    pipe = 5021,
    newselect = 5022,
    sched_yield = 5023,
    mremap = 5024,
    msync = 5025,
    mincore = 5026,
    madvise = 5027,
    shmget = 5028,
    shmat = 5029,
    shmctl = 5030,
    dup = 5031,
    dup2 = 5032,
    pause = 5033,
    nanosleep = 5034,
    getitimer = 5035,
    setitimer = 5036,
    alarm = 5037,
    getpid = 5038,
    sendfile = 5039,
    socket = 5040,
    connect = 5041,
    accept = 5042,
    sendto = 5043,
    recvfrom = 5044,
    sendmsg = 5045,
    recvmsg = 5046,
    shutdown = 5047,
    bind = 5048,
    listen = 5049,
    getsockname = 5050,
    getpeername = 5051,
    socketpair = 5052,
    setsockopt = 5053,
    getsockopt = 5054,
    clone = 5055,
    fork = 5056,
    execve = 5057,
    exit = 5058,
    wait4 = 5059,
    kill = 5060,
    uname = 5061,
    semget = 5062,
    semop = 5063,
    semctl = 5064,
    shmdt = 5065,
    msgget = 5066,
    msgsnd = 5067,
    msgrcv = 5068,
    msgctl = 5069,
    fcntl = 5070,
    flock = 5071,
    fsync = 5072,
    fdatasync = 5073,
    truncate = 5074,
    ftruncate = 5075,
    getdents = 5076,
    getcwd = 5077,
    chdir = 5078,
    fchdir = 5079,
    rename = 5080,
    mkdir = 5081,
    rmdir = 5082,
    creat = 5083,
    link = 5084,
    unlink = 5085,
    symlink = 5086,
    readlink = 5087,
    chmod = 5088,
    fchmod = 5089,
    chown = 5090,
    fchown = 5091,
    lchown = 5092,
    umask = 5093,
    gettimeofday = 5094,
    getrlimit = 5095,
    getrusage = 5096,
    sysinfo = 5097,
    times = 5098,
    ptrace = 5099,
    getuid = 5100,
    syslog = 5101,
    getgid = 5102,
    setuid = 5103,
    setgid = 5104,
    geteuid = 5105,
    getegid = 5106,
    setpgid = 5107,
    getppid = 5108,
    getpgrp = 5109,
    setsid = 5110,
    setreuid = 5111,
    setregid = 5112,
    getgroups = 5113,
    setgroups = 5114,
    setresuid = 5115,
    getresuid = 5116,
    setresgid = 5117,
    getresgid = 5118,
    getpgid = 5119,
    setfsuid = 5120,
    setfsgid = 5121,
    getsid = 5122,
    capget = 5123,
    capset = 5124,
    rt_sigpending = 5125,
    rt_sigtimedwait = 5126,
    rt_sigqueueinfo = 5127,
    rt_sigsuspend = 5128,
    sigaltstack = 5129,
    utime = 5130,
    mknod = 5131,
    personality = 5132,
    ustat = 5133,
    statfs = 5134,
    fstatfs = 5135,
    sysfs = 5136,
    getpriority = 5137,
    setpriority = 5138,
    sched_setparam = 5139,
    sched_getparam = 5140,
    sched_setscheduler = 5141,
    sched_getscheduler = 5142,
    sched_get_priority_max = 5143,
    sched_get_priority_min = 5144,
    sched_rr_get_interval = 5145,
    mlock = 5146,
    munlock = 5147,
    mlockall = 5148,
    munlockall = 5149,
    vhangup = 5150,
    pivot_root = 5151,
    sysctl = 5152,
    prctl = 5153,
    adjtimex = 5154,
    setrlimit = 5155,
    chroot = 5156,
    sync = 5157,
    acct = 5158,
    settimeofday = 5159,
    mount = 5160,
    umount2 = 5161,
    swapon = 5162,
    swapoff = 5163,
    reboot = 5164,
    sethostname = 5165,
    setdomainname = 5166,
    create_module = 5167,
    init_module = 5168,
    delete_module = 5169,
    get_kernel_syms = 5170,
    query_module = 5171,
    quotactl = 5172,
    nfsservctl = 5173,
    getpmsg = 5174,
    putpmsg = 5175,
    afs_syscall = 5176,
    gettid = 5178,
    readahead = 5179,
    setxattr = 5180,
    lsetxattr = 5181,
    fsetxattr = 5182,
    getxattr = 5183,
    lgetxattr = 5184,
    fgetxattr = 5185,
    listxattr = 5186,
    llistxattr = 5187,
    flistxattr = 5188,
    removexattr = 5189,
    lremovexattr = 5190,
    fremovexattr = 5191,
    tkill = 5192,
    futex = 5194,
    sched_setaffinity = 5195,
    sched_getaffinity = 5196,
    cacheflush = 5197,
    cachectl = 5198,
    sysmips = 5199,
    io_setup = 5200,
    io_destroy = 5201,
    io_getevents = 5202,
    io_submit = 5203,
    io_cancel = 5204,
    exit_group = 5205,
    lookup_dcookie = 5206,
    epoll_create = 5207,
    epoll_ctl = 5208,
    epoll_wait = 5209,
    remap_file_pages = 5210,
    rt_sigreturn = 5211,
    set_tid_address = 5212,
    restart_syscall = 5213,
    semtimedop = 5214,
    fadvise64 = 5215,
    timer_create = 5216,
    timer_settime = 5217,
    timer_gettime = 5218,
    timer_getoverrun = 5219,
    timer_delete = 5220,
    clock_settime = 5221,
    clock_gettime = 5222,
    clock_getres = 5223,
    clock_nanosleep = 5224,
    tgkill = 5225,
    utimes = 5226,
    mbind = 5227,
    get_mempolicy = 5228,
    set_mempolicy = 5229,
    mq_open = 5230,
    mq_unlink = 5231,
    mq_timedsend = 5232,
    mq_timedreceive = 5233,
    mq_notify = 5234,
    mq_getsetattr = 5235,
    vserver = 5236,
    waitid = 5237,
    add_key = 5239,
    request_key = 5240,
    keyctl = 5241,
    set_thread_area = 5242,
    inotify_init = 5243,
    inotify_add_watch = 5244,
    inotify_rm_watch = 5245,
    migrate_pages = 5246,
    openat = 5247,
    mkdirat = 5248,
    mknodat = 5249,
    fchownat = 5250,
    futimesat = 5251,
    newfstatat = 5252,
    unlinkat = 5253,
    renameat = 5254,
    linkat = 5255,
    symlinkat = 5256,
    readlinkat = 5257,
    fchmodat = 5258,
    faccessat = 5259,
    pselect6 = 5260,
    ppoll = 5261,
    unshare = 5262,
    splice = 5263,
    sync_file_range = 5264,
    tee = 5265,
    vmsplice = 5266,
    move_pages = 5267,
    set_robust_list = 5268,
    get_robust_list = 5269,
    kexec_load = 5270,
    getcpu = 5271,
    epoll_pwait = 5272,
    ioprio_set = 5273,
    ioprio_get = 5274,
    utimensat = 5275,
    signalfd = 5276,
    timerfd = 5277,
    eventfd = 5278,
    fallocate = 5279,
    timerfd_create = 5280,
    timerfd_gettime = 5281,
    timerfd_settime = 5282,
    signalfd4 = 5283,
    eventfd2 = 5284,
    epoll_create1 = 5285,
    dup3 = 5286,
    pipe2 = 5287,
    inotify_init1 = 5288,
    preadv = 5289,
    pwritev = 5290,
    rt_tgsigqueueinfo = 5291,
    perf_event_open = 5292,
    accept4 = 5293,
    recvmmsg = 5294,
    fanotify_init = 5295,
    fanotify_mark = 5296,
    prlimit64 = 5297,
    name_to_handle_at = 5298,
    open_by_handle_at = 5299,
    clock_adjtime = 5300,
    syncfs = 5301,
    sendmmsg = 5302,
    setns = 5303,
    process_vm_readv = 5304,
    process_vm_writev = 5305,
    kcmp = 5306,
    finit_module = 5307,
    getdents64 = 5308,
    sched_setattr = 5309,
    sched_getattr = 5310,
    renameat2 = 5311,
    seccomp = 5312,
    getrandom = 5313,
    memfd_create = 5314,
    bpf = 5315,
    execveat = 5316,
    userfaultfd = 5317,
    membarrier = 5318,
    mlock2 = 5319,
    copy_file_range = 5320,
    preadv2 = 5321,
    pwritev2 = 5322,
    pkey_mprotect = 5323,
    pkey_alloc = 5324,
    pkey_free = 5325,
    statx = 5326,
    rseq = 5327,
    io_pgetevents = 5328,
    pidfd_send_signal = 5424,
    io_uring_setup = 5425,
    io_uring_enter = 5426,
    io_uring_register = 5427,
    open_tree = 5428,
    move_mount = 5429,
    fsopen = 5430,
    fsconfig = 5431,
    fsmount = 5432,
    fspick = 5433,
    pidfd_open = 5434,
    clone3 = 5435,
    close_range = 5436,
    openat2 = 5437,
    pidfd_getfd = 5438,
    faccessat2 = 5439,
    process_madvise = 5440,
    epoll_pwait2 = 5441,
    mount_setattr = 5442,
    quotactl_fd = 5443,
    landlock_create_ruleset = 5444,
    landlock_add_rule = 5445,
    landlock_restrict_self = 5446,
    process_mrelease = 5448,
    futex_waitv = 5449,
    set_mempolicy_home_node = 5450,
    cachestat = 5451,
    fchmodat2 = 5452,
    map_shadow_stack = 5453,
    futex_wake = 5454,
    futex_wait = 5455,
    futex_requeue = 5456,
    statmount = 5457,
    listmount = 5458,
    lsm_get_self_attr = 5459,
    lsm_set_self_attr = 5460,
    lsm_list_modules = 5461,
    mseal = 5462,
    setxattrat = 5463,
    getxattrat = 5464,
    listxattrat = 5465,
    removexattrat = 5466,
    open_tree_attr = 5467,
    file_getattr = 5468,
    file_setattr = 5469,
    listns = 5470,
    rseq_slice_yield = 5471,
};

pub const MipsN32 = enum(usize) {
    read = 6000,
    write = 6001,
    open = 6002,
    close = 6003,
    stat = 6004,
    fstat = 6005,
    lstat = 6006,
    poll = 6007,
    lseek = 6008,
    mmap = 6009,
    mprotect = 6010,
    munmap = 6011,
    brk = 6012,
    rt_sigaction = 6013,
    rt_sigprocmask = 6014,
    ioctl = 6015,
    pread64 = 6016,
    pwrite64 = 6017,
    readv = 6018,
    writev = 6019,
    access = 6020,
    pipe = 6021,
    newselect = 6022,
    sched_yield = 6023,
    mremap = 6024,
    msync = 6025,
    mincore = 6026,
    madvise = 6027,
    shmget = 6028,
    shmat = 6029,
    shmctl = 6030,
    dup = 6031,
    dup2 = 6032,
    pause = 6033,
    nanosleep = 6034,
    getitimer = 6035,
    setitimer = 6036,
    alarm = 6037,
    getpid = 6038,
    sendfile = 6039,
    socket = 6040,
    connect = 6041,
    accept = 6042,
    sendto = 6043,
    recvfrom = 6044,
    sendmsg = 6045,
    recvmsg = 6046,
    shutdown = 6047,
    bind = 6048,
    listen = 6049,
    getsockname = 6050,
    getpeername = 6051,
    socketpair = 6052,
    setsockopt = 6053,
    getsockopt = 6054,
    clone = 6055,
    fork = 6056,
    execve = 6057,
    exit = 6058,
    wait4 = 6059,
    kill = 6060,
    uname = 6061,
    semget = 6062,
    semop = 6063,
    semctl = 6064,
    shmdt = 6065,
    msgget = 6066,
    msgsnd = 6067,
    msgrcv = 6068,
    msgctl = 6069,
    fcntl = 6070,
    flock = 6071,
    fsync = 6072,
    fdatasync = 6073,
    truncate = 6074,
    ftruncate = 6075,
    getdents = 6076,
    getcwd = 6077,
    chdir = 6078,
    fchdir = 6079,
    rename = 6080,
    mkdir = 6081,
    rmdir = 6082,
    creat = 6083,
    link = 6084,
    unlink = 6085,
    symlink = 6086,
    readlink = 6087,
    chmod = 6088,
    fchmod = 6089,
    chown = 6090,
    fchown = 6091,
    lchown = 6092,
    umask = 6093,
    gettimeofday = 6094,
    getrlimit = 6095,
    getrusage = 6096,
    sysinfo = 6097,
    times = 6098,
    ptrace = 6099,
    getuid = 6100,
    syslog = 6101,
    getgid = 6102,
    setuid = 6103,
    setgid = 6104,
    geteuid = 6105,
    getegid = 6106,
    setpgid = 6107,
    getppid = 6108,
    getpgrp = 6109,
    setsid = 6110,
    setreuid = 6111,
    setregid = 6112,
    getgroups = 6113,
    setgroups = 6114,
    setresuid = 6115,
    getresuid = 6116,
    setresgid = 6117,
    getresgid = 6118,
    getpgid = 6119,
    setfsuid = 6120,
    setfsgid = 6121,
    getsid = 6122,
    capget = 6123,
    capset = 6124,
    rt_sigpending = 6125,
    rt_sigtimedwait = 6126,
    rt_sigqueueinfo = 6127,
    rt_sigsuspend = 6128,
    sigaltstack = 6129,
    utime = 6130,
    mknod = 6131,
    personality = 6132,
    ustat = 6133,
    statfs = 6134,
    fstatfs = 6135,
    sysfs = 6136,
    getpriority = 6137,
    setpriority = 6138,
    sched_setparam = 6139,
    sched_getparam = 6140,
    sched_setscheduler = 6141,
    sched_getscheduler = 6142,
    sched_get_priority_max = 6143,
    sched_get_priority_min = 6144,
    sched_rr_get_interval = 6145,
    mlock = 6146,
    munlock = 6147,
    mlockall = 6148,
    munlockall = 6149,
    vhangup = 6150,
    pivot_root = 6151,
    sysctl = 6152,
    prctl = 6153,
    adjtimex = 6154,
    setrlimit = 6155,
    chroot = 6156,
    sync = 6157,
    acct = 6158,
    settimeofday = 6159,
    mount = 6160,
    umount2 = 6161,
    swapon = 6162,
    swapoff = 6163,
    reboot = 6164,
    sethostname = 6165,
    setdomainname = 6166,
    create_module = 6167,
    init_module = 6168,
    delete_module = 6169,
    get_kernel_syms = 6170,
    query_module = 6171,
    quotactl = 6172,
    nfsservctl = 6173,
    getpmsg = 6174,
    putpmsg = 6175,
    afs_syscall = 6176,
    gettid = 6178,
    readahead = 6179,
    setxattr = 6180,
    lsetxattr = 6181,
    fsetxattr = 6182,
    getxattr = 6183,
    lgetxattr = 6184,
    fgetxattr = 6185,
    listxattr = 6186,
    llistxattr = 6187,
    flistxattr = 6188,
    removexattr = 6189,
    lremovexattr = 6190,
    fremovexattr = 6191,
    tkill = 6192,
    futex = 6194,
    sched_setaffinity = 6195,
    sched_getaffinity = 6196,
    cacheflush = 6197,
    cachectl = 6198,
    sysmips = 6199,
    io_setup = 6200,
    io_destroy = 6201,
    io_getevents = 6202,
    io_submit = 6203,
    io_cancel = 6204,
    exit_group = 6205,
    lookup_dcookie = 6206,
    epoll_create = 6207,
    epoll_ctl = 6208,
    epoll_wait = 6209,
    remap_file_pages = 6210,
    rt_sigreturn = 6211,
    fcntl64 = 6212,
    set_tid_address = 6213,
    restart_syscall = 6214,
    semtimedop = 6215,
    fadvise64 = 6216,
    statfs64 = 6217,
    fstatfs64 = 6218,
    sendfile64 = 6219,
    timer_create = 6220,
    timer_settime = 6221,
    timer_gettime = 6222,
    timer_getoverrun = 6223,
    timer_delete = 6224,
    clock_settime = 6225,
    clock_gettime = 6226,
    clock_getres = 6227,
    clock_nanosleep = 6228,
    tgkill = 6229,
    utimes = 6230,
    mbind = 6231,
    get_mempolicy = 6232,
    set_mempolicy = 6233,
    mq_open = 6234,
    mq_unlink = 6235,
    mq_timedsend = 6236,
    mq_timedreceive = 6237,
    mq_notify = 6238,
    mq_getsetattr = 6239,
    vserver = 6240,
    waitid = 6241,
    add_key = 6243,
    request_key = 6244,
    keyctl = 6245,
    set_thread_area = 6246,
    inotify_init = 6247,
    inotify_add_watch = 6248,
    inotify_rm_watch = 6249,
    migrate_pages = 6250,
    openat = 6251,
    mkdirat = 6252,
    mknodat = 6253,
    fchownat = 6254,
    futimesat = 6255,
    newfstatat = 6256,
    unlinkat = 6257,
    renameat = 6258,
    linkat = 6259,
    symlinkat = 6260,
    readlinkat = 6261,
    fchmodat = 6262,
    faccessat = 6263,
    pselect6 = 6264,
    ppoll = 6265,
    unshare = 6266,
    splice = 6267,
    sync_file_range = 6268,
    tee = 6269,
    vmsplice = 6270,
    move_pages = 6271,
    set_robust_list = 6272,
    get_robust_list = 6273,
    kexec_load = 6274,
    getcpu = 6275,
    epoll_pwait = 6276,
    ioprio_set = 6277,
    ioprio_get = 6278,
    utimensat = 6279,
    signalfd = 6280,
    timerfd = 6281,
    eventfd = 6282,
    fallocate = 6283,
    timerfd_create = 6284,
    timerfd_gettime = 6285,
    timerfd_settime = 6286,
    signalfd4 = 6287,
    eventfd2 = 6288,
    epoll_create1 = 6289,
    dup3 = 6290,
    pipe2 = 6291,
    inotify_init1 = 6292,
    preadv = 6293,
    pwritev = 6294,
    rt_tgsigqueueinfo = 6295,
    perf_event_open = 6296,
    accept4 = 6297,
    recvmmsg = 6298,
    getdents64 = 6299,
    fanotify_init = 6300,
    fanotify_mark = 6301,
    prlimit64 = 6302,
    name_to_handle_at = 6303,
    open_by_handle_at = 6304,
    clock_adjtime = 6305,
    syncfs = 6306,
    sendmmsg = 6307,
    setns = 6308,
    process_vm_readv = 6309,
    process_vm_writev = 6310,
    kcmp = 6311,
    finit_module = 6312,
    sched_setattr = 6313,
    sched_getattr = 6314,
    renameat2 = 6315,
    seccomp = 6316,
    getrandom = 6317,
    memfd_create = 6318,
    bpf = 6319,
    execveat = 6320,
    userfaultfd = 6321,
    membarrier = 6322,
    mlock2 = 6323,
    copy_file_range = 6324,
    preadv2 = 6325,
    pwritev2 = 6326,
    pkey_mprotect = 6327,
    pkey_alloc = 6328,
    pkey_free = 6329,
    statx = 6330,
    rseq = 6331,
    io_pgetevents = 6332,
    clock_gettime64 = 6403,
    clock_settime64 = 6404,
    clock_adjtime64 = 6405,
    clock_getres_time64 = 6406,
    clock_nanosleep_time64 = 6407,
    timer_gettime64 = 6408,
    timer_settime64 = 6409,
    timerfd_gettime64 = 6410,
    timerfd_settime64 = 6411,
    utimensat_time64 = 6412,
    pselect6_time64 = 6413,
    ppoll_time64 = 6414,
    io_pgetevents_time64 = 6416,
    recvmmsg_time64 = 6417,
    mq_timedsend_time64 = 6418,
    mq_timedreceive_time64 = 6419,
    semtimedop_time64 = 6420,
    rt_sigtimedwait_time64 = 6421,
    futex_time64 = 6422,
    sched_rr_get_interval_time64 = 6423,
    pidfd_send_signal = 6424,
    io_uring_setup = 6425,
    io_uring_enter = 6426,
    io_uring_register = 6427,
    open_tree = 6428,
    move_mount = 6429,
    fsopen = 6430,
    fsconfig = 6431,
    fsmount = 6432,
    fspick = 6433,
    pidfd_open = 6434,
    clone3 = 6435,
    close_range = 6436,
    openat2 = 6437,
    pidfd_getfd = 6438,
    faccessat2 = 6439,
    process_madvise = 6440,
    epoll_pwait2 = 6441,
    mount_setattr = 6442,
    quotactl_fd = 6443,
    landlock_create_ruleset = 6444,
    landlock_add_rule = 6445,
    landlock_restrict_self = 6446,
    process_mrelease = 6448,
    futex_waitv = 6449,
    set_mempolicy_home_node = 6450,
    cachestat = 6451,
    fchmodat2 = 6452,
    map_shadow_stack = 6453,
    futex_wake = 6454,
    futex_wait = 6455,
    futex_requeue = 6456,
    statmount = 6457,
    listmount = 6458,
    lsm_get_self_attr = 6459,
    lsm_set_self_attr = 6460,
    lsm_list_modules = 6461,
    mseal = 6462,
    setxattrat = 6463,
    getxattrat = 6464,
    listxattrat = 6465,
    removexattrat = 6466,
    open_tree_attr = 6467,
    file_getattr = 6468,
    file_setattr = 6469,
    listns = 6470,
    rseq_slice_yield = 6471,
};

pub const PowerPC = enum(usize) {
    restart_syscall = 0,
    exit = 1,
    fork = 2,
    read = 3,
    write = 4,
    open = 5,
    close = 6,
    waitpid = 7,
    creat = 8,
    link = 9,
    unlink = 10,
    execve = 11,
    chdir = 12,
    time = 13,
    mknod = 14,
    chmod = 15,
    lchown = 16,
    @"break" = 17,
    oldstat = 18,
    lseek = 19,
    getpid = 20,
    mount = 21,
    umount = 22,
    setuid = 23,
    getuid = 24,
    stime = 25,
    ptrace = 26,
    alarm = 27,
    oldfstat = 28,
    pause = 29,
    utime = 30,
    stty = 31,
    gtty = 32,
    access = 33,
    nice = 34,
    ftime = 35,
    sync = 36,
    kill = 37,
    rename = 38,
    mkdir = 39,
    rmdir = 40,
    dup = 41,
    pipe = 42,
    times = 43,
    prof = 44,
    brk = 45,
    setgid = 46,
    getgid = 47,
    signal = 48,
    geteuid = 49,
    getegid = 50,
    acct = 51,
    umount2 = 52,
    lock = 53,
    ioctl = 54,
    fcntl = 55,
    mpx = 56,
    setpgid = 57,
    ulimit = 58,
    oldolduname = 59,
    umask = 60,
    chroot = 61,
    ustat = 62,
    dup2 = 63,
    getppid = 64,
    getpgrp = 65,
    setsid = 66,
    sigaction = 67,
    sgetmask = 68,
    ssetmask = 69,
    setreuid = 70,
    setregid = 71,
    sigsuspend = 72,
    sigpending = 73,
    sethostname = 74,
    setrlimit = 75,
    getrlimit = 76,
    getrusage = 77,
    gettimeofday = 78,
    settimeofday = 79,
    getgroups = 80,
    setgroups = 81,
    select = 82,
    symlink = 83,
    oldlstat = 84,
    readlink = 85,
    uselib = 86,
    swapon = 87,
    reboot = 88,
    readdir = 89,
    mmap = 90,
    munmap = 91,
    truncate = 92,
    ftruncate = 93,
    fchmod = 94,
    fchown = 95,
    getpriority = 96,
    setpriority = 97,
    profil = 98,
    statfs = 99,
    fstatfs = 100,
    ioperm = 101,
    socketcall = 102,
    syslog = 103,
    setitimer = 104,
    getitimer = 105,
    stat = 106,
    lstat = 107,
    fstat = 108,
    olduname = 109,
    iopl = 110,
    vhangup = 111,
    idle = 112,
    vm86 = 113,
    wait4 = 114,
    swapoff = 115,
    sysinfo = 116,
    ipc = 117,
    fsync = 118,
    sigreturn = 119,
    clone = 120,
    setdomainname = 121,
    uname = 122,
    modify_ldt = 123,
    adjtimex = 124,
    mprotect = 125,
    sigprocmask = 126,
    create_module = 127,
    init_module = 128,
    delete_module = 129,
    get_kernel_syms = 130,
    quotactl = 131,
    getpgid = 132,
    fchdir = 133,
    bdflush = 134,
    sysfs = 135,
    personality = 136,
    afs_syscall = 137,
    setfsuid = 138,
    setfsgid = 139,
    llseek = 140,
    getdents = 141,
    newselect = 142,
    flock = 143,
    msync = 144,
    readv = 145,
    writev = 146,
    getsid = 147,
    fdatasync = 148,
    sysctl = 149,
    mlock = 150,
    munlock = 151,
    mlockall = 152,
    munlockall = 153,
    sched_setparam = 154,
    sched_getparam = 155,
    sched_setscheduler = 156,
    sched_getscheduler = 157,
    sched_yield = 158,
    sched_get_priority_max = 159,
    sched_get_priority_min = 160,
    sched_rr_get_interval = 161,
    nanosleep = 162,
    mremap = 163,
    setresuid = 164,
    getresuid = 165,
    query_module = 166,
    poll = 167,
    nfsservctl = 168,
    setresgid = 169,
    getresgid = 170,
    prctl = 171,
    rt_sigreturn = 172,
    rt_sigaction = 173,
    rt_sigprocmask = 174,
    rt_sigpending = 175,
    rt_sigtimedwait = 176,
    rt_sigqueueinfo = 177,
    rt_sigsuspend = 178,
    pread64 = 179,
    pwrite64 = 180,
    chown = 181,
    getcwd = 182,
    capget = 183,
    capset = 184,
    sigaltstack = 185,
    sendfile = 186,
    getpmsg = 187,
    putpmsg = 188,
    vfork = 189,
    ugetrlimit = 190,
    readahead = 191,
    mmap2 = 192,
    truncate64 = 193,
    ftruncate64 = 194,
    stat64 = 195,
    lstat64 = 196,
    fstat64 = 197,
    pciconfig_read = 198,
    pciconfig_write = 199,
    pciconfig_iobase = 200,
    multiplexer = 201,
    getdents64 = 202,
    pivot_root = 203,
    fcntl64 = 204,
    madvise = 205,
    mincore = 206,
    gettid = 207,
    tkill = 208,
    setxattr = 209,
    lsetxattr = 210,
    fsetxattr = 211,
    getxattr = 212,
    lgetxattr = 213,
    fgetxattr = 214,
    listxattr = 215,
    llistxattr = 216,
    flistxattr = 217,
    removexattr = 218,
    lremovexattr = 219,
    fremovexattr = 220,
    futex = 221,
    sched_setaffinity = 222,
    sched_getaffinity = 223,
    tuxcall = 225,
    sendfile64 = 226,
    io_setup = 227,
    io_destroy = 228,
    io_getevents = 229,
    io_submit = 230,
    io_cancel = 231,
    set_tid_address = 232,
    fadvise64 = 233,
    exit_group = 234,
    lookup_dcookie = 235,
    epoll_create = 236,
    epoll_ctl = 237,
    epoll_wait = 238,
    remap_file_pages = 239,
    timer_create = 240,
    timer_settime = 241,
    timer_gettime = 242,
    timer_getoverrun = 243,
    timer_delete = 244,
    clock_settime = 245,
    clock_gettime = 246,
    clock_getres = 247,
    clock_nanosleep = 248,
    swapcontext = 249,
    tgkill = 250,
    utimes = 251,
    statfs64 = 252,
    fstatfs64 = 253,
    fadvise64_64 = 254,
    rtas = 255,
    sys_debug_setcontext = 256,
    migrate_pages = 258,
    mbind = 259,
    get_mempolicy = 260,
    set_mempolicy = 261,
    mq_open = 262,
    mq_unlink = 263,
    mq_timedsend = 264,
    mq_timedreceive = 265,
    mq_notify = 266,
    mq_getsetattr = 267,
    kexec_load = 268,
    add_key = 269,
    request_key = 270,
    keyctl = 271,
    waitid = 272,
    ioprio_set = 273,
    ioprio_get = 274,
    inotify_init = 275,
    inotify_add_watch = 276,
    inotify_rm_watch = 277,
    spu_run = 278,
    spu_create = 279,
    pselect6 = 280,
    ppoll = 281,
    unshare = 282,
    splice = 283,
    tee = 284,
    vmsplice = 285,
    openat = 286,
    mkdirat = 287,
    mknodat = 288,
    fchownat = 289,
    futimesat = 290,
    fstatat64 = 291,
    unlinkat = 292,
    renameat = 293,
    linkat = 294,
    symlinkat = 295,
    readlinkat = 296,
    fchmodat = 297,
    faccessat = 298,
    get_robust_list = 299,
    set_robust_list = 300,
    move_pages = 301,
    getcpu = 302,
    epoll_pwait = 303,
    utimensat = 304,
    signalfd = 305,
    timerfd_create = 306,
    eventfd = 307,
    sync_file_range2 = 308,
    fallocate = 309,
    subpage_prot = 310,
    timerfd_settime = 311,
    timerfd_gettime = 312,
    signalfd4 = 313,
    eventfd2 = 314,
    epoll_create1 = 315,
    dup3 = 316,
    pipe2 = 317,
    inotify_init1 = 318,
    perf_event_open = 319,
    preadv = 320,
    pwritev = 321,
    rt_tgsigqueueinfo = 322,
    fanotify_init = 323,
    fanotify_mark = 324,
    prlimit64 = 325,
    socket = 326,
    bind = 327,
    connect = 328,
    listen = 329,
    accept = 330,
    getsockname = 331,
    getpeername = 332,
    socketpair = 333,
    send = 334,
    sendto = 335,
    recv = 336,
    recvfrom = 337,
    shutdown = 338,
    setsockopt = 339,
    getsockopt = 340,
    sendmsg = 341,
    recvmsg = 342,
    recvmmsg = 343,
    accept4 = 344,
    name_to_handle_at = 345,
    open_by_handle_at = 346,
    clock_adjtime = 347,
    syncfs = 348,
    sendmmsg = 349,
    setns = 350,
    process_vm_readv = 351,
    process_vm_writev = 352,
    finit_module = 353,
    kcmp = 354,
    sched_setattr = 355,
    sched_getattr = 356,
    renameat2 = 357,
    seccomp = 358,
    getrandom = 359,
    memfd_create = 360,
    bpf = 361,
    execveat = 362,
    switch_endian = 363,
    userfaultfd = 364,
    membarrier = 365,
    mlock2 = 378,
    copy_file_range = 379,
    preadv2 = 380,
    pwritev2 = 381,
    kexec_file_load = 382,
    statx = 383,
    pkey_alloc = 384,
    pkey_free = 385,
    pkey_mprotect = 386,
    rseq = 387,
    io_pgetevents = 388,
    semget = 393,
    semctl = 394,
    shmget = 395,
    shmctl = 396,
    shmat = 397,
    shmdt = 398,
    msgget = 399,
    msgsnd = 400,
    msgrcv = 401,
    msgctl = 402,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const PowerPC64 = enum(usize) {
    restart_syscall = 0,
    exit = 1,
    fork = 2,
    read = 3,
    write = 4,
    open = 5,
    close = 6,
    waitpid = 7,
    creat = 8,
    link = 9,
    unlink = 10,
    execve = 11,
    chdir = 12,
    time = 13,
    mknod = 14,
    chmod = 15,
    lchown = 16,
    @"break" = 17,
    oldstat = 18,
    lseek = 19,
    getpid = 20,
    mount = 21,
    umount = 22,
    setuid = 23,
    getuid = 24,
    stime = 25,
    ptrace = 26,
    alarm = 27,
    oldfstat = 28,
    pause = 29,
    utime = 30,
    stty = 31,
    gtty = 32,
    access = 33,
    nice = 34,
    ftime = 35,
    sync = 36,
    kill = 37,
    rename = 38,
    mkdir = 39,
    rmdir = 40,
    dup = 41,
    pipe = 42,
    times = 43,
    prof = 44,
    brk = 45,
    setgid = 46,
    getgid = 47,
    signal = 48,
    geteuid = 49,
    getegid = 50,
    acct = 51,
    umount2 = 52,
    lock = 53,
    ioctl = 54,
    fcntl = 55,
    mpx = 56,
    setpgid = 57,
    ulimit = 58,
    oldolduname = 59,
    umask = 60,
    chroot = 61,
    ustat = 62,
    dup2 = 63,
    getppid = 64,
    getpgrp = 65,
    setsid = 66,
    sigaction = 67,
    sgetmask = 68,
    ssetmask = 69,
    setreuid = 70,
    setregid = 71,
    sigsuspend = 72,
    sigpending = 73,
    sethostname = 74,
    setrlimit = 75,
    getrlimit = 76,
    getrusage = 77,
    gettimeofday = 78,
    settimeofday = 79,
    getgroups = 80,
    setgroups = 81,
    select = 82,
    symlink = 83,
    oldlstat = 84,
    readlink = 85,
    uselib = 86,
    swapon = 87,
    reboot = 88,
    readdir = 89,
    mmap = 90,
    munmap = 91,
    truncate = 92,
    ftruncate = 93,
    fchmod = 94,
    fchown = 95,
    getpriority = 96,
    setpriority = 97,
    profil = 98,
    statfs = 99,
    fstatfs = 100,
    ioperm = 101,
    socketcall = 102,
    syslog = 103,
    setitimer = 104,
    getitimer = 105,
    stat = 106,
    lstat = 107,
    fstat = 108,
    olduname = 109,
    iopl = 110,
    vhangup = 111,
    idle = 112,
    vm86 = 113,
    wait4 = 114,
    swapoff = 115,
    sysinfo = 116,
    ipc = 117,
    fsync = 118,
    sigreturn = 119,
    clone = 120,
    setdomainname = 121,
    uname = 122,
    modify_ldt = 123,
    adjtimex = 124,
    mprotect = 125,
    sigprocmask = 126,
    create_module = 127,
    init_module = 128,
    delete_module = 129,
    get_kernel_syms = 130,
    quotactl = 131,
    getpgid = 132,
    fchdir = 133,
    bdflush = 134,
    sysfs = 135,
    personality = 136,
    afs_syscall = 137,
    setfsuid = 138,
    setfsgid = 139,
    llseek = 140,
    getdents = 141,
    newselect = 142,
    flock = 143,
    msync = 144,
    readv = 145,
    writev = 146,
    getsid = 147,
    fdatasync = 148,
    sysctl = 149,
    mlock = 150,
    munlock = 151,
    mlockall = 152,
    munlockall = 153,
    sched_setparam = 154,
    sched_getparam = 155,
    sched_setscheduler = 156,
    sched_getscheduler = 157,
    sched_yield = 158,
    sched_get_priority_max = 159,
    sched_get_priority_min = 160,
    sched_rr_get_interval = 161,
    nanosleep = 162,
    mremap = 163,
    setresuid = 164,
    getresuid = 165,
    query_module = 166,
    poll = 167,
    nfsservctl = 168,
    setresgid = 169,
    getresgid = 170,
    prctl = 171,
    rt_sigreturn = 172,
    rt_sigaction = 173,
    rt_sigprocmask = 174,
    rt_sigpending = 175,
    rt_sigtimedwait = 176,
    rt_sigqueueinfo = 177,
    rt_sigsuspend = 178,
    pread64 = 179,
    pwrite64 = 180,
    chown = 181,
    getcwd = 182,
    capget = 183,
    capset = 184,
    sigaltstack = 185,
    sendfile = 186,
    getpmsg = 187,
    putpmsg = 188,
    vfork = 189,
    ugetrlimit = 190,
    readahead = 191,
    pciconfig_read = 198,
    pciconfig_write = 199,
    pciconfig_iobase = 200,
    multiplexer = 201,
    getdents64 = 202,
    pivot_root = 203,
    madvise = 205,
    mincore = 206,
    gettid = 207,
    tkill = 208,
    setxattr = 209,
    lsetxattr = 210,
    fsetxattr = 211,
    getxattr = 212,
    lgetxattr = 213,
    fgetxattr = 214,
    listxattr = 215,
    llistxattr = 216,
    flistxattr = 217,
    removexattr = 218,
    lremovexattr = 219,
    fremovexattr = 220,
    futex = 221,
    sched_setaffinity = 222,
    sched_getaffinity = 223,
    tuxcall = 225,
    io_setup = 227,
    io_destroy = 228,
    io_getevents = 229,
    io_submit = 230,
    io_cancel = 231,
    set_tid_address = 232,
    fadvise64 = 233,
    exit_group = 234,
    lookup_dcookie = 235,
    epoll_create = 236,
    epoll_ctl = 237,
    epoll_wait = 238,
    remap_file_pages = 239,
    timer_create = 240,
    timer_settime = 241,
    timer_gettime = 242,
    timer_getoverrun = 243,
    timer_delete = 244,
    clock_settime = 245,
    clock_gettime = 246,
    clock_getres = 247,
    clock_nanosleep = 248,
    swapcontext = 249,
    tgkill = 250,
    utimes = 251,
    statfs64 = 252,
    fstatfs64 = 253,
    rtas = 255,
    sys_debug_setcontext = 256,
    migrate_pages = 258,
    mbind = 259,
    get_mempolicy = 260,
    set_mempolicy = 261,
    mq_open = 262,
    mq_unlink = 263,
    mq_timedsend = 264,
    mq_timedreceive = 265,
    mq_notify = 266,
    mq_getsetattr = 267,
    kexec_load = 268,
    add_key = 269,
    request_key = 270,
    keyctl = 271,
    waitid = 272,
    ioprio_set = 273,
    ioprio_get = 274,
    inotify_init = 275,
    inotify_add_watch = 276,
    inotify_rm_watch = 277,
    spu_run = 278,
    spu_create = 279,
    pselect6 = 280,
    ppoll = 281,
    unshare = 282,
    splice = 283,
    tee = 284,
    vmsplice = 285,
    openat = 286,
    mkdirat = 287,
    mknodat = 288,
    fchownat = 289,
    futimesat = 290,
    newfstatat = 291,
    unlinkat = 292,
    renameat = 293,
    linkat = 294,
    symlinkat = 295,
    readlinkat = 296,
    fchmodat = 297,
    faccessat = 298,
    get_robust_list = 299,
    set_robust_list = 300,
    move_pages = 301,
    getcpu = 302,
    epoll_pwait = 303,
    utimensat = 304,
    signalfd = 305,
    timerfd_create = 306,
    eventfd = 307,
    sync_file_range2 = 308,
    fallocate = 309,
    subpage_prot = 310,
    timerfd_settime = 311,
    timerfd_gettime = 312,
    signalfd4 = 313,
    eventfd2 = 314,
    epoll_create1 = 315,
    dup3 = 316,
    pipe2 = 317,
    inotify_init1 = 318,
    perf_event_open = 319,
    preadv = 320,
    pwritev = 321,
    rt_tgsigqueueinfo = 322,
    fanotify_init = 323,
    fanotify_mark = 324,
    prlimit64 = 325,
    socket = 326,
    bind = 327,
    connect = 328,
    listen = 329,
    accept = 330,
    getsockname = 331,
    getpeername = 332,
    socketpair = 333,
    send = 334,
    sendto = 335,
    recv = 336,
    recvfrom = 337,
    shutdown = 338,
    setsockopt = 339,
    getsockopt = 340,
    sendmsg = 341,
    recvmsg = 342,
    recvmmsg = 343,
    accept4 = 344,
    name_to_handle_at = 345,
    open_by_handle_at = 346,
    clock_adjtime = 347,
    syncfs = 348,
    sendmmsg = 349,
    setns = 350,
    process_vm_readv = 351,
    process_vm_writev = 352,
    finit_module = 353,
    kcmp = 354,
    sched_setattr = 355,
    sched_getattr = 356,
    renameat2 = 357,
    seccomp = 358,
    getrandom = 359,
    memfd_create = 360,
    bpf = 361,
    execveat = 362,
    switch_endian = 363,
    userfaultfd = 364,
    membarrier = 365,
    mlock2 = 378,
    copy_file_range = 379,
    preadv2 = 380,
    pwritev2 = 381,
    kexec_file_load = 382,
    statx = 383,
    pkey_alloc = 384,
    pkey_free = 385,
    pkey_mprotect = 386,
    rseq = 387,
    io_pgetevents = 388,
    semtimedop = 392,
    semget = 393,
    semctl = 394,
    shmget = 395,
    shmctl = 396,
    shmat = 397,
    shmdt = 398,
    msgget = 399,
    msgsnd = 400,
    msgrcv = 401,
    msgctl = 402,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const S390x = enum(usize) {
    exit = 1,
    fork = 2,
    read = 3,
    write = 4,
    open = 5,
    close = 6,
    restart_syscall = 7,
    creat = 8,
    link = 9,
    unlink = 10,
    execve = 11,
    chdir = 12,
    mknod = 14,
    chmod = 15,
    lseek = 19,
    getpid = 20,
    mount = 21,
    umount = 22,
    ptrace = 26,
    alarm = 27,
    pause = 29,
    utime = 30,
    access = 33,
    nice = 34,
    sync = 36,
    kill = 37,
    rename = 38,
    mkdir = 39,
    rmdir = 40,
    dup = 41,
    pipe = 42,
    times = 43,
    brk = 45,
    signal = 48,
    acct = 51,
    umount2 = 52,
    ioctl = 54,
    fcntl = 55,
    setpgid = 57,
    umask = 60,
    chroot = 61,
    ustat = 62,
    dup2 = 63,
    getppid = 64,
    getpgrp = 65,
    setsid = 66,
    sigaction = 67,
    sigsuspend = 72,
    sigpending = 73,
    sethostname = 74,
    setrlimit = 75,
    getrusage = 77,
    gettimeofday = 78,
    settimeofday = 79,
    symlink = 83,
    readlink = 85,
    uselib = 86,
    swapon = 87,
    reboot = 88,
    readdir = 89,
    mmap = 90,
    munmap = 91,
    truncate = 92,
    ftruncate = 93,
    fchmod = 94,
    getpriority = 96,
    setpriority = 97,
    statfs = 99,
    fstatfs = 100,
    socketcall = 102,
    syslog = 103,
    setitimer = 104,
    getitimer = 105,
    stat = 106,
    lstat = 107,
    fstat = 108,
    lookup_dcookie = 110,
    vhangup = 111,
    idle = 112,
    wait4 = 114,
    swapoff = 115,
    sysinfo = 116,
    ipc = 117,
    fsync = 118,
    sigreturn = 119,
    clone = 120,
    setdomainname = 121,
    uname = 122,
    adjtimex = 124,
    mprotect = 125,
    sigprocmask = 126,
    create_module = 127,
    init_module = 128,
    delete_module = 129,
    get_kernel_syms = 130,
    quotactl = 131,
    getpgid = 132,
    fchdir = 133,
    bdflush = 134,
    sysfs = 135,
    personality = 136,
    afs_syscall = 137,
    getdents = 141,
    select = 142,
    flock = 143,
    msync = 144,
    readv = 145,
    writev = 146,
    getsid = 147,
    fdatasync = 148,
    sysctl = 149,
    mlock = 150,
    munlock = 151,
    mlockall = 152,
    munlockall = 153,
    sched_setparam = 154,
    sched_getparam = 155,
    sched_setscheduler = 156,
    sched_getscheduler = 157,
    sched_yield = 158,
    sched_get_priority_max = 159,
    sched_get_priority_min = 160,
    sched_rr_get_interval = 161,
    nanosleep = 162,
    mremap = 163,
    query_module = 167,
    poll = 168,
    nfsservctl = 169,
    prctl = 172,
    rt_sigreturn = 173,
    rt_sigaction = 174,
    rt_sigprocmask = 175,
    rt_sigpending = 176,
    rt_sigtimedwait = 177,
    rt_sigqueueinfo = 178,
    rt_sigsuspend = 179,
    pread64 = 180,
    pwrite64 = 181,
    getcwd = 183,
    capget = 184,
    capset = 185,
    sigaltstack = 186,
    sendfile = 187,
    getpmsg = 188,
    putpmsg = 189,
    vfork = 190,
    getrlimit = 191,
    lchown = 198,
    getuid = 199,
    getgid = 200,
    geteuid = 201,
    getegid = 202,
    setreuid = 203,
    setregid = 204,
    getgroups = 205,
    setgroups = 206,
    fchown = 207,
    setresuid = 208,
    getresuid = 209,
    setresgid = 210,
    getresgid = 211,
    chown = 212,
    setuid = 213,
    setgid = 214,
    setfsuid = 215,
    setfsgid = 216,
    pivot_root = 217,
    mincore = 218,
    madvise = 219,
    getdents64 = 220,
    readahead = 222,
    setxattr = 224,
    lsetxattr = 225,
    fsetxattr = 226,
    getxattr = 227,
    lgetxattr = 228,
    fgetxattr = 229,
    listxattr = 230,
    llistxattr = 231,
    flistxattr = 232,
    removexattr = 233,
    lremovexattr = 234,
    fremovexattr = 235,
    gettid = 236,
    tkill = 237,
    futex = 238,
    sched_setaffinity = 239,
    sched_getaffinity = 240,
    tgkill = 241,
    io_setup = 243,
    io_destroy = 244,
    io_getevents = 245,
    io_submit = 246,
    io_cancel = 247,
    exit_group = 248,
    epoll_create = 249,
    epoll_ctl = 250,
    epoll_wait = 251,
    set_tid_address = 252,
    fadvise64 = 253,
    timer_create = 254,
    timer_settime = 255,
    timer_gettime = 256,
    timer_getoverrun = 257,
    timer_delete = 258,
    clock_settime = 259,
    clock_gettime = 260,
    clock_getres = 261,
    clock_nanosleep = 262,
    statfs64 = 265,
    fstatfs64 = 266,
    remap_file_pages = 267,
    mbind = 268,
    get_mempolicy = 269,
    set_mempolicy = 270,
    mq_open = 271,
    mq_unlink = 272,
    mq_timedsend = 273,
    mq_timedreceive = 274,
    mq_notify = 275,
    mq_getsetattr = 276,
    kexec_load = 277,
    add_key = 278,
    request_key = 279,
    keyctl = 280,
    waitid = 281,
    ioprio_set = 282,
    ioprio_get = 283,
    inotify_init = 284,
    inotify_add_watch = 285,
    inotify_rm_watch = 286,
    migrate_pages = 287,
    openat = 288,
    mkdirat = 289,
    mknodat = 290,
    fchownat = 291,
    futimesat = 292,
    newfstatat = 293,
    unlinkat = 294,
    renameat = 295,
    linkat = 296,
    symlinkat = 297,
    readlinkat = 298,
    fchmodat = 299,
    faccessat = 300,
    pselect6 = 301,
    ppoll = 302,
    unshare = 303,
    set_robust_list = 304,
    get_robust_list = 305,
    splice = 306,
    sync_file_range = 307,
    tee = 308,
    vmsplice = 309,
    move_pages = 310,
    getcpu = 311,
    epoll_pwait = 312,
    utimes = 313,
    fallocate = 314,
    utimensat = 315,
    signalfd = 316,
    timerfd = 317,
    eventfd = 318,
    timerfd_create = 319,
    timerfd_settime = 320,
    timerfd_gettime = 321,
    signalfd4 = 322,
    eventfd2 = 323,
    inotify_init1 = 324,
    pipe2 = 325,
    dup3 = 326,
    epoll_create1 = 327,
    preadv = 328,
    pwritev = 329,
    rt_tgsigqueueinfo = 330,
    perf_event_open = 331,
    fanotify_init = 332,
    fanotify_mark = 333,
    prlimit64 = 334,
    name_to_handle_at = 335,
    open_by_handle_at = 336,
    clock_adjtime = 337,
    syncfs = 338,
    setns = 339,
    process_vm_readv = 340,
    process_vm_writev = 341,
    s390_runtime_instr = 342,
    kcmp = 343,
    finit_module = 344,
    sched_setattr = 345,
    sched_getattr = 346,
    renameat2 = 347,
    seccomp = 348,
    getrandom = 349,
    memfd_create = 350,
    bpf = 351,
    s390_pci_mmio_write = 352,
    s390_pci_mmio_read = 353,
    execveat = 354,
    userfaultfd = 355,
    membarrier = 356,
    recvmmsg = 357,
    sendmmsg = 358,
    socket = 359,
    socketpair = 360,
    bind = 361,
    connect = 362,
    listen = 363,
    accept4 = 364,
    getsockopt = 365,
    setsockopt = 366,
    getsockname = 367,
    getpeername = 368,
    sendto = 369,
    sendmsg = 370,
    recvfrom = 371,
    recvmsg = 372,
    shutdown = 373,
    mlock2 = 374,
    copy_file_range = 375,
    preadv2 = 376,
    pwritev2 = 377,
    s390_guarded_storage = 378,
    statx = 379,
    s390_sthyi = 380,
    kexec_file_load = 381,
    io_pgetevents = 382,
    rseq = 383,
    pkey_mprotect = 384,
    pkey_alloc = 385,
    pkey_free = 386,
    semtimedop = 392,
    semget = 393,
    semctl = 394,
    shmget = 395,
    shmctl = 396,
    shmat = 397,
    shmdt = 398,
    msgget = 399,
    msgsnd = 400,
    msgrcv = 401,
    msgctl = 402,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    memfd_secret = 447,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const Xtensa = enum(usize) {
    spill = 0,
    xtensa = 1,
    open = 8,
    close = 9,
    dup = 10,
    dup2 = 11,
    read = 12,
    write = 13,
    select = 14,
    lseek = 15,
    poll = 16,
    llseek = 17,
    epoll_wait = 18,
    epoll_ctl = 19,
    epoll_create = 20,
    creat = 21,
    truncate = 22,
    ftruncate = 23,
    readv = 24,
    writev = 25,
    fsync = 26,
    fdatasync = 27,
    truncate64 = 28,
    ftruncate64 = 29,
    pread64 = 30,
    pwrite64 = 31,
    link = 32,
    rename = 33,
    symlink = 34,
    readlink = 35,
    mknod = 36,
    pipe = 37,
    unlink = 38,
    rmdir = 39,
    mkdir = 40,
    chdir = 41,
    fchdir = 42,
    getcwd = 43,
    chmod = 44,
    chown = 45,
    stat = 46,
    stat64 = 47,
    lchown = 48,
    lstat = 49,
    lstat64 = 50,
    fchmod = 52,
    fchown = 53,
    fstat = 54,
    fstat64 = 55,
    flock = 56,
    access = 57,
    umask = 58,
    getdents = 59,
    getdents64 = 60,
    fcntl64 = 61,
    fallocate = 62,
    fadvise64_64 = 63,
    utime = 64,
    utimes = 65,
    ioctl = 66,
    fcntl = 67,
    setxattr = 68,
    getxattr = 69,
    listxattr = 70,
    removexattr = 71,
    lsetxattr = 72,
    lgetxattr = 73,
    llistxattr = 74,
    lremovexattr = 75,
    fsetxattr = 76,
    fgetxattr = 77,
    flistxattr = 78,
    fremovexattr = 79,
    mmap2 = 80,
    munmap = 81,
    mprotect = 82,
    brk = 83,
    mlock = 84,
    munlock = 85,
    mlockall = 86,
    munlockall = 87,
    mremap = 88,
    msync = 89,
    mincore = 90,
    madvise = 91,
    shmget = 92,
    shmat = 93,
    shmctl = 94,
    shmdt = 95,
    socket = 96,
    setsockopt = 97,
    getsockopt = 98,
    shutdown = 99,
    bind = 100,
    connect = 101,
    listen = 102,
    accept = 103,
    getsockname = 104,
    getpeername = 105,
    sendmsg = 106,
    recvmsg = 107,
    send = 108,
    recv = 109,
    sendto = 110,
    recvfrom = 111,
    socketpair = 112,
    sendfile = 113,
    sendfile64 = 114,
    sendmmsg = 115,
    clone = 116,
    execve = 117,
    exit = 118,
    exit_group = 119,
    getpid = 120,
    wait4 = 121,
    waitid = 122,
    kill = 123,
    tkill = 124,
    tgkill = 125,
    set_tid_address = 126,
    gettid = 127,
    setsid = 128,
    getsid = 129,
    prctl = 130,
    personality = 131,
    getpriority = 132,
    setpriority = 133,
    setitimer = 134,
    getitimer = 135,
    setuid = 136,
    getuid = 137,
    setgid = 138,
    getgid = 139,
    geteuid = 140,
    getegid = 141,
    setreuid = 142,
    setregid = 143,
    setresuid = 144,
    getresuid = 145,
    setresgid = 146,
    getresgid = 147,
    setpgid = 148,
    getpgid = 149,
    getppid = 150,
    getpgrp = 151,
    times = 154,
    acct = 155,
    sched_setaffinity = 156,
    sched_getaffinity = 157,
    capget = 158,
    capset = 159,
    ptrace = 160,
    semtimedop = 161,
    semget = 162,
    semop = 163,
    semctl = 164,
    msgget = 166,
    msgsnd = 167,
    msgrcv = 168,
    msgctl = 169,
    umount2 = 171,
    mount = 172,
    swapon = 173,
    chroot = 174,
    pivot_root = 175,
    umount = 176,
    swapoff = 177,
    sync = 178,
    syncfs = 179,
    setfsuid = 180,
    setfsgid = 181,
    sysfs = 182,
    ustat = 183,
    statfs = 184,
    fstatfs = 185,
    statfs64 = 186,
    fstatfs64 = 187,
    setrlimit = 188,
    getrlimit = 189,
    getrusage = 190,
    futex = 191,
    gettimeofday = 192,
    settimeofday = 193,
    adjtimex = 194,
    nanosleep = 195,
    getgroups = 196,
    setgroups = 197,
    sethostname = 198,
    setdomainname = 199,
    syslog = 200,
    vhangup = 201,
    uselib = 202,
    reboot = 203,
    quotactl = 204,
    nfsservctl = 205,
    sysctl = 206,
    bdflush = 207,
    uname = 208,
    sysinfo = 209,
    init_module = 210,
    delete_module = 211,
    sched_setparam = 212,
    sched_getparam = 213,
    sched_setscheduler = 214,
    sched_getscheduler = 215,
    sched_get_priority_max = 216,
    sched_get_priority_min = 217,
    sched_rr_get_interval = 218,
    sched_yield = 219,
    restart_syscall = 223,
    sigaltstack = 224,
    rt_sigreturn = 225,
    rt_sigaction = 226,
    rt_sigprocmask = 227,
    rt_sigpending = 228,
    rt_sigtimedwait = 229,
    rt_sigqueueinfo = 230,
    rt_sigsuspend = 231,
    mq_open = 232,
    mq_unlink = 233,
    mq_timedsend = 234,
    mq_timedreceive = 235,
    mq_notify = 236,
    mq_getsetattr = 237,
    io_setup = 239,
    io_destroy = 240,
    io_submit = 241,
    io_getevents = 242,
    io_cancel = 243,
    clock_settime = 244,
    clock_gettime = 245,
    clock_getres = 246,
    clock_nanosleep = 247,
    timer_create = 248,
    timer_delete = 249,
    timer_settime = 250,
    timer_gettime = 251,
    timer_getoverrun = 252,
    lookup_dcookie = 254,
    add_key = 256,
    request_key = 257,
    keyctl = 258,
    readahead = 260,
    remap_file_pages = 261,
    migrate_pages = 262,
    mbind = 263,
    get_mempolicy = 264,
    set_mempolicy = 265,
    unshare = 266,
    move_pages = 267,
    splice = 268,
    tee = 269,
    vmsplice = 270,
    pselect6 = 272,
    ppoll = 273,
    epoll_pwait = 274,
    epoll_create1 = 275,
    inotify_init = 276,
    inotify_add_watch = 277,
    inotify_rm_watch = 278,
    inotify_init1 = 279,
    getcpu = 280,
    kexec_load = 281,
    ioprio_set = 282,
    ioprio_get = 283,
    set_robust_list = 284,
    get_robust_list = 285,
    openat = 288,
    mkdirat = 289,
    mknodat = 290,
    unlinkat = 291,
    renameat = 292,
    linkat = 293,
    symlinkat = 294,
    readlinkat = 295,
    utimensat = 296,
    fchownat = 297,
    futimesat = 298,
    fstatat64 = 299,
    fchmodat = 300,
    faccessat = 301,
    signalfd = 304,
    eventfd = 306,
    recvmmsg = 307,
    setns = 308,
    signalfd4 = 309,
    dup3 = 310,
    pipe2 = 311,
    timerfd_create = 312,
    timerfd_settime = 313,
    timerfd_gettime = 314,
    eventfd2 = 316,
    preadv = 317,
    pwritev = 318,
    fanotify_init = 320,
    fanotify_mark = 321,
    process_vm_readv = 322,
    process_vm_writev = 323,
    name_to_handle_at = 324,
    open_by_handle_at = 325,
    sync_file_range2 = 326,
    perf_event_open = 327,
    rt_tgsigqueueinfo = 328,
    clock_adjtime = 329,
    prlimit64 = 330,
    kcmp = 331,
    finit_module = 332,
    accept4 = 333,
    sched_setattr = 334,
    sched_getattr = 335,
    renameat2 = 336,
    seccomp = 337,
    getrandom = 338,
    memfd_create = 339,
    bpf = 340,
    execveat = 341,
    userfaultfd = 342,
    membarrier = 343,
    mlock2 = 344,
    copy_file_range = 345,
    preadv2 = 346,
    pwritev2 = 347,
    pkey_mprotect = 348,
    pkey_alloc = 349,
    pkey_free = 350,
    statx = 351,
    rseq = 352,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const Alpha = enum(usize) {
    osf_syscall = 0,
    exit = 1,
    fork = 2,
    read = 3,
    write = 4,
    osf_old_open = 5,
    close = 6,
    osf_wait4 = 7,
    osf_old_creat = 8,
    link = 9,
    unlink = 10,
    osf_execve = 11,
    chdir = 12,
    fchdir = 13,
    mknod = 14,
    chmod = 15,
    chown = 16,
    brk = 17,
    osf_getfsstat = 18,
    lseek = 19,
    getpid = 20,
    osf_mount = 21,
    umount2 = 22,
    setuid = 23,
    getuid = 24,
    exec_with_loader = 25,
    ptrace = 26,
    osf_nrecvmsg = 27,
    osf_nsendmsg = 28,
    osf_nrecvfrom = 29,
    osf_naccept = 30,
    osf_ngetpeername = 31,
    osf_ngetsockname = 32,
    access = 33,
    osf_chflags = 34,
    osf_fchflags = 35,
    sync = 36,
    kill = 37,
    osf_old_stat = 38,
    setpgid = 39,
    osf_old_lstat = 40,
    dup = 41,
    pipe = 42,
    osf_set_program_attributes = 43,
    osf_profil = 44,
    open = 45,
    osf_old_sigaction = 46,
    getgid = 47,
    osf_sigprocmask = 48,
    osf_getlogin = 49,
    osf_setlogin = 50,
    acct = 51,
    sigpending = 52,
    ioctl = 54,
    osf_reboot = 55,
    osf_revoke = 56,
    symlink = 57,
    readlink = 58,
    execve = 59,
    umask = 60,
    chroot = 61,
    osf_old_fstat = 62,
    getpgrp = 63,
    getpagesize = 64,
    osf_mremap = 65,
    vfork = 66,
    stat = 67,
    lstat = 68,
    osf_sbrk = 69,
    osf_sstk = 70,
    mmap = 71,
    osf_old_vadvise = 72,
    munmap = 73,
    mprotect = 74,
    madvise = 75,
    vhangup = 76,
    osf_kmodcall = 77,
    osf_mincore = 78,
    getgroups = 79,
    setgroups = 80,
    osf_old_getpgrp = 81,
    setpgrp = 82,
    osf_setitimer = 83,
    osf_old_wait = 84,
    osf_table = 85,
    osf_getitimer = 86,
    gethostname = 87,
    sethostname = 88,
    getdtablesize = 89,
    dup2 = 90,
    fstat = 91,
    fcntl = 92,
    osf_select = 93,
    poll = 94,
    fsync = 95,
    setpriority = 96,
    socket = 97,
    connect = 98,
    accept = 99,
    getpriority = 100,
    send = 101,
    recv = 102,
    sigreturn = 103,
    bind = 104,
    setsockopt = 105,
    listen = 106,
    osf_plock = 107,
    osf_old_sigvec = 108,
    osf_old_sigblock = 109,
    osf_old_sigsetmask = 110,
    sigsuspend = 111,
    osf_sigstack = 112,
    recvmsg = 113,
    sendmsg = 114,
    osf_old_vtrace = 115,
    osf_gettimeofday = 116,
    osf_getrusage = 117,
    getsockopt = 118,
    readv = 120,
    writev = 121,
    osf_settimeofday = 122,
    fchown = 123,
    fchmod = 124,
    recvfrom = 125,
    setreuid = 126,
    setregid = 127,
    rename = 128,
    truncate = 129,
    ftruncate = 130,
    flock = 131,
    setgid = 132,
    sendto = 133,
    shutdown = 134,
    socketpair = 135,
    mkdir = 136,
    rmdir = 137,
    osf_utimes = 138,
    osf_old_sigreturn = 139,
    osf_adjtime = 140,
    getpeername = 141,
    osf_gethostid = 142,
    osf_sethostid = 143,
    getrlimit = 144,
    setrlimit = 145,
    osf_old_killpg = 146,
    setsid = 147,
    quotactl = 148,
    osf_oldquota = 149,
    getsockname = 150,
    osf_pid_block = 153,
    osf_pid_unblock = 154,
    sigaction = 156,
    osf_sigwaitprim = 157,
    osf_nfssvc = 158,
    osf_getdirentries = 159,
    osf_statfs = 160,
    osf_fstatfs = 161,
    osf_asynch_daemon = 163,
    osf_getfh = 164,
    osf_getdomainname = 165,
    setdomainname = 166,
    osf_exportfs = 169,
    osf_alt_plock = 181,
    osf_getmnt = 184,
    osf_alt_sigpending = 187,
    osf_alt_setsid = 188,
    osf_swapon = 199,
    msgctl = 200,
    msgget = 201,
    msgrcv = 202,
    msgsnd = 203,
    semctl = 204,
    semget = 205,
    semop = 206,
    osf_utsname = 207,
    lchown = 208,
    shmat = 209,
    shmctl = 210,
    shmdt = 211,
    shmget = 212,
    osf_mvalid = 213,
    osf_getaddressconf = 214,
    osf_msleep = 215,
    osf_mwakeup = 216,
    msync = 217,
    osf_signal = 218,
    osf_utc_gettime = 219,
    osf_utc_adjtime = 220,
    osf_security = 222,
    osf_kloadcall = 223,
    osf_stat = 224,
    osf_lstat = 225,
    osf_fstat = 226,
    osf_statfs64 = 227,
    osf_fstatfs64 = 228,
    getpgid = 233,
    getsid = 234,
    sigaltstack = 235,
    osf_waitid = 236,
    osf_priocntlset = 237,
    osf_sigsendset = 238,
    osf_set_speculative = 239,
    osf_msfs_syscall = 240,
    osf_sysinfo = 241,
    osf_uadmin = 242,
    osf_fuser = 243,
    osf_proplist_syscall = 244,
    osf_ntp_adjtime = 245,
    osf_ntp_gettime = 246,
    osf_pathconf = 247,
    osf_fpathconf = 248,
    osf_uswitch = 250,
    osf_usleep_thread = 251,
    osf_audcntl = 252,
    osf_audgen = 253,
    sysfs = 254,
    osf_subsys_info = 255,
    osf_getsysinfo = 256,
    osf_setsysinfo = 257,
    osf_afs_syscall = 258,
    osf_swapctl = 259,
    osf_memcntl = 260,
    osf_fdatasync = 261,
    bdflush = 300,
    sethae = 301,
    mount = 302,
    old_adjtimex = 303,
    swapoff = 304,
    getdents = 305,
    create_module = 306,
    init_module = 307,
    delete_module = 308,
    get_kernel_syms = 309,
    syslog = 310,
    reboot = 311,
    clone = 312,
    uselib = 313,
    mlock = 314,
    munlock = 315,
    mlockall = 316,
    munlockall = 317,
    sysinfo = 318,
    sysctl = 319,
    oldumount = 321,
    swapon = 322,
    times = 323,
    personality = 324,
    setfsuid = 325,
    setfsgid = 326,
    ustat = 327,
    statfs = 328,
    fstatfs = 329,
    sched_setparam = 330,
    sched_getparam = 331,
    sched_setscheduler = 332,
    sched_getscheduler = 333,
    sched_yield = 334,
    sched_get_priority_max = 335,
    sched_get_priority_min = 336,
    sched_rr_get_interval = 337,
    afs_syscall = 338,
    uname = 339,
    nanosleep = 340,
    mremap = 341,
    nfsservctl = 342,
    setresuid = 343,
    getresuid = 344,
    pciconfig_read = 345,
    pciconfig_write = 346,
    query_module = 347,
    prctl = 348,
    pread64 = 349,
    pwrite64 = 350,
    rt_sigreturn = 351,
    rt_sigaction = 352,
    rt_sigprocmask = 353,
    rt_sigpending = 354,
    rt_sigtimedwait = 355,
    rt_sigqueueinfo = 356,
    rt_sigsuspend = 357,
    select = 358,
    gettimeofday = 359,
    settimeofday = 360,
    getitimer = 361,
    setitimer = 362,
    utimes = 363,
    getrusage = 364,
    wait4 = 365,
    adjtimex = 366,
    getcwd = 367,
    capget = 368,
    capset = 369,
    sendfile = 370,
    setresgid = 371,
    getresgid = 372,
    dipc = 373,
    pivot_root = 374,
    mincore = 375,
    pciconfig_iobase = 376,
    getdents64 = 377,
    gettid = 378,
    readahead = 379,
    tkill = 381,
    setxattr = 382,
    lsetxattr = 383,
    fsetxattr = 384,
    getxattr = 385,
    lgetxattr = 386,
    fgetxattr = 387,
    listxattr = 388,
    llistxattr = 389,
    flistxattr = 390,
    removexattr = 391,
    lremovexattr = 392,
    fremovexattr = 393,
    futex = 394,
    sched_setaffinity = 395,
    sched_getaffinity = 396,
    tuxcall = 397,
    io_setup = 398,
    io_destroy = 399,
    io_getevents = 400,
    io_submit = 401,
    io_cancel = 402,
    exit_group = 405,
    lookup_dcookie = 406,
    epoll_create = 407,
    epoll_ctl = 408,
    epoll_wait = 409,
    remap_file_pages = 410,
    set_tid_address = 411,
    restart_syscall = 412,
    fadvise64 = 413,
    timer_create = 414,
    timer_settime = 415,
    timer_gettime = 416,
    timer_getoverrun = 417,
    timer_delete = 418,
    clock_settime = 419,
    clock_gettime = 420,
    clock_getres = 421,
    clock_nanosleep = 422,
    semtimedop = 423,
    tgkill = 424,
    stat64 = 425,
    lstat64 = 426,
    fstat64 = 427,
    vserver = 428,
    mbind = 429,
    get_mempolicy = 430,
    set_mempolicy = 431,
    mq_open = 432,
    mq_unlink = 433,
    mq_timedsend = 434,
    mq_timedreceive = 435,
    mq_notify = 436,
    mq_getsetattr = 437,
    waitid = 438,
    add_key = 439,
    request_key = 440,
    keyctl = 441,
    ioprio_set = 442,
    ioprio_get = 443,
    inotify_init = 444,
    inotify_add_watch = 445,
    inotify_rm_watch = 446,
    fdatasync = 447,
    kexec_load = 448,
    migrate_pages = 449,
    openat = 450,
    mkdirat = 451,
    mknodat = 452,
    fchownat = 453,
    futimesat = 454,
    fstatat64 = 455,
    unlinkat = 456,
    renameat = 457,
    linkat = 458,
    symlinkat = 459,
    readlinkat = 460,
    fchmodat = 461,
    faccessat = 462,
    pselect6 = 463,
    ppoll = 464,
    unshare = 465,
    set_robust_list = 466,
    get_robust_list = 467,
    splice = 468,
    sync_file_range = 469,
    tee = 470,
    vmsplice = 471,
    move_pages = 472,
    getcpu = 473,
    epoll_pwait = 474,
    utimensat = 475,
    signalfd = 476,
    timerfd = 477,
    eventfd = 478,
    recvmmsg = 479,
    fallocate = 480,
    timerfd_create = 481,
    timerfd_settime = 482,
    timerfd_gettime = 483,
    signalfd4 = 484,
    eventfd2 = 485,
    epoll_create1 = 486,
    dup3 = 487,
    pipe2 = 488,
    inotify_init1 = 489,
    preadv = 490,
    pwritev = 491,
    rt_tgsigqueueinfo = 492,
    perf_event_open = 493,
    fanotify_init = 494,
    fanotify_mark = 495,
    prlimit64 = 496,
    name_to_handle_at = 497,
    open_by_handle_at = 498,
    clock_adjtime = 499,
    syncfs = 500,
    setns = 501,
    accept4 = 502,
    sendmmsg = 503,
    process_vm_readv = 504,
    process_vm_writev = 505,
    kcmp = 506,
    finit_module = 507,
    sched_setattr = 508,
    sched_getattr = 509,
    renameat2 = 510,
    getrandom = 511,
    memfd_create = 512,
    execveat = 513,
    seccomp = 514,
    bpf = 515,
    userfaultfd = 516,
    membarrier = 517,
    mlock2 = 518,
    copy_file_range = 519,
    preadv2 = 520,
    pwritev2 = 521,
    statx = 522,
    io_pgetevents = 523,
    pkey_mprotect = 524,
    pkey_alloc = 525,
    pkey_free = 526,
    rseq = 527,
    statfs64 = 528,
    fstatfs64 = 529,
    getegid = 530,
    geteuid = 531,
    getppid = 532,
    pidfd_send_signal = 534,
    io_uring_setup = 535,
    io_uring_enter = 536,
    io_uring_register = 537,
    open_tree = 538,
    move_mount = 539,
    fsopen = 540,
    fsconfig = 541,
    fsmount = 542,
    fspick = 543,
    pidfd_open = 544,
    clone3 = 545,
    close_range = 546,
    openat2 = 547,
    pidfd_getfd = 548,
    faccessat2 = 549,
    process_madvise = 550,
    epoll_pwait2 = 551,
    mount_setattr = 552,
    quotactl_fd = 553,
    landlock_create_ruleset = 554,
    landlock_add_rule = 555,
    landlock_restrict_self = 556,
    process_mrelease = 558,
    futex_waitv = 559,
    set_mempolicy_home_node = 560,
    cachestat = 561,
    fchmodat2 = 562,
    map_shadow_stack = 563,
    futex_wake = 564,
    futex_wait = 565,
    futex_requeue = 566,
    statmount = 567,
    listmount = 568,
    lsm_get_self_attr = 569,
    lsm_set_self_attr = 570,
    lsm_list_modules = 571,
    mseal = 572,
    setxattrat = 573,
    getxattrat = 574,
    listxattrat = 575,
    removexattrat = 576,
    open_tree_attr = 577,
    file_getattr = 578,
    file_setattr = 579,
    listns = 580,
    rseq_slice_yield = 581,
};

pub const Arm64 = enum(usize) {
    io_setup = 0,
    io_destroy = 1,
    io_submit = 2,
    io_cancel = 3,
    io_getevents = 4,
    setxattr = 5,
    lsetxattr = 6,
    fsetxattr = 7,
    getxattr = 8,
    lgetxattr = 9,
    fgetxattr = 10,
    listxattr = 11,
    llistxattr = 12,
    flistxattr = 13,
    removexattr = 14,
    lremovexattr = 15,
    fremovexattr = 16,
    getcwd = 17,
    lookup_dcookie = 18,
    eventfd2 = 19,
    epoll_create1 = 20,
    epoll_ctl = 21,
    epoll_pwait = 22,
    dup = 23,
    dup3 = 24,
    fcntl = 25,
    inotify_init1 = 26,
    inotify_add_watch = 27,
    inotify_rm_watch = 28,
    ioctl = 29,
    ioprio_set = 30,
    ioprio_get = 31,
    flock = 32,
    mknodat = 33,
    mkdirat = 34,
    unlinkat = 35,
    symlinkat = 36,
    linkat = 37,
    renameat = 38,
    umount2 = 39,
    mount = 40,
    pivot_root = 41,
    nfsservctl = 42,
    statfs = 43,
    fstatfs = 44,
    truncate = 45,
    ftruncate = 46,
    fallocate = 47,
    faccessat = 48,
    chdir = 49,
    fchdir = 50,
    chroot = 51,
    fchmod = 52,
    fchmodat = 53,
    fchownat = 54,
    fchown = 55,
    openat = 56,
    close = 57,
    vhangup = 58,
    pipe2 = 59,
    quotactl = 60,
    getdents64 = 61,
    lseek = 62,
    read = 63,
    write = 64,
    readv = 65,
    writev = 66,
    pread64 = 67,
    pwrite64 = 68,
    preadv = 69,
    pwritev = 70,
    sendfile = 71,
    pselect6 = 72,
    ppoll = 73,
    signalfd4 = 74,
    vmsplice = 75,
    splice = 76,
    tee = 77,
    readlinkat = 78,
    newfstatat = 79,
    fstat = 80,
    sync = 81,
    fsync = 82,
    fdatasync = 83,
    sync_file_range = 84,
    timerfd_create = 85,
    timerfd_settime = 86,
    timerfd_gettime = 87,
    utimensat = 88,
    acct = 89,
    capget = 90,
    capset = 91,
    personality = 92,
    exit = 93,
    exit_group = 94,
    waitid = 95,
    set_tid_address = 96,
    unshare = 97,
    futex = 98,
    set_robust_list = 99,
    get_robust_list = 100,
    nanosleep = 101,
    getitimer = 102,
    setitimer = 103,
    kexec_load = 104,
    init_module = 105,
    delete_module = 106,
    timer_create = 107,
    timer_gettime = 108,
    timer_getoverrun = 109,
    timer_settime = 110,
    timer_delete = 111,
    clock_settime = 112,
    clock_gettime = 113,
    clock_getres = 114,
    clock_nanosleep = 115,
    syslog = 116,
    ptrace = 117,
    sched_setparam = 118,
    sched_setscheduler = 119,
    sched_getscheduler = 120,
    sched_getparam = 121,
    sched_setaffinity = 122,
    sched_getaffinity = 123,
    sched_yield = 124,
    sched_get_priority_max = 125,
    sched_get_priority_min = 126,
    sched_rr_get_interval = 127,
    restart_syscall = 128,
    kill = 129,
    tkill = 130,
    tgkill = 131,
    sigaltstack = 132,
    rt_sigsuspend = 133,
    rt_sigaction = 134,
    rt_sigprocmask = 135,
    rt_sigpending = 136,
    rt_sigtimedwait = 137,
    rt_sigqueueinfo = 138,
    rt_sigreturn = 139,
    setpriority = 140,
    getpriority = 141,
    reboot = 142,
    setregid = 143,
    setgid = 144,
    setreuid = 145,
    setuid = 146,
    setresuid = 147,
    getresuid = 148,
    setresgid = 149,
    getresgid = 150,
    setfsuid = 151,
    setfsgid = 152,
    times = 153,
    setpgid = 154,
    getpgid = 155,
    getsid = 156,
    setsid = 157,
    getgroups = 158,
    setgroups = 159,
    uname = 160,
    sethostname = 161,
    setdomainname = 162,
    getrlimit = 163,
    setrlimit = 164,
    getrusage = 165,
    umask = 166,
    prctl = 167,
    getcpu = 168,
    gettimeofday = 169,
    settimeofday = 170,
    adjtimex = 171,
    getpid = 172,
    getppid = 173,
    getuid = 174,
    geteuid = 175,
    getgid = 176,
    getegid = 177,
    gettid = 178,
    sysinfo = 179,
    mq_open = 180,
    mq_unlink = 181,
    mq_timedsend = 182,
    mq_timedreceive = 183,
    mq_notify = 184,
    mq_getsetattr = 185,
    msgget = 186,
    msgctl = 187,
    msgrcv = 188,
    msgsnd = 189,
    semget = 190,
    semctl = 191,
    semtimedop = 192,
    semop = 193,
    shmget = 194,
    shmctl = 195,
    shmat = 196,
    shmdt = 197,
    socket = 198,
    socketpair = 199,
    bind = 200,
    listen = 201,
    accept = 202,
    connect = 203,
    getsockname = 204,
    getpeername = 205,
    sendto = 206,
    recvfrom = 207,
    setsockopt = 208,
    getsockopt = 209,
    shutdown = 210,
    sendmsg = 211,
    recvmsg = 212,
    readahead = 213,
    brk = 214,
    munmap = 215,
    mremap = 216,
    add_key = 217,
    request_key = 218,
    keyctl = 219,
    clone = 220,
    execve = 221,
    mmap = 222,
    fadvise64 = 223,
    swapon = 224,
    swapoff = 225,
    mprotect = 226,
    msync = 227,
    mlock = 228,
    munlock = 229,
    mlockall = 230,
    munlockall = 231,
    mincore = 232,
    madvise = 233,
    remap_file_pages = 234,
    mbind = 235,
    get_mempolicy = 236,
    set_mempolicy = 237,
    migrate_pages = 238,
    move_pages = 239,
    rt_tgsigqueueinfo = 240,
    perf_event_open = 241,
    accept4 = 242,
    recvmmsg = 243,
    wait4 = 260,
    prlimit64 = 261,
    fanotify_init = 262,
    fanotify_mark = 263,
    name_to_handle_at = 264,
    open_by_handle_at = 265,
    clock_adjtime = 266,
    syncfs = 267,
    setns = 268,
    sendmmsg = 269,
    process_vm_readv = 270,
    process_vm_writev = 271,
    kcmp = 272,
    finit_module = 273,
    sched_setattr = 274,
    sched_getattr = 275,
    renameat2 = 276,
    seccomp = 277,
    getrandom = 278,
    memfd_create = 279,
    bpf = 280,
    execveat = 281,
    userfaultfd = 282,
    membarrier = 283,
    mlock2 = 284,
    copy_file_range = 285,
    preadv2 = 286,
    pwritev2 = 287,
    pkey_mprotect = 288,
    pkey_alloc = 289,
    pkey_free = 290,
    statx = 291,
    io_pgetevents = 292,
    rseq = 293,
    kexec_file_load = 294,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    memfd_secret = 447,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const RiscV32 = enum(usize) {
    io_setup = 0,
    io_destroy = 1,
    io_submit = 2,
    io_cancel = 3,
    setxattr = 5,
    lsetxattr = 6,
    fsetxattr = 7,
    getxattr = 8,
    lgetxattr = 9,
    fgetxattr = 10,
    listxattr = 11,
    llistxattr = 12,
    flistxattr = 13,
    removexattr = 14,
    lremovexattr = 15,
    fremovexattr = 16,
    getcwd = 17,
    lookup_dcookie = 18,
    eventfd2 = 19,
    epoll_create1 = 20,
    epoll_ctl = 21,
    epoll_pwait = 22,
    dup = 23,
    dup3 = 24,
    fcntl64 = 25,
    inotify_init1 = 26,
    inotify_add_watch = 27,
    inotify_rm_watch = 28,
    ioctl = 29,
    ioprio_set = 30,
    ioprio_get = 31,
    flock = 32,
    mknodat = 33,
    mkdirat = 34,
    unlinkat = 35,
    symlinkat = 36,
    linkat = 37,
    umount2 = 39,
    mount = 40,
    pivot_root = 41,
    nfsservctl = 42,
    statfs64 = 43,
    fstatfs64 = 44,
    truncate64 = 45,
    ftruncate64 = 46,
    fallocate = 47,
    faccessat = 48,
    chdir = 49,
    fchdir = 50,
    chroot = 51,
    fchmod = 52,
    fchmodat = 53,
    fchownat = 54,
    fchown = 55,
    openat = 56,
    close = 57,
    vhangup = 58,
    pipe2 = 59,
    quotactl = 60,
    getdents64 = 61,
    llseek = 62,
    read = 63,
    write = 64,
    readv = 65,
    writev = 66,
    pread64 = 67,
    pwrite64 = 68,
    preadv = 69,
    pwritev = 70,
    sendfile64 = 71,
    signalfd4 = 74,
    vmsplice = 75,
    splice = 76,
    tee = 77,
    readlinkat = 78,
    sync = 81,
    fsync = 82,
    fdatasync = 83,
    sync_file_range = 84,
    timerfd_create = 85,
    acct = 89,
    capget = 90,
    capset = 91,
    personality = 92,
    exit = 93,
    exit_group = 94,
    waitid = 95,
    set_tid_address = 96,
    unshare = 97,
    set_robust_list = 99,
    get_robust_list = 100,
    getitimer = 102,
    setitimer = 103,
    kexec_load = 104,
    init_module = 105,
    delete_module = 106,
    timer_create = 107,
    timer_getoverrun = 109,
    timer_delete = 111,
    syslog = 116,
    ptrace = 117,
    sched_setparam = 118,
    sched_setscheduler = 119,
    sched_getscheduler = 120,
    sched_getparam = 121,
    sched_setaffinity = 122,
    sched_getaffinity = 123,
    sched_yield = 124,
    sched_get_priority_max = 125,
    sched_get_priority_min = 126,
    restart_syscall = 128,
    kill = 129,
    tkill = 130,
    tgkill = 131,
    sigaltstack = 132,
    rt_sigsuspend = 133,
    rt_sigaction = 134,
    rt_sigprocmask = 135,
    rt_sigpending = 136,
    rt_sigqueueinfo = 138,
    rt_sigreturn = 139,
    setpriority = 140,
    getpriority = 141,
    reboot = 142,
    setregid = 143,
    setgid = 144,
    setreuid = 145,
    setuid = 146,
    setresuid = 147,
    getresuid = 148,
    setresgid = 149,
    getresgid = 150,
    setfsuid = 151,
    setfsgid = 152,
    times = 153,
    setpgid = 154,
    getpgid = 155,
    getsid = 156,
    setsid = 157,
    getgroups = 158,
    setgroups = 159,
    uname = 160,
    sethostname = 161,
    setdomainname = 162,
    getrusage = 165,
    umask = 166,
    prctl = 167,
    getcpu = 168,
    getpid = 172,
    getppid = 173,
    getuid = 174,
    geteuid = 175,
    getgid = 176,
    getegid = 177,
    gettid = 178,
    sysinfo = 179,
    mq_open = 180,
    mq_unlink = 181,
    mq_notify = 184,
    mq_getsetattr = 185,
    msgget = 186,
    msgctl = 187,
    msgrcv = 188,
    msgsnd = 189,
    semget = 190,
    semctl = 191,
    semop = 193,
    shmget = 194,
    shmctl = 195,
    shmat = 196,
    shmdt = 197,
    socket = 198,
    socketpair = 199,
    bind = 200,
    listen = 201,
    accept = 202,
    connect = 203,
    getsockname = 204,
    getpeername = 205,
    sendto = 206,
    recvfrom = 207,
    setsockopt = 208,
    getsockopt = 209,
    shutdown = 210,
    sendmsg = 211,
    recvmsg = 212,
    readahead = 213,
    brk = 214,
    munmap = 215,
    mremap = 216,
    add_key = 217,
    request_key = 218,
    keyctl = 219,
    clone = 220,
    execve = 221,
    mmap2 = 222,
    fadvise64_64 = 223,
    swapon = 224,
    swapoff = 225,
    mprotect = 226,
    msync = 227,
    mlock = 228,
    munlock = 229,
    mlockall = 230,
    munlockall = 231,
    mincore = 232,
    madvise = 233,
    remap_file_pages = 234,
    mbind = 235,
    get_mempolicy = 236,
    set_mempolicy = 237,
    migrate_pages = 238,
    move_pages = 239,
    rt_tgsigqueueinfo = 240,
    perf_event_open = 241,
    accept4 = 242,
    riscv_hwprobe = 258,
    riscv_flush_icache = 259,
    prlimit64 = 261,
    fanotify_init = 262,
    fanotify_mark = 263,
    name_to_handle_at = 264,
    open_by_handle_at = 265,
    syncfs = 267,
    setns = 268,
    sendmmsg = 269,
    process_vm_readv = 270,
    process_vm_writev = 271,
    kcmp = 272,
    finit_module = 273,
    sched_setattr = 274,
    sched_getattr = 275,
    renameat2 = 276,
    seccomp = 277,
    getrandom = 278,
    memfd_create = 279,
    bpf = 280,
    execveat = 281,
    userfaultfd = 282,
    membarrier = 283,
    mlock2 = 284,
    copy_file_range = 285,
    preadv2 = 286,
    pwritev2 = 287,
    pkey_mprotect = 288,
    pkey_alloc = 289,
    pkey_free = 290,
    statx = 291,
    rseq = 293,
    kexec_file_load = 294,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    memfd_secret = 447,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const RiscV64 = enum(usize) {
    io_setup = 0,
    io_destroy = 1,
    io_submit = 2,
    io_cancel = 3,
    io_getevents = 4,
    setxattr = 5,
    lsetxattr = 6,
    fsetxattr = 7,
    getxattr = 8,
    lgetxattr = 9,
    fgetxattr = 10,
    listxattr = 11,
    llistxattr = 12,
    flistxattr = 13,
    removexattr = 14,
    lremovexattr = 15,
    fremovexattr = 16,
    getcwd = 17,
    lookup_dcookie = 18,
    eventfd2 = 19,
    epoll_create1 = 20,
    epoll_ctl = 21,
    epoll_pwait = 22,
    dup = 23,
    dup3 = 24,
    fcntl = 25,
    inotify_init1 = 26,
    inotify_add_watch = 27,
    inotify_rm_watch = 28,
    ioctl = 29,
    ioprio_set = 30,
    ioprio_get = 31,
    flock = 32,
    mknodat = 33,
    mkdirat = 34,
    unlinkat = 35,
    symlinkat = 36,
    linkat = 37,
    umount2 = 39,
    mount = 40,
    pivot_root = 41,
    nfsservctl = 42,
    statfs = 43,
    fstatfs = 44,
    truncate = 45,
    ftruncate = 46,
    fallocate = 47,
    faccessat = 48,
    chdir = 49,
    fchdir = 50,
    chroot = 51,
    fchmod = 52,
    fchmodat = 53,
    fchownat = 54,
    fchown = 55,
    openat = 56,
    close = 57,
    vhangup = 58,
    pipe2 = 59,
    quotactl = 60,
    getdents64 = 61,
    lseek = 62,
    read = 63,
    write = 64,
    readv = 65,
    writev = 66,
    pread64 = 67,
    pwrite64 = 68,
    preadv = 69,
    pwritev = 70,
    sendfile = 71,
    pselect6 = 72,
    ppoll = 73,
    signalfd4 = 74,
    vmsplice = 75,
    splice = 76,
    tee = 77,
    readlinkat = 78,
    newfstatat = 79,
    fstat = 80,
    sync = 81,
    fsync = 82,
    fdatasync = 83,
    sync_file_range = 84,
    timerfd_create = 85,
    timerfd_settime = 86,
    timerfd_gettime = 87,
    utimensat = 88,
    acct = 89,
    capget = 90,
    capset = 91,
    personality = 92,
    exit = 93,
    exit_group = 94,
    waitid = 95,
    set_tid_address = 96,
    unshare = 97,
    futex = 98,
    set_robust_list = 99,
    get_robust_list = 100,
    nanosleep = 101,
    getitimer = 102,
    setitimer = 103,
    kexec_load = 104,
    init_module = 105,
    delete_module = 106,
    timer_create = 107,
    timer_gettime = 108,
    timer_getoverrun = 109,
    timer_settime = 110,
    timer_delete = 111,
    clock_settime = 112,
    clock_gettime = 113,
    clock_getres = 114,
    clock_nanosleep = 115,
    syslog = 116,
    ptrace = 117,
    sched_setparam = 118,
    sched_setscheduler = 119,
    sched_getscheduler = 120,
    sched_getparam = 121,
    sched_setaffinity = 122,
    sched_getaffinity = 123,
    sched_yield = 124,
    sched_get_priority_max = 125,
    sched_get_priority_min = 126,
    sched_rr_get_interval = 127,
    restart_syscall = 128,
    kill = 129,
    tkill = 130,
    tgkill = 131,
    sigaltstack = 132,
    rt_sigsuspend = 133,
    rt_sigaction = 134,
    rt_sigprocmask = 135,
    rt_sigpending = 136,
    rt_sigtimedwait = 137,
    rt_sigqueueinfo = 138,
    rt_sigreturn = 139,
    setpriority = 140,
    getpriority = 141,
    reboot = 142,
    setregid = 143,
    setgid = 144,
    setreuid = 145,
    setuid = 146,
    setresuid = 147,
    getresuid = 148,
    setresgid = 149,
    getresgid = 150,
    setfsuid = 151,
    setfsgid = 152,
    times = 153,
    setpgid = 154,
    getpgid = 155,
    getsid = 156,
    setsid = 157,
    getgroups = 158,
    setgroups = 159,
    uname = 160,
    sethostname = 161,
    setdomainname = 162,
    getrlimit = 163,
    setrlimit = 164,
    getrusage = 165,
    umask = 166,
    prctl = 167,
    getcpu = 168,
    gettimeofday = 169,
    settimeofday = 170,
    adjtimex = 171,
    getpid = 172,
    getppid = 173,
    getuid = 174,
    geteuid = 175,
    getgid = 176,
    getegid = 177,
    gettid = 178,
    sysinfo = 179,
    mq_open = 180,
    mq_unlink = 181,
    mq_timedsend = 182,
    mq_timedreceive = 183,
    mq_notify = 184,
    mq_getsetattr = 185,
    msgget = 186,
    msgctl = 187,
    msgrcv = 188,
    msgsnd = 189,
    semget = 190,
    semctl = 191,
    semtimedop = 192,
    semop = 193,
    shmget = 194,
    shmctl = 195,
    shmat = 196,
    shmdt = 197,
    socket = 198,
    socketpair = 199,
    bind = 200,
    listen = 201,
    accept = 202,
    connect = 203,
    getsockname = 204,
    getpeername = 205,
    sendto = 206,
    recvfrom = 207,
    setsockopt = 208,
    getsockopt = 209,
    shutdown = 210,
    sendmsg = 211,
    recvmsg = 212,
    readahead = 213,
    brk = 214,
    munmap = 215,
    mremap = 216,
    add_key = 217,
    request_key = 218,
    keyctl = 219,
    clone = 220,
    execve = 221,
    mmap = 222,
    fadvise64 = 223,
    swapon = 224,
    swapoff = 225,
    mprotect = 226,
    msync = 227,
    mlock = 228,
    munlock = 229,
    mlockall = 230,
    munlockall = 231,
    mincore = 232,
    madvise = 233,
    remap_file_pages = 234,
    mbind = 235,
    get_mempolicy = 236,
    set_mempolicy = 237,
    migrate_pages = 238,
    move_pages = 239,
    rt_tgsigqueueinfo = 240,
    perf_event_open = 241,
    accept4 = 242,
    recvmmsg = 243,
    riscv_hwprobe = 258,
    riscv_flush_icache = 259,
    wait4 = 260,
    prlimit64 = 261,
    fanotify_init = 262,
    fanotify_mark = 263,
    name_to_handle_at = 264,
    open_by_handle_at = 265,
    clock_adjtime = 266,
    syncfs = 267,
    setns = 268,
    sendmmsg = 269,
    process_vm_readv = 270,
    process_vm_writev = 271,
    kcmp = 272,
    finit_module = 273,
    sched_setattr = 274,
    sched_getattr = 275,
    renameat2 = 276,
    seccomp = 277,
    getrandom = 278,
    memfd_create = 279,
    bpf = 280,
    execveat = 281,
    userfaultfd = 282,
    membarrier = 283,
    mlock2 = 284,
    copy_file_range = 285,
    preadv2 = 286,
    pwritev2 = 287,
    pkey_mprotect = 288,
    pkey_alloc = 289,
    pkey_free = 290,
    statx = 291,
    io_pgetevents = 292,
    rseq = 293,
    kexec_file_load = 294,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    memfd_secret = 447,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const LoongArch32 = enum(usize) {
    io_setup = 0,
    io_destroy = 1,
    io_submit = 2,
    io_cancel = 3,
    setxattr = 5,
    lsetxattr = 6,
    fsetxattr = 7,
    getxattr = 8,
    lgetxattr = 9,
    fgetxattr = 10,
    listxattr = 11,
    llistxattr = 12,
    flistxattr = 13,
    removexattr = 14,
    lremovexattr = 15,
    fremovexattr = 16,
    getcwd = 17,
    lookup_dcookie = 18,
    eventfd2 = 19,
    epoll_create1 = 20,
    epoll_ctl = 21,
    epoll_pwait = 22,
    dup = 23,
    dup3 = 24,
    fcntl64 = 25,
    inotify_init1 = 26,
    inotify_add_watch = 27,
    inotify_rm_watch = 28,
    ioctl = 29,
    ioprio_set = 30,
    ioprio_get = 31,
    flock = 32,
    mknodat = 33,
    mkdirat = 34,
    unlinkat = 35,
    symlinkat = 36,
    linkat = 37,
    umount2 = 39,
    mount = 40,
    pivot_root = 41,
    nfsservctl = 42,
    statfs64 = 43,
    fstatfs64 = 44,
    truncate64 = 45,
    ftruncate64 = 46,
    fallocate = 47,
    faccessat = 48,
    chdir = 49,
    fchdir = 50,
    chroot = 51,
    fchmod = 52,
    fchmodat = 53,
    fchownat = 54,
    fchown = 55,
    openat = 56,
    close = 57,
    vhangup = 58,
    pipe2 = 59,
    quotactl = 60,
    getdents64 = 61,
    llseek = 62,
    read = 63,
    write = 64,
    readv = 65,
    writev = 66,
    pread64 = 67,
    pwrite64 = 68,
    preadv = 69,
    pwritev = 70,
    sendfile64 = 71,
    signalfd4 = 74,
    vmsplice = 75,
    splice = 76,
    tee = 77,
    readlinkat = 78,
    sync = 81,
    fsync = 82,
    fdatasync = 83,
    sync_file_range = 84,
    timerfd_create = 85,
    acct = 89,
    capget = 90,
    capset = 91,
    personality = 92,
    exit = 93,
    exit_group = 94,
    waitid = 95,
    set_tid_address = 96,
    unshare = 97,
    set_robust_list = 99,
    get_robust_list = 100,
    getitimer = 102,
    setitimer = 103,
    kexec_load = 104,
    init_module = 105,
    delete_module = 106,
    timer_create = 107,
    timer_getoverrun = 109,
    timer_delete = 111,
    syslog = 116,
    ptrace = 117,
    sched_setparam = 118,
    sched_setscheduler = 119,
    sched_getscheduler = 120,
    sched_getparam = 121,
    sched_setaffinity = 122,
    sched_getaffinity = 123,
    sched_yield = 124,
    sched_get_priority_max = 125,
    sched_get_priority_min = 126,
    restart_syscall = 128,
    kill = 129,
    tkill = 130,
    tgkill = 131,
    sigaltstack = 132,
    rt_sigsuspend = 133,
    rt_sigaction = 134,
    rt_sigprocmask = 135,
    rt_sigpending = 136,
    rt_sigqueueinfo = 138,
    rt_sigreturn = 139,
    setpriority = 140,
    getpriority = 141,
    reboot = 142,
    setregid = 143,
    setgid = 144,
    setreuid = 145,
    setuid = 146,
    setresuid = 147,
    getresuid = 148,
    setresgid = 149,
    getresgid = 150,
    setfsuid = 151,
    setfsgid = 152,
    times = 153,
    setpgid = 154,
    getpgid = 155,
    getsid = 156,
    setsid = 157,
    getgroups = 158,
    setgroups = 159,
    uname = 160,
    sethostname = 161,
    setdomainname = 162,
    getrusage = 165,
    umask = 166,
    prctl = 167,
    getcpu = 168,
    getpid = 172,
    getppid = 173,
    getuid = 174,
    geteuid = 175,
    getgid = 176,
    getegid = 177,
    gettid = 178,
    sysinfo = 179,
    mq_open = 180,
    mq_unlink = 181,
    mq_notify = 184,
    mq_getsetattr = 185,
    msgget = 186,
    msgctl = 187,
    msgrcv = 188,
    msgsnd = 189,
    semget = 190,
    semctl = 191,
    semop = 193,
    shmget = 194,
    shmctl = 195,
    shmat = 196,
    shmdt = 197,
    socket = 198,
    socketpair = 199,
    bind = 200,
    listen = 201,
    accept = 202,
    connect = 203,
    getsockname = 204,
    getpeername = 205,
    sendto = 206,
    recvfrom = 207,
    setsockopt = 208,
    getsockopt = 209,
    shutdown = 210,
    sendmsg = 211,
    recvmsg = 212,
    readahead = 213,
    brk = 214,
    munmap = 215,
    mremap = 216,
    add_key = 217,
    request_key = 218,
    keyctl = 219,
    clone = 220,
    execve = 221,
    mmap2 = 222,
    fadvise64_64 = 223,
    swapon = 224,
    swapoff = 225,
    mprotect = 226,
    msync = 227,
    mlock = 228,
    munlock = 229,
    mlockall = 230,
    munlockall = 231,
    mincore = 232,
    madvise = 233,
    remap_file_pages = 234,
    mbind = 235,
    get_mempolicy = 236,
    set_mempolicy = 237,
    migrate_pages = 238,
    move_pages = 239,
    rt_tgsigqueueinfo = 240,
    perf_event_open = 241,
    accept4 = 242,
    prlimit64 = 261,
    fanotify_init = 262,
    fanotify_mark = 263,
    name_to_handle_at = 264,
    open_by_handle_at = 265,
    syncfs = 267,
    setns = 268,
    sendmmsg = 269,
    process_vm_readv = 270,
    process_vm_writev = 271,
    kcmp = 272,
    finit_module = 273,
    sched_setattr = 274,
    sched_getattr = 275,
    renameat2 = 276,
    seccomp = 277,
    getrandom = 278,
    memfd_create = 279,
    bpf = 280,
    execveat = 281,
    userfaultfd = 282,
    membarrier = 283,
    mlock2 = 284,
    copy_file_range = 285,
    preadv2 = 286,
    pwritev2 = 287,
    pkey_mprotect = 288,
    pkey_alloc = 289,
    pkey_free = 290,
    statx = 291,
    rseq = 293,
    kexec_file_load = 294,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const LoongArch64 = enum(usize) {
    io_setup = 0,
    io_destroy = 1,
    io_submit = 2,
    io_cancel = 3,
    io_getevents = 4,
    setxattr = 5,
    lsetxattr = 6,
    fsetxattr = 7,
    getxattr = 8,
    lgetxattr = 9,
    fgetxattr = 10,
    listxattr = 11,
    llistxattr = 12,
    flistxattr = 13,
    removexattr = 14,
    lremovexattr = 15,
    fremovexattr = 16,
    getcwd = 17,
    lookup_dcookie = 18,
    eventfd2 = 19,
    epoll_create1 = 20,
    epoll_ctl = 21,
    epoll_pwait = 22,
    dup = 23,
    dup3 = 24,
    fcntl = 25,
    inotify_init1 = 26,
    inotify_add_watch = 27,
    inotify_rm_watch = 28,
    ioctl = 29,
    ioprio_set = 30,
    ioprio_get = 31,
    flock = 32,
    mknodat = 33,
    mkdirat = 34,
    unlinkat = 35,
    symlinkat = 36,
    linkat = 37,
    umount2 = 39,
    mount = 40,
    pivot_root = 41,
    nfsservctl = 42,
    statfs = 43,
    fstatfs = 44,
    truncate = 45,
    ftruncate = 46,
    fallocate = 47,
    faccessat = 48,
    chdir = 49,
    fchdir = 50,
    chroot = 51,
    fchmod = 52,
    fchmodat = 53,
    fchownat = 54,
    fchown = 55,
    openat = 56,
    close = 57,
    vhangup = 58,
    pipe2 = 59,
    quotactl = 60,
    getdents64 = 61,
    lseek = 62,
    read = 63,
    write = 64,
    readv = 65,
    writev = 66,
    pread64 = 67,
    pwrite64 = 68,
    preadv = 69,
    pwritev = 70,
    sendfile = 71,
    pselect6 = 72,
    ppoll = 73,
    signalfd4 = 74,
    vmsplice = 75,
    splice = 76,
    tee = 77,
    readlinkat = 78,
    newfstatat = 79,
    fstat = 80,
    sync = 81,
    fsync = 82,
    fdatasync = 83,
    sync_file_range = 84,
    timerfd_create = 85,
    timerfd_settime = 86,
    timerfd_gettime = 87,
    utimensat = 88,
    acct = 89,
    capget = 90,
    capset = 91,
    personality = 92,
    exit = 93,
    exit_group = 94,
    waitid = 95,
    set_tid_address = 96,
    unshare = 97,
    futex = 98,
    set_robust_list = 99,
    get_robust_list = 100,
    nanosleep = 101,
    getitimer = 102,
    setitimer = 103,
    kexec_load = 104,
    init_module = 105,
    delete_module = 106,
    timer_create = 107,
    timer_gettime = 108,
    timer_getoverrun = 109,
    timer_settime = 110,
    timer_delete = 111,
    clock_settime = 112,
    clock_gettime = 113,
    clock_getres = 114,
    clock_nanosleep = 115,
    syslog = 116,
    ptrace = 117,
    sched_setparam = 118,
    sched_setscheduler = 119,
    sched_getscheduler = 120,
    sched_getparam = 121,
    sched_setaffinity = 122,
    sched_getaffinity = 123,
    sched_yield = 124,
    sched_get_priority_max = 125,
    sched_get_priority_min = 126,
    sched_rr_get_interval = 127,
    restart_syscall = 128,
    kill = 129,
    tkill = 130,
    tgkill = 131,
    sigaltstack = 132,
    rt_sigsuspend = 133,
    rt_sigaction = 134,
    rt_sigprocmask = 135,
    rt_sigpending = 136,
    rt_sigtimedwait = 137,
    rt_sigqueueinfo = 138,
    rt_sigreturn = 139,
    setpriority = 140,
    getpriority = 141,
    reboot = 142,
    setregid = 143,
    setgid = 144,
    setreuid = 145,
    setuid = 146,
    setresuid = 147,
    getresuid = 148,
    setresgid = 149,
    getresgid = 150,
    setfsuid = 151,
    setfsgid = 152,
    times = 153,
    setpgid = 154,
    getpgid = 155,
    getsid = 156,
    setsid = 157,
    getgroups = 158,
    setgroups = 159,
    uname = 160,
    sethostname = 161,
    setdomainname = 162,
    getrusage = 165,
    umask = 166,
    prctl = 167,
    getcpu = 168,
    gettimeofday = 169,
    settimeofday = 170,
    adjtimex = 171,
    getpid = 172,
    getppid = 173,
    getuid = 174,
    geteuid = 175,
    getgid = 176,
    getegid = 177,
    gettid = 178,
    sysinfo = 179,
    mq_open = 180,
    mq_unlink = 181,
    mq_timedsend = 182,
    mq_timedreceive = 183,
    mq_notify = 184,
    mq_getsetattr = 185,
    msgget = 186,
    msgctl = 187,
    msgrcv = 188,
    msgsnd = 189,
    semget = 190,
    semctl = 191,
    semtimedop = 192,
    semop = 193,
    shmget = 194,
    shmctl = 195,
    shmat = 196,
    shmdt = 197,
    socket = 198,
    socketpair = 199,
    bind = 200,
    listen = 201,
    accept = 202,
    connect = 203,
    getsockname = 204,
    getpeername = 205,
    sendto = 206,
    recvfrom = 207,
    setsockopt = 208,
    getsockopt = 209,
    shutdown = 210,
    sendmsg = 211,
    recvmsg = 212,
    readahead = 213,
    brk = 214,
    munmap = 215,
    mremap = 216,
    add_key = 217,
    request_key = 218,
    keyctl = 219,
    clone = 220,
    execve = 221,
    mmap = 222,
    fadvise64 = 223,
    swapon = 224,
    swapoff = 225,
    mprotect = 226,
    msync = 227,
    mlock = 228,
    munlock = 229,
    mlockall = 230,
    munlockall = 231,
    mincore = 232,
    madvise = 233,
    remap_file_pages = 234,
    mbind = 235,
    get_mempolicy = 236,
    set_mempolicy = 237,
    migrate_pages = 238,
    move_pages = 239,
    rt_tgsigqueueinfo = 240,
    perf_event_open = 241,
    accept4 = 242,
    recvmmsg = 243,
    wait4 = 260,
    prlimit64 = 261,
    fanotify_init = 262,
    fanotify_mark = 263,
    name_to_handle_at = 264,
    open_by_handle_at = 265,
    clock_adjtime = 266,
    syncfs = 267,
    setns = 268,
    sendmmsg = 269,
    process_vm_readv = 270,
    process_vm_writev = 271,
    kcmp = 272,
    finit_module = 273,
    sched_setattr = 274,
    sched_getattr = 275,
    renameat2 = 276,
    seccomp = 277,
    getrandom = 278,
    memfd_create = 279,
    bpf = 280,
    execveat = 281,
    userfaultfd = 282,
    membarrier = 283,
    mlock2 = 284,
    copy_file_range = 285,
    preadv2 = 286,
    pwritev2 = 287,
    pkey_mprotect = 288,
    pkey_alloc = 289,
    pkey_free = 290,
    statx = 291,
    io_pgetevents = 292,
    rseq = 293,
    kexec_file_load = 294,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const Arc = enum(usize) {
    io_setup = 0,
    io_destroy = 1,
    io_submit = 2,
    io_cancel = 3,
    io_getevents = 4,
    setxattr = 5,
    lsetxattr = 6,
    fsetxattr = 7,
    getxattr = 8,
    lgetxattr = 9,
    fgetxattr = 10,
    listxattr = 11,
    llistxattr = 12,
    flistxattr = 13,
    removexattr = 14,
    lremovexattr = 15,
    fremovexattr = 16,
    getcwd = 17,
    lookup_dcookie = 18,
    eventfd2 = 19,
    epoll_create1 = 20,
    epoll_ctl = 21,
    epoll_pwait = 22,
    dup = 23,
    dup3 = 24,
    fcntl64 = 25,
    inotify_init1 = 26,
    inotify_add_watch = 27,
    inotify_rm_watch = 28,
    ioctl = 29,
    ioprio_set = 30,
    ioprio_get = 31,
    flock = 32,
    mknodat = 33,
    mkdirat = 34,
    unlinkat = 35,
    symlinkat = 36,
    linkat = 37,
    renameat = 38,
    umount2 = 39,
    mount = 40,
    pivot_root = 41,
    nfsservctl = 42,
    statfs64 = 43,
    fstatfs64 = 44,
    truncate64 = 45,
    ftruncate64 = 46,
    fallocate = 47,
    faccessat = 48,
    chdir = 49,
    fchdir = 50,
    chroot = 51,
    fchmod = 52,
    fchmodat = 53,
    fchownat = 54,
    fchown = 55,
    openat = 56,
    close = 57,
    vhangup = 58,
    pipe2 = 59,
    quotactl = 60,
    getdents64 = 61,
    llseek = 62,
    read = 63,
    write = 64,
    readv = 65,
    writev = 66,
    pread64 = 67,
    pwrite64 = 68,
    preadv = 69,
    pwritev = 70,
    sendfile64 = 71,
    pselect6 = 72,
    ppoll = 73,
    signalfd4 = 74,
    vmsplice = 75,
    splice = 76,
    tee = 77,
    readlinkat = 78,
    fstatat64 = 79,
    fstat64 = 80,
    sync = 81,
    fsync = 82,
    fdatasync = 83,
    sync_file_range = 84,
    timerfd_create = 85,
    timerfd_settime = 86,
    timerfd_gettime = 87,
    utimensat = 88,
    acct = 89,
    capget = 90,
    capset = 91,
    personality = 92,
    exit = 93,
    exit_group = 94,
    waitid = 95,
    set_tid_address = 96,
    unshare = 97,
    futex = 98,
    set_robust_list = 99,
    get_robust_list = 100,
    nanosleep = 101,
    getitimer = 102,
    setitimer = 103,
    kexec_load = 104,
    init_module = 105,
    delete_module = 106,
    timer_create = 107,
    timer_gettime = 108,
    timer_getoverrun = 109,
    timer_settime = 110,
    timer_delete = 111,
    clock_settime = 112,
    clock_gettime = 113,
    clock_getres = 114,
    clock_nanosleep = 115,
    syslog = 116,
    ptrace = 117,
    sched_setparam = 118,
    sched_setscheduler = 119,
    sched_getscheduler = 120,
    sched_getparam = 121,
    sched_setaffinity = 122,
    sched_getaffinity = 123,
    sched_yield = 124,
    sched_get_priority_max = 125,
    sched_get_priority_min = 126,
    sched_rr_get_interval = 127,
    restart_syscall = 128,
    kill = 129,
    tkill = 130,
    tgkill = 131,
    sigaltstack = 132,
    rt_sigsuspend = 133,
    rt_sigaction = 134,
    rt_sigprocmask = 135,
    rt_sigpending = 136,
    rt_sigtimedwait = 137,
    rt_sigqueueinfo = 138,
    rt_sigreturn = 139,
    setpriority = 140,
    getpriority = 141,
    reboot = 142,
    setregid = 143,
    setgid = 144,
    setreuid = 145,
    setuid = 146,
    setresuid = 147,
    getresuid = 148,
    setresgid = 149,
    getresgid = 150,
    setfsuid = 151,
    setfsgid = 152,
    times = 153,
    setpgid = 154,
    getpgid = 155,
    getsid = 156,
    setsid = 157,
    getgroups = 158,
    setgroups = 159,
    uname = 160,
    sethostname = 161,
    setdomainname = 162,
    getrlimit = 163,
    setrlimit = 164,
    getrusage = 165,
    umask = 166,
    prctl = 167,
    getcpu = 168,
    gettimeofday = 169,
    settimeofday = 170,
    adjtimex = 171,
    getpid = 172,
    getppid = 173,
    getuid = 174,
    geteuid = 175,
    getgid = 176,
    getegid = 177,
    gettid = 178,
    sysinfo = 179,
    mq_open = 180,
    mq_unlink = 181,
    mq_timedsend = 182,
    mq_timedreceive = 183,
    mq_notify = 184,
    mq_getsetattr = 185,
    msgget = 186,
    msgctl = 187,
    msgrcv = 188,
    msgsnd = 189,
    semget = 190,
    semctl = 191,
    semtimedop = 192,
    semop = 193,
    shmget = 194,
    shmctl = 195,
    shmat = 196,
    shmdt = 197,
    socket = 198,
    socketpair = 199,
    bind = 200,
    listen = 201,
    accept = 202,
    connect = 203,
    getsockname = 204,
    getpeername = 205,
    sendto = 206,
    recvfrom = 207,
    setsockopt = 208,
    getsockopt = 209,
    shutdown = 210,
    sendmsg = 211,
    recvmsg = 212,
    readahead = 213,
    brk = 214,
    munmap = 215,
    mremap = 216,
    add_key = 217,
    request_key = 218,
    keyctl = 219,
    clone = 220,
    execve = 221,
    mmap2 = 222,
    fadvise64_64 = 223,
    swapon = 224,
    swapoff = 225,
    mprotect = 226,
    msync = 227,
    mlock = 228,
    munlock = 229,
    mlockall = 230,
    munlockall = 231,
    mincore = 232,
    madvise = 233,
    remap_file_pages = 234,
    mbind = 235,
    get_mempolicy = 236,
    set_mempolicy = 237,
    migrate_pages = 238,
    move_pages = 239,
    rt_tgsigqueueinfo = 240,
    perf_event_open = 241,
    accept4 = 242,
    recvmmsg = 243,
    cacheflush = 244,
    arc_settls = 245,
    arc_gettls = 246,
    sysfs = 247,
    arc_usr_cmpxchg = 248,
    wait4 = 260,
    prlimit64 = 261,
    fanotify_init = 262,
    fanotify_mark = 263,
    name_to_handle_at = 264,
    open_by_handle_at = 265,
    clock_adjtime = 266,
    syncfs = 267,
    setns = 268,
    sendmmsg = 269,
    process_vm_readv = 270,
    process_vm_writev = 271,
    kcmp = 272,
    finit_module = 273,
    sched_setattr = 274,
    sched_getattr = 275,
    renameat2 = 276,
    seccomp = 277,
    getrandom = 278,
    memfd_create = 279,
    bpf = 280,
    execveat = 281,
    userfaultfd = 282,
    membarrier = 283,
    mlock2 = 284,
    copy_file_range = 285,
    preadv2 = 286,
    pwritev2 = 287,
    pkey_mprotect = 288,
    pkey_alloc = 289,
    pkey_free = 290,
    statx = 291,
    io_pgetevents = 292,
    rseq = 293,
    kexec_file_load = 294,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const CSky = enum(usize) {
    io_setup = 0,
    io_destroy = 1,
    io_submit = 2,
    io_cancel = 3,
    io_getevents = 4,
    setxattr = 5,
    lsetxattr = 6,
    fsetxattr = 7,
    getxattr = 8,
    lgetxattr = 9,
    fgetxattr = 10,
    listxattr = 11,
    llistxattr = 12,
    flistxattr = 13,
    removexattr = 14,
    lremovexattr = 15,
    fremovexattr = 16,
    getcwd = 17,
    lookup_dcookie = 18,
    eventfd2 = 19,
    epoll_create1 = 20,
    epoll_ctl = 21,
    epoll_pwait = 22,
    dup = 23,
    dup3 = 24,
    fcntl64 = 25,
    inotify_init1 = 26,
    inotify_add_watch = 27,
    inotify_rm_watch = 28,
    ioctl = 29,
    ioprio_set = 30,
    ioprio_get = 31,
    flock = 32,
    mknodat = 33,
    mkdirat = 34,
    unlinkat = 35,
    symlinkat = 36,
    linkat = 37,
    umount2 = 39,
    mount = 40,
    pivot_root = 41,
    nfsservctl = 42,
    statfs64 = 43,
    fstatfs64 = 44,
    truncate64 = 45,
    ftruncate64 = 46,
    fallocate = 47,
    faccessat = 48,
    chdir = 49,
    fchdir = 50,
    chroot = 51,
    fchmod = 52,
    fchmodat = 53,
    fchownat = 54,
    fchown = 55,
    openat = 56,
    close = 57,
    vhangup = 58,
    pipe2 = 59,
    quotactl = 60,
    getdents64 = 61,
    llseek = 62,
    read = 63,
    write = 64,
    readv = 65,
    writev = 66,
    pread64 = 67,
    pwrite64 = 68,
    preadv = 69,
    pwritev = 70,
    sendfile64 = 71,
    pselect6 = 72,
    ppoll = 73,
    signalfd4 = 74,
    vmsplice = 75,
    splice = 76,
    tee = 77,
    readlinkat = 78,
    fstatat64 = 79,
    fstat64 = 80,
    sync = 81,
    fsync = 82,
    fdatasync = 83,
    sync_file_range = 84,
    timerfd_create = 85,
    timerfd_settime = 86,
    timerfd_gettime = 87,
    utimensat = 88,
    acct = 89,
    capget = 90,
    capset = 91,
    personality = 92,
    exit = 93,
    exit_group = 94,
    waitid = 95,
    set_tid_address = 96,
    unshare = 97,
    futex = 98,
    set_robust_list = 99,
    get_robust_list = 100,
    nanosleep = 101,
    getitimer = 102,
    setitimer = 103,
    kexec_load = 104,
    init_module = 105,
    delete_module = 106,
    timer_create = 107,
    timer_gettime = 108,
    timer_getoverrun = 109,
    timer_settime = 110,
    timer_delete = 111,
    clock_settime = 112,
    clock_gettime = 113,
    clock_getres = 114,
    clock_nanosleep = 115,
    syslog = 116,
    ptrace = 117,
    sched_setparam = 118,
    sched_setscheduler = 119,
    sched_getscheduler = 120,
    sched_getparam = 121,
    sched_setaffinity = 122,
    sched_getaffinity = 123,
    sched_yield = 124,
    sched_get_priority_max = 125,
    sched_get_priority_min = 126,
    sched_rr_get_interval = 127,
    restart_syscall = 128,
    kill = 129,
    tkill = 130,
    tgkill = 131,
    sigaltstack = 132,
    rt_sigsuspend = 133,
    rt_sigaction = 134,
    rt_sigprocmask = 135,
    rt_sigpending = 136,
    rt_sigtimedwait = 137,
    rt_sigqueueinfo = 138,
    rt_sigreturn = 139,
    setpriority = 140,
    getpriority = 141,
    reboot = 142,
    setregid = 143,
    setgid = 144,
    setreuid = 145,
    setuid = 146,
    setresuid = 147,
    getresuid = 148,
    setresgid = 149,
    getresgid = 150,
    setfsuid = 151,
    setfsgid = 152,
    times = 153,
    setpgid = 154,
    getpgid = 155,
    getsid = 156,
    setsid = 157,
    getgroups = 158,
    setgroups = 159,
    uname = 160,
    sethostname = 161,
    setdomainname = 162,
    getrlimit = 163,
    setrlimit = 164,
    getrusage = 165,
    umask = 166,
    prctl = 167,
    getcpu = 168,
    gettimeofday = 169,
    settimeofday = 170,
    adjtimex = 171,
    getpid = 172,
    getppid = 173,
    getuid = 174,
    geteuid = 175,
    getgid = 176,
    getegid = 177,
    gettid = 178,
    sysinfo = 179,
    mq_open = 180,
    mq_unlink = 181,
    mq_timedsend = 182,
    mq_timedreceive = 183,
    mq_notify = 184,
    mq_getsetattr = 185,
    msgget = 186,
    msgctl = 187,
    msgrcv = 188,
    msgsnd = 189,
    semget = 190,
    semctl = 191,
    semtimedop = 192,
    semop = 193,
    shmget = 194,
    shmctl = 195,
    shmat = 196,
    shmdt = 197,
    socket = 198,
    socketpair = 199,
    bind = 200,
    listen = 201,
    accept = 202,
    connect = 203,
    getsockname = 204,
    getpeername = 205,
    sendto = 206,
    recvfrom = 207,
    setsockopt = 208,
    getsockopt = 209,
    shutdown = 210,
    sendmsg = 211,
    recvmsg = 212,
    readahead = 213,
    brk = 214,
    munmap = 215,
    mremap = 216,
    add_key = 217,
    request_key = 218,
    keyctl = 219,
    clone = 220,
    execve = 221,
    mmap2 = 222,
    fadvise64_64 = 223,
    swapon = 224,
    swapoff = 225,
    mprotect = 226,
    msync = 227,
    mlock = 228,
    munlock = 229,
    mlockall = 230,
    munlockall = 231,
    mincore = 232,
    madvise = 233,
    remap_file_pages = 234,
    mbind = 235,
    get_mempolicy = 236,
    set_mempolicy = 237,
    migrate_pages = 238,
    move_pages = 239,
    rt_tgsigqueueinfo = 240,
    perf_event_open = 241,
    accept4 = 242,
    recvmmsg = 243,
    set_thread_area = 244,
    cacheflush = 245,
    wait4 = 260,
    prlimit64 = 261,
    fanotify_init = 262,
    fanotify_mark = 263,
    name_to_handle_at = 264,
    open_by_handle_at = 265,
    clock_adjtime = 266,
    syncfs = 267,
    setns = 268,
    sendmmsg = 269,
    process_vm_readv = 270,
    process_vm_writev = 271,
    kcmp = 272,
    finit_module = 273,
    sched_setattr = 274,
    sched_getattr = 275,
    renameat2 = 276,
    seccomp = 277,
    getrandom = 278,
    memfd_create = 279,
    bpf = 280,
    execveat = 281,
    userfaultfd = 282,
    membarrier = 283,
    mlock2 = 284,
    copy_file_range = 285,
    preadv2 = 286,
    pwritev2 = 287,
    pkey_mprotect = 288,
    pkey_alloc = 289,
    pkey_free = 290,
    statx = 291,
    io_pgetevents = 292,
    rseq = 293,
    kexec_file_load = 294,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const Hexagon = enum(usize) {
    io_setup = 0,
    io_destroy = 1,
    io_submit = 2,
    io_cancel = 3,
    io_getevents = 4,
    setxattr = 5,
    lsetxattr = 6,
    fsetxattr = 7,
    getxattr = 8,
    lgetxattr = 9,
    fgetxattr = 10,
    listxattr = 11,
    llistxattr = 12,
    flistxattr = 13,
    removexattr = 14,
    lremovexattr = 15,
    fremovexattr = 16,
    getcwd = 17,
    lookup_dcookie = 18,
    eventfd2 = 19,
    epoll_create1 = 20,
    epoll_ctl = 21,
    epoll_pwait = 22,
    dup = 23,
    dup3 = 24,
    fcntl64 = 25,
    inotify_init1 = 26,
    inotify_add_watch = 27,
    inotify_rm_watch = 28,
    ioctl = 29,
    ioprio_set = 30,
    ioprio_get = 31,
    flock = 32,
    mknodat = 33,
    mkdirat = 34,
    unlinkat = 35,
    symlinkat = 36,
    linkat = 37,
    renameat = 38,
    umount2 = 39,
    mount = 40,
    pivot_root = 41,
    nfsservctl = 42,
    statfs64 = 43,
    fstatfs64 = 44,
    truncate64 = 45,
    ftruncate64 = 46,
    fallocate = 47,
    faccessat = 48,
    chdir = 49,
    fchdir = 50,
    chroot = 51,
    fchmod = 52,
    fchmodat = 53,
    fchownat = 54,
    fchown = 55,
    openat = 56,
    close = 57,
    vhangup = 58,
    pipe2 = 59,
    quotactl = 60,
    getdents64 = 61,
    llseek = 62,
    read = 63,
    write = 64,
    readv = 65,
    writev = 66,
    pread64 = 67,
    pwrite64 = 68,
    preadv = 69,
    pwritev = 70,
    sendfile64 = 71,
    pselect6 = 72,
    ppoll = 73,
    signalfd4 = 74,
    vmsplice = 75,
    splice = 76,
    tee = 77,
    readlinkat = 78,
    fstatat64 = 79,
    fstat64 = 80,
    sync = 81,
    fsync = 82,
    fdatasync = 83,
    sync_file_range = 84,
    timerfd_create = 85,
    timerfd_settime = 86,
    timerfd_gettime = 87,
    utimensat = 88,
    acct = 89,
    capget = 90,
    capset = 91,
    personality = 92,
    exit = 93,
    exit_group = 94,
    waitid = 95,
    set_tid_address = 96,
    unshare = 97,
    futex = 98,
    set_robust_list = 99,
    get_robust_list = 100,
    nanosleep = 101,
    getitimer = 102,
    setitimer = 103,
    kexec_load = 104,
    init_module = 105,
    delete_module = 106,
    timer_create = 107,
    timer_gettime = 108,
    timer_getoverrun = 109,
    timer_settime = 110,
    timer_delete = 111,
    clock_settime = 112,
    clock_gettime = 113,
    clock_getres = 114,
    clock_nanosleep = 115,
    syslog = 116,
    ptrace = 117,
    sched_setparam = 118,
    sched_setscheduler = 119,
    sched_getscheduler = 120,
    sched_getparam = 121,
    sched_setaffinity = 122,
    sched_getaffinity = 123,
    sched_yield = 124,
    sched_get_priority_max = 125,
    sched_get_priority_min = 126,
    sched_rr_get_interval = 127,
    restart_syscall = 128,
    kill = 129,
    tkill = 130,
    tgkill = 131,
    sigaltstack = 132,
    rt_sigsuspend = 133,
    rt_sigaction = 134,
    rt_sigprocmask = 135,
    rt_sigpending = 136,
    rt_sigtimedwait = 137,
    rt_sigqueueinfo = 138,
    rt_sigreturn = 139,
    setpriority = 140,
    getpriority = 141,
    reboot = 142,
    setregid = 143,
    setgid = 144,
    setreuid = 145,
    setuid = 146,
    setresuid = 147,
    getresuid = 148,
    setresgid = 149,
    getresgid = 150,
    setfsuid = 151,
    setfsgid = 152,
    times = 153,
    setpgid = 154,
    getpgid = 155,
    getsid = 156,
    setsid = 157,
    getgroups = 158,
    setgroups = 159,
    uname = 160,
    sethostname = 161,
    setdomainname = 162,
    getrlimit = 163,
    setrlimit = 164,
    getrusage = 165,
    umask = 166,
    prctl = 167,
    getcpu = 168,
    gettimeofday = 169,
    settimeofday = 170,
    adjtimex = 171,
    getpid = 172,
    getppid = 173,
    getuid = 174,
    geteuid = 175,
    getgid = 176,
    getegid = 177,
    gettid = 178,
    sysinfo = 179,
    mq_open = 180,
    mq_unlink = 181,
    mq_timedsend = 182,
    mq_timedreceive = 183,
    mq_notify = 184,
    mq_getsetattr = 185,
    msgget = 186,
    msgctl = 187,
    msgrcv = 188,
    msgsnd = 189,
    semget = 190,
    semctl = 191,
    semtimedop = 192,
    semop = 193,
    shmget = 194,
    shmctl = 195,
    shmat = 196,
    shmdt = 197,
    socket = 198,
    socketpair = 199,
    bind = 200,
    listen = 201,
    accept = 202,
    connect = 203,
    getsockname = 204,
    getpeername = 205,
    sendto = 206,
    recvfrom = 207,
    setsockopt = 208,
    getsockopt = 209,
    shutdown = 210,
    sendmsg = 211,
    recvmsg = 212,
    readahead = 213,
    brk = 214,
    munmap = 215,
    mremap = 216,
    add_key = 217,
    request_key = 218,
    keyctl = 219,
    clone = 220,
    execve = 221,
    mmap2 = 222,
    fadvise64_64 = 223,
    swapon = 224,
    swapoff = 225,
    mprotect = 226,
    msync = 227,
    mlock = 228,
    munlock = 229,
    mlockall = 230,
    munlockall = 231,
    mincore = 232,
    madvise = 233,
    remap_file_pages = 234,
    mbind = 235,
    get_mempolicy = 236,
    set_mempolicy = 237,
    migrate_pages = 238,
    move_pages = 239,
    rt_tgsigqueueinfo = 240,
    perf_event_open = 241,
    accept4 = 242,
    recvmmsg = 243,
    wait4 = 260,
    prlimit64 = 261,
    fanotify_init = 262,
    fanotify_mark = 263,
    name_to_handle_at = 264,
    open_by_handle_at = 265,
    clock_adjtime = 266,
    syncfs = 267,
    setns = 268,
    sendmmsg = 269,
    process_vm_readv = 270,
    process_vm_writev = 271,
    kcmp = 272,
    finit_module = 273,
    sched_setattr = 274,
    sched_getattr = 275,
    renameat2 = 276,
    seccomp = 277,
    getrandom = 278,
    memfd_create = 279,
    bpf = 280,
    execveat = 281,
    userfaultfd = 282,
    membarrier = 283,
    mlock2 = 284,
    copy_file_range = 285,
    preadv2 = 286,
    pwritev2 = 287,
    pkey_mprotect = 288,
    pkey_alloc = 289,
    pkey_free = 290,
    statx = 291,
    io_pgetevents = 292,
    rseq = 293,
    kexec_file_load = 294,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const OpenRisc = enum(usize) {
    io_setup = 0,
    io_destroy = 1,
    io_submit = 2,
    io_cancel = 3,
    io_getevents = 4,
    setxattr = 5,
    lsetxattr = 6,
    fsetxattr = 7,
    getxattr = 8,
    lgetxattr = 9,
    fgetxattr = 10,
    listxattr = 11,
    llistxattr = 12,
    flistxattr = 13,
    removexattr = 14,
    lremovexattr = 15,
    fremovexattr = 16,
    getcwd = 17,
    lookup_dcookie = 18,
    eventfd2 = 19,
    epoll_create1 = 20,
    epoll_ctl = 21,
    epoll_pwait = 22,
    dup = 23,
    dup3 = 24,
    fcntl64 = 25,
    inotify_init1 = 26,
    inotify_add_watch = 27,
    inotify_rm_watch = 28,
    ioctl = 29,
    ioprio_set = 30,
    ioprio_get = 31,
    flock = 32,
    mknodat = 33,
    mkdirat = 34,
    unlinkat = 35,
    symlinkat = 36,
    linkat = 37,
    renameat = 38,
    umount2 = 39,
    mount = 40,
    pivot_root = 41,
    nfsservctl = 42,
    statfs64 = 43,
    fstatfs64 = 44,
    truncate64 = 45,
    ftruncate64 = 46,
    fallocate = 47,
    faccessat = 48,
    chdir = 49,
    fchdir = 50,
    chroot = 51,
    fchmod = 52,
    fchmodat = 53,
    fchownat = 54,
    fchown = 55,
    openat = 56,
    close = 57,
    vhangup = 58,
    pipe2 = 59,
    quotactl = 60,
    getdents64 = 61,
    llseek = 62,
    read = 63,
    write = 64,
    readv = 65,
    writev = 66,
    pread64 = 67,
    pwrite64 = 68,
    preadv = 69,
    pwritev = 70,
    sendfile64 = 71,
    pselect6 = 72,
    ppoll = 73,
    signalfd4 = 74,
    vmsplice = 75,
    splice = 76,
    tee = 77,
    readlinkat = 78,
    fstatat64 = 79,
    fstat64 = 80,
    sync = 81,
    fsync = 82,
    fdatasync = 83,
    sync_file_range = 84,
    timerfd_create = 85,
    timerfd_settime = 86,
    timerfd_gettime = 87,
    utimensat = 88,
    acct = 89,
    capget = 90,
    capset = 91,
    personality = 92,
    exit = 93,
    exit_group = 94,
    waitid = 95,
    set_tid_address = 96,
    unshare = 97,
    futex = 98,
    set_robust_list = 99,
    get_robust_list = 100,
    nanosleep = 101,
    getitimer = 102,
    setitimer = 103,
    kexec_load = 104,
    init_module = 105,
    delete_module = 106,
    timer_create = 107,
    timer_gettime = 108,
    timer_getoverrun = 109,
    timer_settime = 110,
    timer_delete = 111,
    clock_settime = 112,
    clock_gettime = 113,
    clock_getres = 114,
    clock_nanosleep = 115,
    syslog = 116,
    ptrace = 117,
    sched_setparam = 118,
    sched_setscheduler = 119,
    sched_getscheduler = 120,
    sched_getparam = 121,
    sched_setaffinity = 122,
    sched_getaffinity = 123,
    sched_yield = 124,
    sched_get_priority_max = 125,
    sched_get_priority_min = 126,
    sched_rr_get_interval = 127,
    restart_syscall = 128,
    kill = 129,
    tkill = 130,
    tgkill = 131,
    sigaltstack = 132,
    rt_sigsuspend = 133,
    rt_sigaction = 134,
    rt_sigprocmask = 135,
    rt_sigpending = 136,
    rt_sigtimedwait = 137,
    rt_sigqueueinfo = 138,
    rt_sigreturn = 139,
    setpriority = 140,
    getpriority = 141,
    reboot = 142,
    setregid = 143,
    setgid = 144,
    setreuid = 145,
    setuid = 146,
    setresuid = 147,
    getresuid = 148,
    setresgid = 149,
    getresgid = 150,
    setfsuid = 151,
    setfsgid = 152,
    times = 153,
    setpgid = 154,
    getpgid = 155,
    getsid = 156,
    setsid = 157,
    getgroups = 158,
    setgroups = 159,
    uname = 160,
    sethostname = 161,
    setdomainname = 162,
    getrlimit = 163,
    setrlimit = 164,
    getrusage = 165,
    umask = 166,
    prctl = 167,
    getcpu = 168,
    gettimeofday = 169,
    settimeofday = 170,
    adjtimex = 171,
    getpid = 172,
    getppid = 173,
    getuid = 174,
    geteuid = 175,
    getgid = 176,
    getegid = 177,
    gettid = 178,
    sysinfo = 179,
    mq_open = 180,
    mq_unlink = 181,
    mq_timedsend = 182,
    mq_timedreceive = 183,
    mq_notify = 184,
    mq_getsetattr = 185,
    msgget = 186,
    msgctl = 187,
    msgrcv = 188,
    msgsnd = 189,
    semget = 190,
    semctl = 191,
    semtimedop = 192,
    semop = 193,
    shmget = 194,
    shmctl = 195,
    shmat = 196,
    shmdt = 197,
    socket = 198,
    socketpair = 199,
    bind = 200,
    listen = 201,
    accept = 202,
    connect = 203,
    getsockname = 204,
    getpeername = 205,
    sendto = 206,
    recvfrom = 207,
    setsockopt = 208,
    getsockopt = 209,
    shutdown = 210,
    sendmsg = 211,
    recvmsg = 212,
    readahead = 213,
    brk = 214,
    munmap = 215,
    mremap = 216,
    add_key = 217,
    request_key = 218,
    keyctl = 219,
    clone = 220,
    execve = 221,
    mmap2 = 222,
    fadvise64_64 = 223,
    swapon = 224,
    swapoff = 225,
    mprotect = 226,
    msync = 227,
    mlock = 228,
    munlock = 229,
    mlockall = 230,
    munlockall = 231,
    mincore = 232,
    madvise = 233,
    remap_file_pages = 234,
    mbind = 235,
    get_mempolicy = 236,
    set_mempolicy = 237,
    migrate_pages = 238,
    move_pages = 239,
    rt_tgsigqueueinfo = 240,
    perf_event_open = 241,
    accept4 = 242,
    recvmmsg = 243,
    or1k_atomic = 244,
    wait4 = 260,
    prlimit64 = 261,
    fanotify_init = 262,
    fanotify_mark = 263,
    name_to_handle_at = 264,
    open_by_handle_at = 265,
    clock_adjtime = 266,
    syncfs = 267,
    setns = 268,
    sendmmsg = 269,
    process_vm_readv = 270,
    process_vm_writev = 271,
    kcmp = 272,
    finit_module = 273,
    sched_setattr = 274,
    sched_getattr = 275,
    renameat2 = 276,
    seccomp = 277,
    getrandom = 278,
    memfd_create = 279,
    bpf = 280,
    execveat = 281,
    userfaultfd = 282,
    membarrier = 283,
    mlock2 = 284,
    copy_file_range = 285,
    preadv2 = 286,
    pwritev2 = 287,
    pkey_mprotect = 288,
    pkey_alloc = 289,
    pkey_free = 290,
    statx = 291,
    io_pgetevents = 292,
    rseq = 293,
    kexec_file_load = 294,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const Hppa = enum(usize) {
    restart_syscall = 0,
    exit = 1,
    fork = 2,
    read = 3,
    write = 4,
    open = 5,
    close = 6,
    waitpid = 7,
    creat = 8,
    link = 9,
    unlink = 10,
    execve = 11,
    chdir = 12,
    time = 13,
    mknod = 14,
    chmod = 15,
    lchown = 16,
    socket = 17,
    stat = 18,
    lseek = 19,
    getpid = 20,
    mount = 21,
    bind = 22,
    setuid = 23,
    getuid = 24,
    stime = 25,
    ptrace = 26,
    alarm = 27,
    fstat = 28,
    pause = 29,
    utime = 30,
    connect = 31,
    listen = 32,
    access = 33,
    nice = 34,
    accept = 35,
    sync = 36,
    kill = 37,
    rename = 38,
    mkdir = 39,
    rmdir = 40,
    dup = 41,
    pipe = 42,
    times = 43,
    getsockname = 44,
    brk = 45,
    setgid = 46,
    getgid = 47,
    signal = 48,
    geteuid = 49,
    getegid = 50,
    acct = 51,
    umount2 = 52,
    getpeername = 53,
    ioctl = 54,
    fcntl = 55,
    socketpair = 56,
    setpgid = 57,
    send = 58,
    uname = 59,
    umask = 60,
    chroot = 61,
    ustat = 62,
    dup2 = 63,
    getppid = 64,
    getpgrp = 65,
    setsid = 66,
    pivot_root = 67,
    sgetmask = 68,
    ssetmask = 69,
    setreuid = 70,
    setregid = 71,
    mincore = 72,
    sigpending = 73,
    sethostname = 74,
    setrlimit = 75,
    getrlimit = 76,
    getrusage = 77,
    gettimeofday = 78,
    settimeofday = 79,
    getgroups = 80,
    setgroups = 81,
    sendto = 82,
    symlink = 83,
    lstat = 84,
    readlink = 85,
    uselib = 86,
    swapon = 87,
    reboot = 88,
    mmap2 = 89,
    mmap = 90,
    munmap = 91,
    truncate = 92,
    ftruncate = 93,
    fchmod = 94,
    fchown = 95,
    getpriority = 96,
    setpriority = 97,
    recv = 98,
    statfs = 99,
    fstatfs = 100,
    stat64 = 101,
    syslog = 103,
    setitimer = 104,
    getitimer = 105,
    capget = 106,
    capset = 107,
    pread64 = 108,
    pwrite64 = 109,
    getcwd = 110,
    vhangup = 111,
    fstat64 = 112,
    vfork = 113,
    wait4 = 114,
    swapoff = 115,
    sysinfo = 116,
    shutdown = 117,
    fsync = 118,
    madvise = 119,
    clone = 120,
    setdomainname = 121,
    sendfile = 122,
    recvfrom = 123,
    adjtimex = 124,
    mprotect = 125,
    sigprocmask = 126,
    init_module = 128,
    delete_module = 129,
    quotactl = 131,
    getpgid = 132,
    fchdir = 133,
    bdflush = 134,
    sysfs = 135,
    personality = 136,
    setfsuid = 138,
    setfsgid = 139,
    llseek = 140,
    getdents = 141,
    newselect = 142,
    flock = 143,
    msync = 144,
    readv = 145,
    writev = 146,
    getsid = 147,
    fdatasync = 148,
    sysctl = 149,
    mlock = 150,
    munlock = 151,
    mlockall = 152,
    munlockall = 153,
    sched_setparam = 154,
    sched_getparam = 155,
    sched_setscheduler = 156,
    sched_getscheduler = 157,
    sched_yield = 158,
    sched_get_priority_max = 159,
    sched_get_priority_min = 160,
    sched_rr_get_interval = 161,
    nanosleep = 162,
    mremap = 163,
    setresuid = 164,
    getresuid = 165,
    sigaltstack = 166,
    poll = 168,
    setresgid = 170,
    getresgid = 171,
    prctl = 172,
    rt_sigreturn = 173,
    rt_sigaction = 174,
    rt_sigprocmask = 175,
    rt_sigpending = 176,
    rt_sigtimedwait = 177,
    rt_sigqueueinfo = 178,
    rt_sigsuspend = 179,
    chown = 180,
    setsockopt = 181,
    getsockopt = 182,
    sendmsg = 183,
    recvmsg = 184,
    semop = 185,
    semget = 186,
    semctl = 187,
    msgsnd = 188,
    msgrcv = 189,
    msgget = 190,
    msgctl = 191,
    shmat = 192,
    shmdt = 193,
    shmget = 194,
    shmctl = 195,
    lstat64 = 198,
    truncate64 = 199,
    ftruncate64 = 200,
    getdents64 = 201,
    fcntl64 = 202,
    gettid = 206,
    readahead = 207,
    tkill = 208,
    sendfile64 = 209,
    futex = 210,
    sched_setaffinity = 211,
    sched_getaffinity = 212,
    io_setup = 215,
    io_destroy = 216,
    io_getevents = 217,
    io_submit = 218,
    io_cancel = 219,
    exit_group = 222,
    lookup_dcookie = 223,
    epoll_create = 224,
    epoll_ctl = 225,
    epoll_wait = 226,
    remap_file_pages = 227,
    semtimedop = 228,
    mq_open = 229,
    mq_unlink = 230,
    mq_timedsend = 231,
    mq_timedreceive = 232,
    mq_notify = 233,
    mq_getsetattr = 234,
    waitid = 235,
    fadvise64_64 = 236,
    set_tid_address = 237,
    setxattr = 238,
    lsetxattr = 239,
    fsetxattr = 240,
    getxattr = 241,
    lgetxattr = 242,
    fgetxattr = 243,
    listxattr = 244,
    llistxattr = 245,
    flistxattr = 246,
    removexattr = 247,
    lremovexattr = 248,
    fremovexattr = 249,
    timer_create = 250,
    timer_settime = 251,
    timer_gettime = 252,
    timer_getoverrun = 253,
    timer_delete = 254,
    clock_settime = 255,
    clock_gettime = 256,
    clock_getres = 257,
    clock_nanosleep = 258,
    tgkill = 259,
    mbind = 260,
    get_mempolicy = 261,
    set_mempolicy = 262,
    add_key = 264,
    request_key = 265,
    keyctl = 266,
    ioprio_set = 267,
    ioprio_get = 268,
    inotify_init = 269,
    inotify_add_watch = 270,
    inotify_rm_watch = 271,
    migrate_pages = 272,
    pselect6 = 273,
    ppoll = 274,
    openat = 275,
    mkdirat = 276,
    mknodat = 277,
    fchownat = 278,
    futimesat = 279,
    fstatat64 = 280,
    unlinkat = 281,
    renameat = 282,
    linkat = 283,
    symlinkat = 284,
    readlinkat = 285,
    fchmodat = 286,
    faccessat = 287,
    unshare = 288,
    set_robust_list = 289,
    get_robust_list = 290,
    splice = 291,
    sync_file_range = 292,
    tee = 293,
    vmsplice = 294,
    move_pages = 295,
    getcpu = 296,
    epoll_pwait = 297,
    statfs64 = 298,
    fstatfs64 = 299,
    kexec_load = 300,
    utimensat = 301,
    signalfd = 302,
    eventfd = 304,
    fallocate = 305,
    timerfd_create = 306,
    timerfd_settime = 307,
    timerfd_gettime = 308,
    signalfd4 = 309,
    eventfd2 = 310,
    epoll_create1 = 311,
    dup3 = 312,
    pipe2 = 313,
    inotify_init1 = 314,
    preadv = 315,
    pwritev = 316,
    rt_tgsigqueueinfo = 317,
    perf_event_open = 318,
    recvmmsg = 319,
    accept4 = 320,
    prlimit64 = 321,
    fanotify_init = 322,
    fanotify_mark = 323,
    clock_adjtime = 324,
    name_to_handle_at = 325,
    open_by_handle_at = 326,
    syncfs = 327,
    setns = 328,
    sendmmsg = 329,
    process_vm_readv = 330,
    process_vm_writev = 331,
    kcmp = 332,
    finit_module = 333,
    sched_setattr = 334,
    sched_getattr = 335,
    utimes = 336,
    renameat2 = 337,
    seccomp = 338,
    getrandom = 339,
    memfd_create = 340,
    bpf = 341,
    execveat = 342,
    membarrier = 343,
    userfaultfd = 344,
    mlock2 = 345,
    copy_file_range = 346,
    preadv2 = 347,
    pwritev2 = 348,
    statx = 349,
    io_pgetevents = 350,
    pkey_mprotect = 351,
    pkey_alloc = 352,
    pkey_free = 353,
    rseq = 354,
    kexec_file_load = 355,
    cacheflush = 356,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const Hppa64 = enum(usize) {
    restart_syscall = 0,
    exit = 1,
    fork = 2,
    read = 3,
    write = 4,
    open = 5,
    close = 6,
    waitpid = 7,
    creat = 8,
    link = 9,
    unlink = 10,
    execve = 11,
    chdir = 12,
    time = 13,
    mknod = 14,
    chmod = 15,
    lchown = 16,
    socket = 17,
    stat = 18,
    lseek = 19,
    getpid = 20,
    mount = 21,
    bind = 22,
    setuid = 23,
    getuid = 24,
    stime = 25,
    ptrace = 26,
    alarm = 27,
    fstat = 28,
    pause = 29,
    utime = 30,
    connect = 31,
    listen = 32,
    access = 33,
    nice = 34,
    accept = 35,
    sync = 36,
    kill = 37,
    rename = 38,
    mkdir = 39,
    rmdir = 40,
    dup = 41,
    pipe = 42,
    times = 43,
    getsockname = 44,
    brk = 45,
    setgid = 46,
    getgid = 47,
    signal = 48,
    geteuid = 49,
    getegid = 50,
    acct = 51,
    umount2 = 52,
    getpeername = 53,
    ioctl = 54,
    fcntl = 55,
    socketpair = 56,
    setpgid = 57,
    send = 58,
    uname = 59,
    umask = 60,
    chroot = 61,
    ustat = 62,
    dup2 = 63,
    getppid = 64,
    getpgrp = 65,
    setsid = 66,
    pivot_root = 67,
    sgetmask = 68,
    ssetmask = 69,
    setreuid = 70,
    setregid = 71,
    mincore = 72,
    sigpending = 73,
    sethostname = 74,
    setrlimit = 75,
    getrlimit = 76,
    getrusage = 77,
    gettimeofday = 78,
    settimeofday = 79,
    getgroups = 80,
    setgroups = 81,
    sendto = 82,
    symlink = 83,
    lstat = 84,
    readlink = 85,
    uselib = 86,
    swapon = 87,
    reboot = 88,
    mmap2 = 89,
    mmap = 90,
    munmap = 91,
    truncate = 92,
    ftruncate = 93,
    fchmod = 94,
    fchown = 95,
    getpriority = 96,
    setpriority = 97,
    recv = 98,
    statfs = 99,
    fstatfs = 100,
    stat64 = 101,
    syslog = 103,
    setitimer = 104,
    getitimer = 105,
    capget = 106,
    capset = 107,
    pread64 = 108,
    pwrite64 = 109,
    getcwd = 110,
    vhangup = 111,
    fstat64 = 112,
    vfork = 113,
    wait4 = 114,
    swapoff = 115,
    sysinfo = 116,
    shutdown = 117,
    fsync = 118,
    madvise = 119,
    clone = 120,
    setdomainname = 121,
    sendfile = 122,
    recvfrom = 123,
    adjtimex = 124,
    mprotect = 125,
    sigprocmask = 126,
    init_module = 128,
    delete_module = 129,
    quotactl = 131,
    getpgid = 132,
    fchdir = 133,
    bdflush = 134,
    sysfs = 135,
    personality = 136,
    setfsuid = 138,
    setfsgid = 139,
    getdents = 141,
    newselect = 142,
    flock = 143,
    msync = 144,
    readv = 145,
    writev = 146,
    getsid = 147,
    fdatasync = 148,
    sysctl = 149,
    mlock = 150,
    munlock = 151,
    mlockall = 152,
    munlockall = 153,
    sched_setparam = 154,
    sched_getparam = 155,
    sched_setscheduler = 156,
    sched_getscheduler = 157,
    sched_yield = 158,
    sched_get_priority_max = 159,
    sched_get_priority_min = 160,
    sched_rr_get_interval = 161,
    nanosleep = 162,
    mremap = 163,
    setresuid = 164,
    getresuid = 165,
    sigaltstack = 166,
    poll = 168,
    setresgid = 170,
    getresgid = 171,
    prctl = 172,
    rt_sigreturn = 173,
    rt_sigaction = 174,
    rt_sigprocmask = 175,
    rt_sigpending = 176,
    rt_sigtimedwait = 177,
    rt_sigqueueinfo = 178,
    rt_sigsuspend = 179,
    chown = 180,
    setsockopt = 181,
    getsockopt = 182,
    sendmsg = 183,
    recvmsg = 184,
    semop = 185,
    semget = 186,
    semctl = 187,
    msgsnd = 188,
    msgrcv = 189,
    msgget = 190,
    msgctl = 191,
    shmat = 192,
    shmdt = 193,
    shmget = 194,
    shmctl = 195,
    lstat64 = 198,
    truncate64 = 199,
    ftruncate64 = 200,
    getdents64 = 201,
    fcntl64 = 202,
    gettid = 206,
    readahead = 207,
    tkill = 208,
    sendfile64 = 209,
    futex = 210,
    sched_setaffinity = 211,
    sched_getaffinity = 212,
    io_setup = 215,
    io_destroy = 216,
    io_getevents = 217,
    io_submit = 218,
    io_cancel = 219,
    exit_group = 222,
    lookup_dcookie = 223,
    epoll_create = 224,
    epoll_ctl = 225,
    epoll_wait = 226,
    remap_file_pages = 227,
    semtimedop = 228,
    mq_open = 229,
    mq_unlink = 230,
    mq_timedsend = 231,
    mq_timedreceive = 232,
    mq_notify = 233,
    mq_getsetattr = 234,
    waitid = 235,
    fadvise64_64 = 236,
    set_tid_address = 237,
    setxattr = 238,
    lsetxattr = 239,
    fsetxattr = 240,
    getxattr = 241,
    lgetxattr = 242,
    fgetxattr = 243,
    listxattr = 244,
    llistxattr = 245,
    flistxattr = 246,
    removexattr = 247,
    lremovexattr = 248,
    fremovexattr = 249,
    timer_create = 250,
    timer_settime = 251,
    timer_gettime = 252,
    timer_getoverrun = 253,
    timer_delete = 254,
    clock_settime = 255,
    clock_gettime = 256,
    clock_getres = 257,
    clock_nanosleep = 258,
    tgkill = 259,
    mbind = 260,
    get_mempolicy = 261,
    set_mempolicy = 262,
    add_key = 264,
    request_key = 265,
    keyctl = 266,
    ioprio_set = 267,
    ioprio_get = 268,
    inotify_init = 269,
    inotify_add_watch = 270,
    inotify_rm_watch = 271,
    migrate_pages = 272,
    pselect6 = 273,
    ppoll = 274,
    openat = 275,
    mkdirat = 276,
    mknodat = 277,
    fchownat = 278,
    futimesat = 279,
    fstatat64 = 280,
    unlinkat = 281,
    renameat = 282,
    linkat = 283,
    symlinkat = 284,
    readlinkat = 285,
    fchmodat = 286,
    faccessat = 287,
    unshare = 288,
    set_robust_list = 289,
    get_robust_list = 290,
    splice = 291,
    sync_file_range = 292,
    tee = 293,
    vmsplice = 294,
    move_pages = 295,
    getcpu = 296,
    epoll_pwait = 297,
    statfs64 = 298,
    fstatfs64 = 299,
    kexec_load = 300,
    utimensat = 301,
    signalfd = 302,
    eventfd = 304,
    fallocate = 305,
    timerfd_create = 306,
    timerfd_settime = 307,
    timerfd_gettime = 308,
    signalfd4 = 309,
    eventfd2 = 310,
    epoll_create1 = 311,
    dup3 = 312,
    pipe2 = 313,
    inotify_init1 = 314,
    preadv = 315,
    pwritev = 316,
    rt_tgsigqueueinfo = 317,
    perf_event_open = 318,
    recvmmsg = 319,
    accept4 = 320,
    prlimit64 = 321,
    fanotify_init = 322,
    fanotify_mark = 323,
    clock_adjtime = 324,
    name_to_handle_at = 325,
    open_by_handle_at = 326,
    syncfs = 327,
    setns = 328,
    sendmmsg = 329,
    process_vm_readv = 330,
    process_vm_writev = 331,
    kcmp = 332,
    finit_module = 333,
    sched_setattr = 334,
    sched_getattr = 335,
    utimes = 336,
    renameat2 = 337,
    seccomp = 338,
    getrandom = 339,
    memfd_create = 340,
    bpf = 341,
    execveat = 342,
    membarrier = 343,
    userfaultfd = 344,
    mlock2 = 345,
    copy_file_range = 346,
    preadv2 = 347,
    pwritev2 = 348,
    statx = 349,
    io_pgetevents = 350,
    pkey_mprotect = 351,
    pkey_alloc = 352,
    pkey_free = 353,
    rseq = 354,
    kexec_file_load = 355,
    cacheflush = 356,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const Sh = enum(usize) {
    restart_syscall = 0,
    exit = 1,
    fork = 2,
    read = 3,
    write = 4,
    open = 5,
    close = 6,
    waitpid = 7,
    creat = 8,
    link = 9,
    unlink = 10,
    execve = 11,
    chdir = 12,
    time = 13,
    mknod = 14,
    chmod = 15,
    lchown = 16,
    oldstat = 18,
    lseek = 19,
    getpid = 20,
    mount = 21,
    umount = 22,
    setuid = 23,
    getuid = 24,
    stime = 25,
    ptrace = 26,
    alarm = 27,
    oldfstat = 28,
    pause = 29,
    utime = 30,
    access = 33,
    nice = 34,
    sync = 36,
    kill = 37,
    rename = 38,
    mkdir = 39,
    rmdir = 40,
    dup = 41,
    pipe = 42,
    times = 43,
    brk = 45,
    setgid = 46,
    getgid = 47,
    signal = 48,
    geteuid = 49,
    getegid = 50,
    acct = 51,
    umount2 = 52,
    ioctl = 54,
    fcntl = 55,
    setpgid = 57,
    umask = 60,
    chroot = 61,
    ustat = 62,
    dup2 = 63,
    getppid = 64,
    getpgrp = 65,
    setsid = 66,
    sigaction = 67,
    sgetmask = 68,
    ssetmask = 69,
    setreuid = 70,
    setregid = 71,
    sigsuspend = 72,
    sigpending = 73,
    sethostname = 74,
    setrlimit = 75,
    getrlimit = 76,
    getrusage = 77,
    gettimeofday = 78,
    settimeofday = 79,
    getgroups = 80,
    setgroups = 81,
    symlink = 83,
    oldlstat = 84,
    readlink = 85,
    uselib = 86,
    swapon = 87,
    reboot = 88,
    readdir = 89,
    mmap = 90,
    munmap = 91,
    truncate = 92,
    ftruncate = 93,
    fchmod = 94,
    fchown = 95,
    getpriority = 96,
    setpriority = 97,
    statfs = 99,
    fstatfs = 100,
    socketcall = 102,
    syslog = 103,
    setitimer = 104,
    getitimer = 105,
    stat = 106,
    lstat = 107,
    fstat = 108,
    olduname = 109,
    vhangup = 111,
    wait4 = 114,
    swapoff = 115,
    sysinfo = 116,
    ipc = 117,
    fsync = 118,
    sigreturn = 119,
    clone = 120,
    setdomainname = 121,
    uname = 122,
    cacheflush = 123,
    adjtimex = 124,
    mprotect = 125,
    sigprocmask = 126,
    init_module = 128,
    delete_module = 129,
    quotactl = 131,
    getpgid = 132,
    fchdir = 133,
    bdflush = 134,
    sysfs = 135,
    personality = 136,
    setfsuid = 138,
    setfsgid = 139,
    llseek = 140,
    getdents = 141,
    newselect = 142,
    flock = 143,
    msync = 144,
    readv = 145,
    writev = 146,
    getsid = 147,
    fdatasync = 148,
    sysctl = 149,
    mlock = 150,
    munlock = 151,
    mlockall = 152,
    munlockall = 153,
    sched_setparam = 154,
    sched_getparam = 155,
    sched_setscheduler = 156,
    sched_getscheduler = 157,
    sched_yield = 158,
    sched_get_priority_max = 159,
    sched_get_priority_min = 160,
    sched_rr_get_interval = 161,
    nanosleep = 162,
    mremap = 163,
    setresuid = 164,
    getresuid = 165,
    poll = 168,
    nfsservctl = 169,
    setresgid = 170,
    getresgid = 171,
    prctl = 172,
    rt_sigreturn = 173,
    rt_sigaction = 174,
    rt_sigprocmask = 175,
    rt_sigpending = 176,
    rt_sigtimedwait = 177,
    rt_sigqueueinfo = 178,
    rt_sigsuspend = 179,
    pread64 = 180,
    pwrite64 = 181,
    chown = 182,
    getcwd = 183,
    capget = 184,
    capset = 185,
    sigaltstack = 186,
    sendfile = 187,
    vfork = 190,
    ugetrlimit = 191,
    mmap2 = 192,
    truncate64 = 193,
    ftruncate64 = 194,
    stat64 = 195,
    lstat64 = 196,
    fstat64 = 197,
    lchown32 = 198,
    getuid32 = 199,
    getgid32 = 200,
    geteuid32 = 201,
    getegid32 = 202,
    setreuid32 = 203,
    setregid32 = 204,
    getgroups32 = 205,
    setgroups32 = 206,
    fchown32 = 207,
    setresuid32 = 208,
    getresuid32 = 209,
    setresgid32 = 210,
    getresgid32 = 211,
    chown32 = 212,
    setuid32 = 213,
    setgid32 = 214,
    setfsuid32 = 215,
    setfsgid32 = 216,
    pivot_root = 217,
    mincore = 218,
    madvise = 219,
    getdents64 = 220,
    fcntl64 = 221,
    gettid = 224,
    readahead = 225,
    setxattr = 226,
    lsetxattr = 227,
    fsetxattr = 228,
    getxattr = 229,
    lgetxattr = 230,
    fgetxattr = 231,
    listxattr = 232,
    llistxattr = 233,
    flistxattr = 234,
    removexattr = 235,
    lremovexattr = 236,
    fremovexattr = 237,
    tkill = 238,
    sendfile64 = 239,
    futex = 240,
    sched_setaffinity = 241,
    sched_getaffinity = 242,
    io_setup = 245,
    io_destroy = 246,
    io_getevents = 247,
    io_submit = 248,
    io_cancel = 249,
    fadvise64 = 250,
    exit_group = 252,
    lookup_dcookie = 253,
    epoll_create = 254,
    epoll_ctl = 255,
    epoll_wait = 256,
    remap_file_pages = 257,
    set_tid_address = 258,
    timer_create = 259,
    timer_settime = 260,
    timer_gettime = 261,
    timer_getoverrun = 262,
    timer_delete = 263,
    clock_settime = 264,
    clock_gettime = 265,
    clock_getres = 266,
    clock_nanosleep = 267,
    statfs64 = 268,
    fstatfs64 = 269,
    tgkill = 270,
    utimes = 271,
    fadvise64_64 = 272,
    mbind = 274,
    get_mempolicy = 275,
    set_mempolicy = 276,
    mq_open = 277,
    mq_unlink = 278,
    mq_timedsend = 279,
    mq_timedreceive = 280,
    mq_notify = 281,
    mq_getsetattr = 282,
    kexec_load = 283,
    waitid = 284,
    add_key = 285,
    request_key = 286,
    keyctl = 287,
    ioprio_set = 288,
    ioprio_get = 289,
    inotify_init = 290,
    inotify_add_watch = 291,
    inotify_rm_watch = 292,
    migrate_pages = 294,
    openat = 295,
    mkdirat = 296,
    mknodat = 297,
    fchownat = 298,
    futimesat = 299,
    fstatat64 = 300,
    unlinkat = 301,
    renameat = 302,
    linkat = 303,
    symlinkat = 304,
    readlinkat = 305,
    fchmodat = 306,
    faccessat = 307,
    pselect6 = 308,
    ppoll = 309,
    unshare = 310,
    set_robust_list = 311,
    get_robust_list = 312,
    splice = 313,
    sync_file_range = 314,
    tee = 315,
    vmsplice = 316,
    move_pages = 317,
    getcpu = 318,
    epoll_pwait = 319,
    utimensat = 320,
    signalfd = 321,
    timerfd_create = 322,
    eventfd = 323,
    fallocate = 324,
    timerfd_settime = 325,
    timerfd_gettime = 326,
    signalfd4 = 327,
    eventfd2 = 328,
    epoll_create1 = 329,
    dup3 = 330,
    pipe2 = 331,
    inotify_init1 = 332,
    preadv = 333,
    pwritev = 334,
    rt_tgsigqueueinfo = 335,
    perf_event_open = 336,
    fanotify_init = 337,
    fanotify_mark = 338,
    prlimit64 = 339,
    socket = 340,
    bind = 341,
    connect = 342,
    listen = 343,
    accept = 344,
    getsockname = 345,
    getpeername = 346,
    socketpair = 347,
    send = 348,
    sendto = 349,
    recv = 350,
    recvfrom = 351,
    shutdown = 352,
    setsockopt = 353,
    getsockopt = 354,
    sendmsg = 355,
    recvmsg = 356,
    recvmmsg = 357,
    accept4 = 358,
    name_to_handle_at = 359,
    open_by_handle_at = 360,
    clock_adjtime = 361,
    syncfs = 362,
    sendmmsg = 363,
    setns = 364,
    process_vm_readv = 365,
    process_vm_writev = 366,
    kcmp = 367,
    finit_module = 368,
    sched_getattr = 369,
    sched_setattr = 370,
    renameat2 = 371,
    seccomp = 372,
    getrandom = 373,
    memfd_create = 374,
    bpf = 375,
    execveat = 376,
    userfaultfd = 377,
    membarrier = 378,
    mlock2 = 379,
    copy_file_range = 380,
    preadv2 = 381,
    pwritev2 = 382,
    statx = 383,
    pkey_mprotect = 384,
    pkey_alloc = 385,
    pkey_free = 386,
    rseq = 387,
    sync_file_range2 = 388,
    semget = 393,
    semctl = 394,
    shmget = 395,
    shmctl = 396,
    shmat = 397,
    shmdt = 398,
    msgget = 399,
    msgsnd = 400,
    msgrcv = 401,
    msgctl = 402,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};

pub const Microblaze = enum(usize) {
    restart_syscall = 0,
    exit = 1,
    fork = 2,
    read = 3,
    write = 4,
    open = 5,
    close = 6,
    waitpid = 7,
    creat = 8,
    link = 9,
    unlink = 10,
    execve = 11,
    chdir = 12,
    time = 13,
    mknod = 14,
    chmod = 15,
    lchown = 16,
    @"break" = 17,
    oldstat = 18,
    lseek = 19,
    getpid = 20,
    mount = 21,
    umount = 22,
    setuid = 23,
    getuid = 24,
    stime = 25,
    ptrace = 26,
    alarm = 27,
    oldfstat = 28,
    pause = 29,
    utime = 30,
    stty = 31,
    gtty = 32,
    access = 33,
    nice = 34,
    ftime = 35,
    sync = 36,
    kill = 37,
    rename = 38,
    mkdir = 39,
    rmdir = 40,
    dup = 41,
    pipe = 42,
    times = 43,
    prof = 44,
    brk = 45,
    setgid = 46,
    getgid = 47,
    signal = 48,
    geteuid = 49,
    getegid = 50,
    acct = 51,
    umount2 = 52,
    lock = 53,
    ioctl = 54,
    fcntl = 55,
    mpx = 56,
    setpgid = 57,
    ulimit = 58,
    oldolduname = 59,
    umask = 60,
    chroot = 61,
    ustat = 62,
    dup2 = 63,
    getppid = 64,
    getpgrp = 65,
    setsid = 66,
    sigaction = 67,
    sgetmask = 68,
    ssetmask = 69,
    setreuid = 70,
    setregid = 71,
    sigsuspend = 72,
    sigpending = 73,
    sethostname = 74,
    setrlimit = 75,
    getrlimit = 76,
    getrusage = 77,
    gettimeofday = 78,
    settimeofday = 79,
    getgroups = 80,
    setgroups = 81,
    select = 82,
    symlink = 83,
    oldlstat = 84,
    readlink = 85,
    uselib = 86,
    swapon = 87,
    reboot = 88,
    readdir = 89,
    mmap = 90,
    munmap = 91,
    truncate = 92,
    ftruncate = 93,
    fchmod = 94,
    fchown = 95,
    getpriority = 96,
    setpriority = 97,
    profil = 98,
    statfs = 99,
    fstatfs = 100,
    ioperm = 101,
    socketcall = 102,
    syslog = 103,
    setitimer = 104,
    getitimer = 105,
    stat = 106,
    lstat = 107,
    fstat = 108,
    olduname = 109,
    iopl = 110,
    vhangup = 111,
    idle = 112,
    vm86old = 113,
    wait4 = 114,
    swapoff = 115,
    sysinfo = 116,
    ipc = 117,
    fsync = 118,
    sigreturn = 119,
    clone = 120,
    setdomainname = 121,
    uname = 122,
    modify_ldt = 123,
    adjtimex = 124,
    mprotect = 125,
    sigprocmask = 126,
    create_module = 127,
    init_module = 128,
    delete_module = 129,
    get_kernel_syms = 130,
    quotactl = 131,
    getpgid = 132,
    fchdir = 133,
    bdflush = 134,
    sysfs = 135,
    personality = 136,
    afs_syscall = 137,
    setfsuid = 138,
    setfsgid = 139,
    llseek = 140,
    getdents = 141,
    newselect = 142,
    flock = 143,
    msync = 144,
    readv = 145,
    writev = 146,
    getsid = 147,
    fdatasync = 148,
    sysctl = 149,
    mlock = 150,
    munlock = 151,
    mlockall = 152,
    munlockall = 153,
    sched_setparam = 154,
    sched_getparam = 155,
    sched_setscheduler = 156,
    sched_getscheduler = 157,
    sched_yield = 158,
    sched_get_priority_max = 159,
    sched_get_priority_min = 160,
    sched_rr_get_interval = 161,
    nanosleep = 162,
    mremap = 163,
    setresuid = 164,
    getresuid = 165,
    vm86 = 166,
    query_module = 167,
    poll = 168,
    nfsservctl = 169,
    setresgid = 170,
    getresgid = 171,
    prctl = 172,
    rt_sigreturn = 173,
    rt_sigaction = 174,
    rt_sigprocmask = 175,
    rt_sigpending = 176,
    rt_sigtimedwait = 177,
    rt_sigqueueinfo = 178,
    rt_sigsuspend = 179,
    pread64 = 180,
    pwrite64 = 181,
    chown = 182,
    getcwd = 183,
    capget = 184,
    capset = 185,
    sigaltstack = 186,
    sendfile = 187,
    getpmsg = 188,
    putpmsg = 189,
    vfork = 190,
    ugetrlimit = 191,
    mmap2 = 192,
    truncate64 = 193,
    ftruncate64 = 194,
    stat64 = 195,
    lstat64 = 196,
    fstat64 = 197,
    lchown32 = 198,
    getuid32 = 199,
    getgid32 = 200,
    geteuid32 = 201,
    getegid32 = 202,
    setreuid32 = 203,
    setregid32 = 204,
    getgroups32 = 205,
    setgroups32 = 206,
    fchown32 = 207,
    setresuid32 = 208,
    getresuid32 = 209,
    setresgid32 = 210,
    getresgid32 = 211,
    chown32 = 212,
    setuid32 = 213,
    setgid32 = 214,
    setfsuid32 = 215,
    setfsgid32 = 216,
    pivot_root = 217,
    mincore = 218,
    madvise = 219,
    getdents64 = 220,
    fcntl64 = 221,
    gettid = 224,
    readahead = 225,
    setxattr = 226,
    lsetxattr = 227,
    fsetxattr = 228,
    getxattr = 229,
    lgetxattr = 230,
    fgetxattr = 231,
    listxattr = 232,
    llistxattr = 233,
    flistxattr = 234,
    removexattr = 235,
    lremovexattr = 236,
    fremovexattr = 237,
    tkill = 238,
    sendfile64 = 239,
    futex = 240,
    sched_setaffinity = 241,
    sched_getaffinity = 242,
    set_thread_area = 243,
    get_thread_area = 244,
    io_setup = 245,
    io_destroy = 246,
    io_getevents = 247,
    io_submit = 248,
    io_cancel = 249,
    fadvise64 = 250,
    exit_group = 252,
    lookup_dcookie = 253,
    epoll_create = 254,
    epoll_ctl = 255,
    epoll_wait = 256,
    remap_file_pages = 257,
    set_tid_address = 258,
    timer_create = 259,
    timer_settime = 260,
    timer_gettime = 261,
    timer_getoverrun = 262,
    timer_delete = 263,
    clock_settime = 264,
    clock_gettime = 265,
    clock_getres = 266,
    clock_nanosleep = 267,
    statfs64 = 268,
    fstatfs64 = 269,
    tgkill = 270,
    utimes = 271,
    fadvise64_64 = 272,
    vserver = 273,
    mbind = 274,
    get_mempolicy = 275,
    set_mempolicy = 276,
    mq_open = 277,
    mq_unlink = 278,
    mq_timedsend = 279,
    mq_timedreceive = 280,
    mq_notify = 281,
    mq_getsetattr = 282,
    kexec_load = 283,
    waitid = 284,
    add_key = 286,
    request_key = 287,
    keyctl = 288,
    ioprio_set = 289,
    ioprio_get = 290,
    inotify_init = 291,
    inotify_add_watch = 292,
    inotify_rm_watch = 293,
    migrate_pages = 294,
    openat = 295,
    mkdirat = 296,
    mknodat = 297,
    fchownat = 298,
    futimesat = 299,
    fstatat64 = 300,
    unlinkat = 301,
    renameat = 302,
    linkat = 303,
    symlinkat = 304,
    readlinkat = 305,
    fchmodat = 306,
    faccessat = 307,
    pselect6 = 308,
    ppoll = 309,
    unshare = 310,
    set_robust_list = 311,
    get_robust_list = 312,
    splice = 313,
    sync_file_range = 314,
    tee = 315,
    vmsplice = 316,
    move_pages = 317,
    getcpu = 318,
    epoll_pwait = 319,
    utimensat = 320,
    signalfd = 321,
    timerfd_create = 322,
    eventfd = 323,
    fallocate = 324,
    semtimedop = 325,
    timerfd_settime = 326,
    timerfd_gettime = 327,
    semctl = 328,
    semget = 329,
    semop = 330,
    msgctl = 331,
    msgget = 332,
    msgrcv = 333,
    msgsnd = 334,
    shmat = 335,
    shmctl = 336,
    shmdt = 337,
    shmget = 338,
    signalfd4 = 339,
    eventfd2 = 340,
    epoll_create1 = 341,
    dup3 = 342,
    pipe2 = 343,
    inotify_init1 = 344,
    socket = 345,
    socketpair = 346,
    bind = 347,
    listen = 348,
    accept = 349,
    connect = 350,
    getsockname = 351,
    getpeername = 352,
    sendto = 353,
    send = 354,
    recvfrom = 355,
    recv = 356,
    setsockopt = 357,
    getsockopt = 358,
    shutdown = 359,
    sendmsg = 360,
    recvmsg = 361,
    accept4 = 362,
    preadv = 363,
    pwritev = 364,
    rt_tgsigqueueinfo = 365,
    perf_event_open = 366,
    recvmmsg = 367,
    fanotify_init = 368,
    fanotify_mark = 369,
    prlimit64 = 370,
    name_to_handle_at = 371,
    open_by_handle_at = 372,
    clock_adjtime = 373,
    syncfs = 374,
    setns = 375,
    sendmmsg = 376,
    process_vm_readv = 377,
    process_vm_writev = 378,
    kcmp = 379,
    finit_module = 380,
    sched_setattr = 381,
    sched_getattr = 382,
    renameat2 = 383,
    seccomp = 384,
    getrandom = 385,
    memfd_create = 386,
    bpf = 387,
    execveat = 388,
    userfaultfd = 389,
    membarrier = 390,
    mlock2 = 391,
    copy_file_range = 392,
    preadv2 = 393,
    pwritev2 = 394,
    pkey_mprotect = 395,
    pkey_alloc = 396,
    pkey_free = 397,
    statx = 398,
    io_pgetevents = 399,
    rseq = 400,
    clock_gettime64 = 403,
    clock_settime64 = 404,
    clock_adjtime64 = 405,
    clock_getres_time64 = 406,
    clock_nanosleep_time64 = 407,
    timer_gettime64 = 408,
    timer_settime64 = 409,
    timerfd_gettime64 = 410,
    timerfd_settime64 = 411,
    utimensat_time64 = 412,
    pselect6_time64 = 413,
    ppoll_time64 = 414,
    io_pgetevents_time64 = 416,
    recvmmsg_time64 = 417,
    mq_timedsend_time64 = 418,
    mq_timedreceive_time64 = 419,
    semtimedop_time64 = 420,
    rt_sigtimedwait_time64 = 421,
    futex_time64 = 422,
    sched_rr_get_interval_time64 = 423,
    pidfd_send_signal = 424,
    io_uring_setup = 425,
    io_uring_enter = 426,
    io_uring_register = 427,
    open_tree = 428,
    move_mount = 429,
    fsopen = 430,
    fsconfig = 431,
    fsmount = 432,
    fspick = 433,
    pidfd_open = 434,
    clone3 = 435,
    close_range = 436,
    openat2 = 437,
    pidfd_getfd = 438,
    faccessat2 = 439,
    process_madvise = 440,
    epoll_pwait2 = 441,
    mount_setattr = 442,
    quotactl_fd = 443,
    landlock_create_ruleset = 444,
    landlock_add_rule = 445,
    landlock_restrict_self = 446,
    process_mrelease = 448,
    futex_waitv = 449,
    set_mempolicy_home_node = 450,
    cachestat = 451,
    fchmodat2 = 452,
    map_shadow_stack = 453,
    futex_wake = 454,
    futex_wait = 455,
    futex_requeue = 456,
    statmount = 457,
    listmount = 458,
    lsm_get_self_attr = 459,
    lsm_set_self_attr = 460,
    lsm_list_modules = 461,
    mseal = 462,
    setxattrat = 463,
    getxattrat = 464,
    listxattrat = 465,
    removexattrat = 466,
    open_tree_attr = 467,
    file_getattr = 468,
    file_setattr = 469,
    listns = 470,
    rseq_slice_yield = 471,
};
